/*---------------------------------------------------------------------------*/
/*!
 * @file PCA_PP.cpp
 * @brief Multidimensionnal analysis and dimension reduction by Principal Component Analysis and Projection Pursuit
 *
 * This procedure  :
 * \li first analysis a multidimensionnal feature space composed by several images bands by Principal Component Analysis (PCA). It produces as an output the principal components. 
 * \li and second analysis a reduced multidimensionnal feature space composed by only principal components representing a signal energie percentage (fixed by the user)  by Projection Pursuit (PP). It produces as an output the extracted projections. This analysis requires quite a lot of computation power
 * 
 *
 * @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/
#include <iostream>
#include <string>  
#include <vector>
#include <fstream>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <ctime>

//format of the input images
#define PIX short int
#define inf 1e20

using namespace std;

/*! 	@brief  Usage function 	
	@param argv0 String containing the imput output file/paramters*/
void usage(string argv0) {
cout <<"*************************************************************************************************************"<<endl;
    cout << "\nusage: " << ">"
              << argv0 
              << " input_image_file_name output_PPimage_file_name output_PCAimage_file_name nb_lignes nb_colonnes PCAEnergiePercentage PPquantile" << endl; 
    cout << "\nOptions for cutting: " 
    	 << "offset_colonnes offset_lignes nb_colonnes_original"<< endl;
	 
    cout << "\nOptional parameters for Projection Pursuit : nb rand_departure, nb_maximum_of_iterations, epsilon,  factor_decreasing" <<endl;	
   
    cout << "\n Default parameters:\n nb rand_departure = 5\n nb_maximum_of_iterations = 50\n epsilon =1e-3\n factor_decreasing = 2\n quantile = 0.7\n PCAEnergiePercentage= 90\n"<<endl;
 
 cout <<"*************************************************************************************************************"<<endl;
  
}

//auxiliary functions
/*! @brief  Evaluation of the projection index by a chi-square distance 
    @param x  bidimensionnal data
    @param lenght sample size
    @return projection index*/
float Evaluate_PI(float **x, unsigned long int lenght);

/*! @brief  Procedure searching an orthogonal base to the found projection  
    @param plan_init  Found projection plan
    @param new_base Orthogonal base returned   
    @param nb_dim Space dimensionnality */
void base_ortho_plan_finder(float **plan_init,float **new_base,unsigned long int nb_dim);

/*! @brief  Procedure projecting the data in a new orthogonal base  
    @param Y  input Data and output projected data
    @param new_base Orthogonal base for projection 
    @param dim number of samples   
    @param nb_dim Space dimensionnality */
void projec_on_base(float **Y,float **newbase,unsigned long int dim, unsigned long int nb_dim);

/*! @brief  Procedure projecting the data in the original orthogonal base 
    @param Y  input Data and output projected data
    @param new_base Old Orthogonal base for projection 
    @param dim number of samples   
    @param nb_dim Space dimensionnality  */
void projec_on_oldbase(float **Y,float **newbase,unsigned long int dim, unsigned long int nb_dim);

/*! @brief  Procedure desctructuring : transforming the distribution so that it becomes a Normal distribution
    @param Y  bidimensionnal data input/output
    @param dim sample size*/
void destructuration(float **Y,unsigned long int dim);

/*! @brief  Returns the sum of an aray
    @param a  array of float
    @param lenght sum*/
double Total(float *a, unsigned long int lenght);

//numerical recepies
/*! @brief  imported from Numerical Reciepes functions. See Numerical Reciepes documentation for further details */
void eigsrt(float d[], float **v, int n);
/*! @brief  imported from Numerical Reciepes functions. See Numerical Reciepes documentation for further details */
void jacobi(float **a, int n, float d[], float **v, int *nrot);
/*! @brief  imported from Numerical Reciepes functions. See Numerical Reciepes documentation for further details */
float gasdev(long *idum); 
/*! @brief  imported from Numerical Reciepes functions : Normal(0,1) distributed random number generator See Numerical Reciepes documentation for further details */
void gaussj(float **a, int n, float **b, int m);
	
/*!	@brief Main procedure where PCA and PP analysis are chained
*
*After a step of memory allocation and initialisation , the input data is read according to the cutting and sampling input parameters.\n Then, the data is analysed by PCA. The analyse relies on finding the eigenvalues and eigenvectors of the data covariance matrix. The eigenvalue decompositionis performed using Numerical recepies function " jacobi(...)" and "eigsrt(...)". The principal components are then saved on the disk.\n
The PP optimisation procedure is then launch on a reduced space composed by principal components comprising a user provided signal energy percentage (input parameter "PCAEnergiePercentage"):
\li 1) first, a plan (bi-dimenionnal projection) is initialized randomly and the data is projected on it. The non-Gaussianity of the distribution within this plan is  is evaluated by the chi-2 projection index function :"Evaluate_PI(...)".
\li 2) then, the search of the projection index minima is performed simultaneously by 4 different sub-routines, by an iterative algorithm. For each sub routine, 2 different moves in the feature space are generated  randomly;  they reveal, in parallel, 2 different plans, on which the data is projected and the non-Gaussianity of the projections  are simultaneously evaluated; the best candidate between the two moves is preserved for each of the 4 sub-routines and the procedure is iterated.\n
The scale of the search reduces while iterating  the algorithm, for 3 of the 4 sub-routines, by deviding by a given factor (user provided : input parameter "factor_decreasing") the importance of the moves. The change of the scale search is done when the number of iteration reaches a limit fixed by the user   (user provided : input parameter "nb_maximum_of_iterations"). But, 1 of the 4 search is performed always using the same search scale. Once the scale has changed, the algorithm is reiterated with this new search scale.  The algorithm stops iterating when the size of the move is lower than a given value (user provided : input parameter "epsilon").
\li 3) The best projection plan found is then saved on the disk and the structure found in the projection is removed : the data is projected in an orthogonal based comprising the bidimensionnal plan; the distribution in this plan is then transform into a Gaussian normalized distribution and the data is projected back in the original feature space. Note, that even if the data has been modified  by structure removals, the saved projections are not affected because the are generated from the original data.\n
 The step 2) is then reiterated a given number of times (user provided : input parameter "rand_departure")
\li 4) While the projection index limit  provided by the user (input parameter "quantile") is not reached, that is to stay while intersting projection remain in the feature space, Step 2) and 3) are iterated. 
	@param argv0[1]  Text file name of input image bands 
*    	@param argv0[2]  Text file name of output image bands created by PP 
*	@param argv0[3]  Text file name of output image bands created by PCA which represent a given energie percentage 
*	@param argv0[4]  Number of lines of the images
*	@param argv0[5]  Number of columns of the images 
*	@param argv0[6]  Signal energie percentage to preserve with PCA 
	@param argv0[7]  Quantile : index projection limit to stop the search 
*	
*	Optional parameters for selecting a spatial window  :
*	@param argv0[8]  Column offset when reading the image file 
*	@param argv0[9]  Line offset when reading the image file
*	@param argv0[10]  Number of columns of the image file 
*
*	Optional parameters for Projection Pursuit  :
*	@param argv0[11]  Number of rand_departures 
*	@param argv0[12]  Maximum number of iterations
*	@param argv0[13]  Precision of the convergence (epsilon)
*	@param argv0[14]  Factor decreasing the scale of the search at each iteration
*/	
int main(int argc, char *argv []){   
  
srand((unsigned)time(NULL));

string ifn1;
string ifn2;
string ifn3;

unsigned long int	    dimx           = 0;
unsigned long int	    dimy	   = 0;
int		            ofx		   = 0;
int         		    ofy		   = 0;
int		            Dy	 	   = 0;
float 			    quantile	   = 0.7;
float 			    keepEnergie	   = 90.;
//Default parameters....

float eigenthreshold=0.000001;
int nb_rand_departure=5;//nb of random departure 
float cbegin=tan(80*M_PI/180);//c initial value
float eps=1e-3;//limit of c
float factor=2;//factor which devides c[1,2,3] for each iteration
float factor_cons=1;//factor which devides c[0] only once
int nb_step, nb_max_step=50;//nb max of iteration before c=c\2

switch (argc-1) {

	 case 14:  factor		= atof(argv[14]);
	 case 13:  eps  		= atof(argv[13]);
 	 case 12:  nb_max_step		= atoi(argv[12]);
         case 11:  nb_rand_departure	= atoi(argv[11]);
	 case 10:  Dy			= atoi(argv[10]);
         case 9:  ofy 			= atoi(argv[9]);
	 case 8:  ofx 			= atoi(argv[8]);
	 case 7:  quantile  		= atof(argv[7]);
	 case 6:  keepEnergie	 	= atof(argv[6]);
	 case 5:  dimy       		= atoi(argv[5]);
         case 4:  dimx       		= atoi(argv[4]);
         case 3:  ifn3	 		= string(argv[3]);
	 case 2:  ifn2	 		= string(argv[2]);
         case 1:  ifn1	 		= string(argv[1]); break;
         case 0:
        default: usage(argv[0]); return -1;
    }
    if ((argc-1) < 8) {Dy=dimx;} 
    if ((argc-1) < 5) { usage(argv[0]); return -1;}
    
     cout<<"\n*************************************************************************************************************"<<endl;cout<<"Multidimensionnal analysis and dimension reduction by Principal Component Analysis and Projection Pursuit\n Author : Patrick HEAS  2005"<<endl;cout<<"*************************************************************************************************************\n"<<endl;
    
    cout << "> " << argv[0] 
              << " " << ifn1 
              << " " << ifn2
	      << " " << ifn3 
	      << " " << dimx 
              << " " << dimy 
	      << " " << keepEnergie 
	      << " " << quantile
	      << " " << ofx
	      << " " << ofy  
              << " " << Dy 	
              << " " << nb_rand_departure 
              << " " << nb_max_step 
	      << " " << eps
	      << " " << factor << endl;
	


/*************************************************************************************************************/

time_t tdeb, t1,t2;
time(&tdeb);
t1=tdeb;


unsigned long  int i=0,j=0,k=0,l=0,m=0,n=0;;
unsigned long  int dim=dimx*dimy;
 
unsigned long int nb_bands=1;

string line;
ifstream inaux;
inaux.open(ifn1.c_str(),ifstream::in);
ifstream in;
in.open(ifn1.c_str(),ifstream::in);
	
n=0;
while(getline(inaux,line)) {n++;} 	
unsigned long  int nb_images=n, nb_dim=n;
cout << "number of images : " << nb_images << std::endl;

char ***path= new char **[nb_images];
for(m=0;m<nb_images;++m){path[m] =new char*[1];} 
for(m=0;m<nb_images;++m){for(k=0;k<1;++k){path[m][k]=new char[100];}}
char ***save_path= new char **[(int)nb_images/2];
for(m=0;m<(unsigned long int)nb_images/2;++m){save_path[m] =new char*[2];} 
for(m=0;m<(unsigned long int)nb_images/2;++m){for(k=0;(int)k<2;++k){save_path[m][k]=new char[100];}}
char ***save_path2= new char **[nb_images];
for(m=0;m<nb_images;++m){save_path2[m] =new char*[1];} 
for(m=0;m<nb_images;++m){for(k=0;k<1;++k){save_path2[m][k]=new char[100];}}

/*******************************Reading text file names*********************************************/

for (i=0;i<(unsigned long int)nb_images;i++){
	for (j=0;j<(unsigned long int)nb_bands;j++){
 		   	getline(in,line);
		  	strcpy(path[i][j],(char *)line.c_str());
 	}	
}

string line1;
ifstream inaux1;
inaux1.open(ifn2.c_str(),ifstream::in);
ifstream in1;
in1.open(ifn2.c_str(),ifstream::in);
	
cout <<" PP Save Files :"<<endl;				
for (i=0;i<(unsigned long int)nb_images/2;i++){
	for (j=0;j<2;j++){
 		   	getline(in1,line1);
		  	strcpy(save_path[i][j],(char *)line1.c_str());
		  	std::cout<<save_path[i][j]<<std::endl;
 	}	
}
string line2;
ifstream inaux2;
inaux2.open(ifn3.c_str(),ifstream::in);
ifstream in2;
in2.open(ifn3.c_str(),ifstream::in);

cout<<" PCA Save Files :"<<endl;				
for (i=0;i<(unsigned long int)nb_images;i++){
	for (j=0;j<(unsigned long int)nb_bands;j++){
 		   	getline(in2,line2);
		  	strcpy(save_path2[i][j],(char *)line2.c_str());
		  	std::cout<<save_path2[i][j]<<std::endl;
 	}	
}


/*******************************Reading the images*********************************************/
cout<<"\n";
printf("Reading images\n");
 PIX a[dimx][dimy]    ;                                            
       float **clust= new float *[nb_dim]; //vector clust contains the data
        for(m=0;m<nb_dim;++m)
             clust[m] =new float[dim];

	 char * ifn;
         FILE * ifp;
	 m=0;n=0;
	 for (l = 0 ;l <nb_dim; l++){
	   ifn=path[m][n];printf("%s\n",ifn); // Reading  dimensions R, G and B
	   ifp=NULL;
	   assert (ifp=fopen(ifn,"r"));
	   assert (0==fseek(ifp,(sizeof(PIX)*Dy*ofy+sizeof(PIX)*ofx), SEEK_SET));
	   assert (dimy==fread(&a[0][0],sizeof (PIX),dimy,ifp));    
	   for (i=1;i<dimx;i++){
	     assert (0==fseek(ifp,(sizeof(PIX)*(Dy-dimy)), SEEK_CUR));
	     assert (dimy==fread(&a[i][0],sizeof (PIX),dimy,ifp));
	   }
	   fclose(ifp); 
	   if (n==nb_bands-1){n=0;m++;}
	   else{n++;}
	   for (i = 0 ;i < (dimx); i++){ k=(dimy)*i;          
	     for (j = 0; j <(dimy); j++) { 
	       clust[l][k]=a[i][j];
	       k=k+1;
	     }
	   }
	 }  
	
/*********************************Spherisation of the data*************************************************/
cout<<"\nDecorralating procedure"<<endl;	

float **Y= new float *[nb_dim]; //    Sphered data 
for(m=0;m<nb_dim;++m)
{ Y[m]=new float[dim];}
float **Y1= new float *[nb_dim]; //    Sphered data bis
for(m=0;m<nb_dim;++m)
{ Y1[m]=new float[dim];}
 float *sauvACP= new float [dim];//vector to save
float *moy= new float [nb_dim]; //    mean vector 
float *M= new float [nb_dim];  //   Normalizing diag Matrix 
float *eigenvalues= new float [nb_dim+1];//eigen values (index belong to [1;nb_dim])
float **cov= new float *[nb_dim+1];//    covariance Matrix (index belong to [1;nb_dim][1;nb_dim])
float **eigenvectors= new float *[nb_dim+1];//eigen vectors(index belong to [1;nb_dim][1;nb_dim])
for(m=0;m<nb_dim+1;++m)
{cov[m]=new float[nb_dim+1]; 
 eigenvectors[m]=new float[nb_dim+1]; }

int *nrot= new int [1];
double truc;


//printf("Mean vector\n");
for (l=0;l<nb_dim;l++)
{  truc=Total(clust[l],dim);
   moy[l]=(float)(truc/dim);
   M[l]=0;   
   for (i=0;i<dim;i++){
      M[l]=((clust[l][i]-moy[l])*(clust[l][i]-moy[l]))+M[l];
   }
   M[l]= sqrt(1/((double)dim)*M[l]);
  // printf("%f\t",M[l]);printf("\n");
} 

//printf("Covariance Matrix");
for (l=0;l<nb_dim;l++){//printf("\n");
  for (k=l;k<nb_dim;k++){
    cov[l+1][k+1]=0;
    for (i=0;i<dim;i++){
      cov[l+1][k+1]=(clust[l][i]-moy[l])*(clust[k][i]-moy[k])+cov[l+1][k+1];
    }
    cov[l+1][k+1]=cov[l+1][k+1]/(dim*M[l]*M[k]);
    cov[k+1][l+1]=cov[l+1][k+1];
   // printf("%f\t",cov[l+1][k+1]);
  }
} 

jacobi(cov,nb_dim,eigenvalues,eigenvectors,nrot);//compute eigensysteme
eigsrt(eigenvalues,eigenvectors,nb_dim);//sort eigen values and vectors

 

for (i=0;i<dim;i++){
  for (l=0;l<nb_dim;l++){ Y[l][i]=0;
    for (k=0;k<nb_dim;k++){
      Y[l][i]= Y[l][i]+(clust[k][i]-moy[k])/M[k]*eigenvectors[k+1][l+1];
    }
    Y[l][i]=Y[l][i]/(sqrt(eigenvalues[l+1]));Y1[l][i]=Y[l][i];
  }
}

/*Keeping eigen values representing more than "keepEnergie" pourcent of the signal energie*/
float TotalEnergie=0;
for (l=1;l<nb_dim+1;l++){
	TotalEnergie+=eigenvalues[l];
}
float Energie=0;
for (l=1;l<nb_dim+1;l++){
	Energie+=eigenvalues[l];
	if (Energie/TotalEnergie*100.>keepEnergie){eigenthreshold=eigenvalues[l];break;}
}

int dim_less=0;
for (l=1;l<nb_dim+1;l++)
  {
  //cout<<eigenvalues[l]<<" "<<eigenthreshold<<endl;
  if (eigenvalues[l]<eigenthreshold){ dim_less++; }
  }
nb_dim=nb_dim-dim_less;// Component not very descriminent which have a eigenvalue under eigenthreshold
cout<<"Number of component representing "<<keepEnergie<<" of the signal energie : "<<nb_dim<<endl; 
                      // These dimensions are removed and we work now with b_dim-dim_less dimensions


printf("saving PCA results\n");
for (j=0;j<nb_dim;j++){
  for (i=0;i<dim;i++){sauvACP[i]=(float)Y[j][i];} 
  ifp=NULL; 
  ifn=save_path2[j][0];
  assert (ifp=fopen(ifn,"wb")); 
  assert ( dim==fwrite(&sauvACP[0], sizeof(float),dim,ifp)); 
  assert (EOF!=fclose(ifp));
}

for(m=0;m<nb_dim;++m)
            { delete [] cov[m];
              delete [] eigenvectors[m] ;
	    }
//delete [] cov ;
delete [] eigenvectors;
//delete [] moy;
//delete [] M;
delete [] eigenvalues;
delete [] sauvACP;

/*********************************************************************************************************/
/*********************************Global Optimisation **************************************************/

cout<<"\nInitialisation of optimisation procedure"<<endl;
int nb_proj=0;//number of random departure
int rand_dep;
float *temp1 = new float[4];float *temp2 = new float[4];
float tmp,aux1,aux2, p_i, p_i_old,p_i_opt;
float *c = new float[4];//factor fixing the size of the neighberhood visited
float **new_base = new float*[nb_dim];//orthonormal base ortho gonal to the plan for structure removal
for (m=0;m<nb_dim;++m){new_base[m]= new float[nb_dim];}
float **v_random = new float*[4];
for (m=0;m<4;++m){v_random[m]= new float[nb_dim];}
float **plan_init= new float*[2];
for (m=0;m<2;++m){ plan_init[m]= new float[nb_dim];} 
float **ab1= new float*[8];    
float **ab2= new float*[8];
for (m=0;m<8;++m){  ab1[m]= new float[nb_dim]; ab2[m]= new float[nb_dim];}
float **projec_plan= new float*[8];
for (m=0;m<8;++m){ projec_plan[m]= new float[dim]; }
float **arg= new float*[2];
for (m=0;m<2;++m){ arg[m]= new float[dim]; }
float *sauv= new float [2*dim];//vector to save

int sub,skip=0;
p_i=1000;
p_i_opt=p_i;

/*****************************************New projection research*********************************/
printf("Optimisation procedure\n");
while ((p_i_opt>=quantile) && (nb_proj<((int)(nb_images/2)))){

nb_proj++ ;
printf("%s %d\n","index of current significant projection",nb_proj);
p_i_opt=-1;nb_step=0;
for (rand_dep=0;rand_dep<nb_rand_departure;rand_dep++) // New random departure.........
{
//printf("new random departure\n");
 if (skip==0){
//initialisation of a random plan (alpha, beta)
tmp=0;
 for (l=0;l<nb_dim;l++){plan_init[0][l]=(((float)rand())/RAND_MAX-0.5);}
 for (l=0;l<nb_dim;l++){tmp=plan_init[0][l]*plan_init[0][l]+tmp;}
 tmp=sqrt(tmp);
 for (l=0;l<nb_dim;l++){
 plan_init[0][l]=plan_init[0][l]/tmp;}//unitary condition 

tmp=0;//beta
int last=(int)((((float)rand())/RAND_MAX)*nb_dim-0.5);
for (l=0;l<nb_dim;l++){plan_init[1][l]=(((float)rand())/RAND_MAX-0.5);} 
for (l=0;l<nb_dim;l++){tmp=plan_init[0][l]*plan_init[1][l]+tmp;}
tmp=tmp-plan_init[0][last]*plan_init[1][last];
plan_init[1][last]=-1.*tmp/(plan_init[0][last]);//orthogonal condition
tmp=0;
for (l=0;l<nb_dim;l++){tmp=plan_init[1][l]*plan_init[1][l]+tmp;}
tmp=sqrt(tmp);
for (l=0;l<nb_dim;l++){plan_init[1][l]=plan_init[1][l]/tmp;}//unitary condition
 }
//projection on the plan of the initial solution


for (i=0;i<dim;i++) 
{
  arg[0][i]=0;arg[1][i]=0;
  for (l=0;l<nb_dim;l++)
  {   
    arg[0][i]=Y[l][i]*plan_init[0][l]+arg[0][i]; 
    arg[1][i]=Y[l][i]*plan_init[1][l]+arg[1][i];
   }
}


p_i=Evaluate_PI(arg,dim);p_i_old=p_i;// chi2
//printf("%s %e\n","Projection indice initial=",p_i);

//searching the maxima.........................................................................

// 4 parallele sub_routines.....
for(i=0;i<4;i++){c[i]=cbegin;}//initaialisating at tan(80)=5.67
while (c[1]>eps)
  {// printf("%e\n",c[1]);

  //initialisation of 4 random vector v
  for(sub=0;sub<4;sub++)
  {
  tmp=0;
  for (l=0;l<nb_dim;l++){ v_random[sub][l]=(((float)rand())/RAND_MAX);}
  for (l=0;l<nb_dim;l++){tmp= v_random[sub][l]*v_random[sub][l]+tmp;}
  tmp=sqrt(tmp);
  for (l=0;l<nb_dim;l++){ v_random[sub][l]= v_random[sub][l]/tmp;}//unitary condition
  }

  //calculation of 2 couple of  vectors for the 2 different plans for EACH sub-routine
  
  temp1[0]=0;temp1[1]=0;temp1[2]=0;temp1[3]=0;
  temp2[0]=0;temp2[1]=0;temp2[2]=0;temp2[3]=0;
  for (i=0;i<nb_dim;i++)
    {
    for (sub=0;sub<4;sub++)
      {
      ab1[sub*2][i]=plan_init[1][i]+c[sub]*v_random[sub][i]; 
      ab2[sub*2][i]=plan_init[1][i]-c[sub]*v_random[sub][i];
      temp1[sub]=temp1[sub]+ab1[sub*2][i]*ab1[sub*2][i];
      temp2[sub]=temp2[sub]+ab2[sub*2][i]*ab2[sub*2][i];
      }
    }
  for (sub=0;sub<4;sub++)  
    {
      temp1[sub]=sqrt(temp1[sub]);temp2[sub]=sqrt(temp2[sub]); 
      for (i=0;i<nb_dim;i++) 
	{ ab1[2*sub][i]=ab1[2*sub][i]/temp1[sub]; 
	ab2[2*sub][i]=ab2[2*sub][i]/temp2[sub];
	}
    }
  
   
   for (sub=0;sub<4;sub++) 
    {
    temp1[0]=0;temp1[1]=0;temp1[2]=0;temp1[3]=0;
    temp2[0]=0;temp2[1]=0;temp2[2]=0;temp2[3]=0; 
    for (i=0;i<nb_dim;i++)
      {temp1[sub]=temp1[sub]+ab1[2*sub][i]*plan_init[0][i];
      temp2[sub]=temp2[sub]+ab2[2*sub][i]*plan_init[0][i];
      }
    aux1=0;aux2=0;
    for (i=0;i<nb_dim;i++)
      {
	ab1[2*sub+1][i]=plan_init[0][i]-temp1[sub]*ab1[2*sub][i];
	ab2[2*sub+1][i]=plan_init[0][i]-temp2[sub]*ab2[2*sub][i];
	aux1=aux1+ab1[2*sub+1][i]*ab1[2*sub+1][i];
	aux2=aux2+ab2[2*sub+1][i]*ab2[2*sub+1][i];
      }
    aux1=sqrt(aux1);
    aux2=sqrt(aux2);
    for (i=0;i<nb_dim;i++) { ab1[2*sub+1][i]= ab1[2*sub+1][i]/aux1; ab2[2*sub+1][i]= ab2[2*sub+1][i]/aux2;

    }  
    }

  //projection on the plan  defined by the first candidate for EACH sub-routine

      for (i=0;i<dim;i++)
	{
	  for (sub=0;sub<4;sub++)  
	    {
	      projec_plan[2*sub][i]=0;projec_plan[2*sub+1][i]=0;
	      for (l=0;l<nb_dim;l++)
		{
		  projec_plan[2*sub][i]=Y[l][i]*ab1[2*sub+1][l]+projec_plan[2*sub][i]; 
		  projec_plan[2*sub+1][i]=Y[l][i]*ab1[2*sub][l]+projec_plan[2*sub+1][i]; 
		}
	    }
	}
    
     tmp=Evaluate_PI(projec_plan,dim);
     if (tmp>p_i)
       {p_i=tmp;p_i_old=p_i;nb_step=0; 
       for (l=0;l<nb_dim;l++){plan_init[0][l]=ab1[0][l];plan_init[1][l]=ab1[1][l];}
       }
     for (sub=1;sub<4;sub++)  
       {
	 for (i=0;i<dim;i++){arg[0][i]= projec_plan[2*sub][i];arg[1][i]= projec_plan[2*sub+1][i];}
	 tmp=Evaluate_PI(arg,dim);
	 if (tmp>p_i)
	   {p_i=tmp;p_i_old=p_i;nb_step=0;
	   for (l=0;l<nb_dim;l++){plan_init[0][l]=ab1[2*sub][l];plan_init[1][l]=ab1[2*sub+1][l];}
	   }
       }
  //projection on the plan defined by the second candidate for EACH sub-routine

      for (i=0;i<dim;i++)
	{
	  for (sub=0;sub<4;sub++)  
	    {
	      projec_plan[2*sub][i]=0;projec_plan[2*sub+1][i]=0;
	      for (l=0;l<nb_dim;l++)
		{
		  projec_plan[2*sub][i]=Y[l][i]*ab2[2*sub+1][l]+projec_plan[2*sub][i]; 
		  projec_plan[2*sub+1][i]=Y[l][i]*ab2[2*sub][l]+projec_plan[2*sub+1][i]; 
		}
	    }
	}
     tmp=Evaluate_PI(projec_plan,dim);
     if (tmp>p_i)
       {p_i=tmp;p_i_old=p_i;nb_step=0;
        for (l=0;l<nb_dim;l++){plan_init[0][l]=ab2[0][l];plan_init[1][l]=ab2[1][l];}
       }
     for (sub=1;sub<4;sub++)  
       {
	 for (i=0;i<dim;i++){arg[0][i]= projec_plan[2*sub][i];arg[1][i]= projec_plan[2*sub+1][i];}
	 tmp=Evaluate_PI(arg,dim);
	 if (tmp>p_i)
	   {p_i=tmp;p_i_old=p_i;nb_step=0; 
	   for (l=0;l<nb_dim;l++){plan_init[0][l]=ab2[2*sub][l];plan_init[1][l]=ab2[2*sub+1][l];}
	   }
       }
     if ((p_i==p_i_old)&&(nb_step<=nb_max_step)){nb_step++;}
     if ((p_i==p_i_old)&&(nb_step>nb_max_step)) {c[0]=cbegin/factor_cons;c[1]=c[1]/factor;c[2]=c[2]/factor;c[3]=c[3]/factor;nb_step=0;}

    

} //end while loop
/*********************************************************************************************************/

/*****************************saving best projection of the random departures********************************/
//projection on the plan of the final solution
printf("%s %e\n","Intermediate projection index=",p_i);
if (p_i>p_i_opt) {
p_i_opt=p_i;
 


for (i=0;i<dim;i++)
{
   arg[0][i]=0;arg[1][i]=0;
   for (l=0;l<nb_dim;l++)
   {
     arg[0][i]=Y1[l][i]*plan_init[0][l]+arg[0][i];
     arg[1][i]=Y1[l][i]*plan_init[1][l]+arg[1][i]; 
   }
}   

if (nb_proj==((int)(nb_images/2-1))){
	for (i=0;i<dim;i++)
	{
     	arg[0][i]=Y1[0][i];
     	arg[1][i]=Y1[1][i];   
	}   
}
for (i=0;i<dim;i++){
  for (l=0;l<2;l++){sauv[l*nb_dim+i]=(float)arg[l][i];}}
ifp=NULL; 
ifn=save_path[nb_proj-1][0];
assert (ifp=fopen(ifn,"wb"));
assert ( dim==fwrite(&sauv[0], sizeof(float),dim,ifp));
assert (EOF!=fclose(ifp));

for (i=0;i<dim;i++){
  for (l=0;l<2;l++){sauv[l*nb_dim+i]=(float)arg[1-l][i];}}

ifp=NULL; 
ifn=save_path[nb_proj-1][1];
assert (ifp=fopen(ifn,"wb"));
assert ( dim==fwrite(&sauv[0], sizeof(float),dim,ifp));
assert (EOF!=fclose(ifp));
}// end overwrite  
skip=0;
} //end for loop of new random departure
printf("%s %e\n","Final projection index=",p_i_opt);

/*********************************************************************************************************/

/*************************Structure removal or dimension removal*******************/
base_ortho_plan_finder(plan_init,new_base,nb_dim);
projec_on_base(Y,new_base,dim,nb_dim);
destructuration(Y,dim);
projec_on_oldbase(Y,new_base,dim,nb_dim);


skip=1;
/**********************************************************************************************************/
}//end of while p_i>=quantile loop
printf("All the interesting projections have been found...\n");
time(&t2);
cout<< "Processing time duration : "<<t2-t1<<" seconds."<<endl;
return 0;
/************************************End of projection research******************************************/
}

/**************************************Auxiliary Fonction **************************************************/


/******************************CHI2 index************************************/
/*!@brief Evaluation of the projection index of a given plan.
*
*The non-Gaussianity of the plan distribution is evaluated using a chi2 projection index distance between the data distribution and a Gaussian normalized distribution. The integration requiered for this distance evaluation (see Theoritical note) is performed by deviding the integration domain into 432 small boxes. For further details, we refer the reader to Posse (1993) thesis.
@param x bidimensionnal data
@param dim number of data samples
@return projection index*/
float Evaluate_PI(float **x, unsigned long int dim)
{
float res=0, p_i=0;
unsigned long int *box=new unsigned long int[433];//index from 1 to 432
unsigned long int i,j,k,ir,aux;
float pi_help,xt1,xt2,r_i,theta_i,help,R;
R=1.25;
float *Ftheo= new float [6];// density in the boxes of the theorical normal distribution (normalized)
Ftheo[0]=0.00387949;
Ftheo[1]=0.010849;
Ftheo[2]=0.0159186;
Ftheo[3]=0.0184859;
Ftheo[4]=0.0186418;
Ftheo[5]=0.0572252;

for(k=0;k<433;k++)
  {box[k]=0;}
for(i=0;i<dim;i++)
  {
  xt1=x[0][i];xt2=x[1][i];
  r_i=sqrt(xt1*xt1+xt2*xt2);
  aux=(int)(5*r_i/R);
  if (aux<5) {j=72*aux+1;}
  else {j=72*5+1;}
  theta_i=atan(xt2/xt1);
  if (xt1<0){theta_i=theta_i+M_PI;}
  else if (xt2<0){theta_i=theta_i+2*M_PI;}
  k=j +(int)(theta_i*36/M_PI);
  box[k]=box[k]+1;
  }

for (i=1;i<7;i++)
  {
    ir=(i-1)*72;pi_help=0;
    help=box[ir+1]+box[ir+2]+box[ir+3]+box[ir+4]+box[ir+5]+box[ir+6]+box[ir+7]+box[ir+8]+box[ir+9];
    pi_help=pi_help+help*help;
    for (j=1;j<64;j++) 
      {
       help=help-box[ir+j]+box[ir+j+9];
       pi_help=pi_help+ help*help;
      }
    for (j=1;j<9;j++) 

      {
       help=help-box[ir+j+63]+box[ir+j];
       pi_help=pi_help+ help*help;
      }
    p_i=p_i+pi_help/(Ftheo[i-1]);
  }
p_i=p_i/(9*(float)dim*(float)dim)-1;
res=p_i; 
delete [] box;
delete [] Ftheo;
return res;
}

/*********************************************************************************************************/
/*!@brief Method finding the orthogonal base comprising a given plan.
*
*The analysis is based on Gauss Jordan linear equation solution. the analysis uses Numerical recepies function
@param plan_init input plan that will be contained in the orthogonal base 
@param new_base output orthogonal base 
@param nb_dim number of dimension of the base*/
void base_ortho_plan_finder(float **plan_init,float **new_base,unsigned long int nb_dim)
{
unsigned long int l,m,k;
float tmp;
//using Gauss Jordan linear equation solution
for (l=0;l<nb_dim;l++){new_base[0][l]=plan_init[0][l];new_base[1][l]=plan_init[1][l];}
for (m=2;m<nb_dim;m++)
  {
  tmp=0;
  for (l=m;l<nb_dim;l++)
    {
    new_base[m][l]=(((float)rand())/RAND_MAX) ;//fixing degree of freedom 
    }
  }
float **a= new float*[nb_dim+1];
for (m=0;m<nb_dim+1;++m){a[m]= new float[nb_dim+1];} 
 float **b= new float*[nb_dim+1];for (m=0;m<nb_dim+1;++m){b[m]=new float[2];}
for (m=2;m<nb_dim;++m)//nb_dim-2 equation systeme to be solved
{
  //defining the systeme to be solved
  for (l=0;l<m;l++)
    {
    b[l+1][1]=0;//zero block
    for (k=0;k<nb_dim;k++)
      {
	a[l+1][k+1]=(float)new_base[l][k];//already fixed vectors
      }
    }
  for (l=m;l<nb_dim;l++)
    {
    b[l+1][1]=(float) new_base[m][l];//block fixed randomly
    for (k=0;k<nb_dim;k++)
      {
      a[l+1][k+1]=0;//zero block
      }
    a[l+1][l+1]=1;//unity block
    }
  //solving the system
  gaussj(a,nb_dim,b,1);
  for (k=0;k<nb_dim;k++){ new_base[m][k]=(float)b[k+1][1];}//filling the base with one vector more
  tmp=0;
  for (l=0;l<nb_dim;l++){tmp=new_base[m][l]*new_base[m][l]+tmp;}
  tmp=sqrt(tmp);
  for (l=0;l<nb_dim;l++){new_base[m][l]=new_base[m][l]/tmp;}//normalizing the new vector
}

for (m=0;m<nb_dim+1;++m){delete [] a[m];delete [] b[m];}
delete [] a;delete [] b;
}
/*********************************************************************************************************/
/*!@brief Method projecting the data on a given orthogonal base.
*
@param Y input data to project
@param newbase  base for projection 
@param dim number of data samples 
@param nb_dim number of dimension of the base*/
void projec_on_base(float **Y,float **newbase,unsigned long int dim,unsigned long int nb_dim)
{
unsigned long int i,m,l;
float **arg= new float *[nb_dim]; 
for(m=0;m<nb_dim;++m)
{ arg[m]=new float[dim];}
for (i=0;i<dim;i++)
  {
    for (m=0;m<nb_dim;m++)
      {	 
	arg[m][i]=0;
	for (l=0;l<nb_dim;l++){arg[m][i]=Y[l][i]*newbase[m][l]+arg[m][i];}  	  
      } 
  }
for (i=0;i<dim;i++) {for (m=0;m<nb_dim;m++){Y[m][i]=arg[m][i];}}

for (m=0;m<nb_dim;++m){delete [] arg[m];}
delete [] arg;
}
/*********************************************************************************************************/
/*!@brief Method projecting the data in a new coordinate system in  the old orthogonal base.
*
*the anlysis is based on a matrix inversion. the analysis uses Numerical recepies function
@param Y input data  in a new coordinate system to project
@param newbase  old base for projection 
@param dim number of data samples 
@param nb_dim number of dimension of the base*/
void projec_on_oldbase(float **Y,float **newbase,unsigned long int dim,unsigned long int nb_dim)
{
//defining old base in new base coordinates
unsigned long int i,m,l;
float **arg= new float *[nb_dim]; 
for(m=0;m<nb_dim;++m){ arg[m]=new float[dim];}
float **a= new float*[nb_dim+1];float **b= new float*[nb_dim+1];
for (m=0;m<nb_dim+1;++m){a[m]= new float[nb_dim+1];b[m]= new float[nb_dim+1];} 

//matrix inverse
for(m=0;m<nb_dim;m++){ for(l=0;l<nb_dim;l++){ a[m+1][l+1]=newbase[m][l];b[m+1][l+1]=0;}}
for(m=0;m<nb_dim;m++){b[m+1][m+1]=1;}
 gaussj(a,nb_dim,b,nb_dim);//inversion
//multiplication
for (i=0;i<dim;i++)
  {
    for (m=0;m<nb_dim;m++)
      {	 
	arg[m][i]=0;
	for (l=0;l<nb_dim;l++){arg[m][i]=Y[l][i]*a[m+1][l+1]+arg[m][i];}  	  
      } 
  }
for (i=0;i<dim;i++) {for (m=0;m<nb_dim;m++){Y[m][i]=arg[m][i];}}
for (m=0;m<nb_dim;++m){delete [] arg[m];}
delete [] arg;
for (m=0;m<nb_dim+1;++m){delete [] a[m];delete [] b[m];}
delete [] a;delete [] b;
}
/*********************************************************************************************************/
/*!@brief Transformation of the distribution in a plan into a Gaussian bivariate distribution.
*
*the analysis uses Numerical recepies function
@param Y input bidimensionnal data  to modify
@param dim number of data samples */
void destructuration(float **Y,unsigned long int dim)
{
unsigned long int i;
long int *aux=new long int[1];
aux[0]=-(int)(((float)rand())/RAND_MAX*10000);
for (i=0;i<dim;i++)
  {
   Y[0][i]=gasdev(aux);Y[1][i]=gasdev(aux);
  }
}
/*********************************************************************************************************/
/*!@brief Sum of an array*/
double Total(float *a, unsigned long int lenght)

{
unsigned long int i;
double res=0;
 for (i=0;i<lenght;i++){ res= res+(double)a[i];}
return res;
}

/*********************************************************************************************************/
