/*---------------------------------------------------------------------------*/
/*!
 * @file Inliers.hh
 * @brief File defining an object for managing an evolving collection of clustered data samples\n
 FOR ANY EXTENSION, THIS OBJECT REQUIRES TO BE RE-IMPLEMENTED 
 *
 */
 /*---------------------------------------------------------------------------*/
#ifndef INLIERS_H
#define INLIERS_H

#include "./ReferenceVector.hh"

/*!@brief object for managing an evolving  collection of clustered data samples\n
 FOR ANY EXTENSION, THIS  OBJECT REQUIRES TO BE RE-IMPLEMENTED 
 *
 *This object possesses the adaptation method, the selection method, and also a bunch of other private methods which are used by these 2 main methods or by the constructor. Other public methods are used  directly fro initialisation or other purposes. */
class Inliers
{
	public:
	
	// Attributes
		/*! @brief number of samples*/
		unsigned long int nb_inliers;
		/*! @brief number of components in the mixture*/
		int nb_RefVector;
		/*! @brief variable=1 if convergence of the EM algorithme achieved, 0 otherwise*/
		int Change;
		/*! @brief Mean number of iteration before convergence (survey variable, could be removed!) */
		int nbMeanIt;
		/*! @brief Mean number of adaptation steps periteration (survey variable, could be removed!) */
		int nbMeanAdapt;		
		
		/*!  @brief the Gaussian mean vectors after the simulated removal of each Gaussian component (should be part of RefVector.hh!!!)*/
		float ***center_RefVector_PostRemoval;
		/*! @brief Index of the samples  NOT USEFUL because there are no more outliers!!*/
		unsigned long int *Index_Inliers;
		/*! @brief Pointer to the multiband image data*/
		Image<PIXEL>* im_point;
		
		//assignement
		/*!  @brief First  Gaussian component index assigned to the samples*/
		int * firstNearestA;
		/*! @brief Second Gaussian component index assigned to the samples*/
		int * secondNearestA;
		/*! @brief mean of the Gaussian components*/
		//Ref_Vector *A;
        vector<Ref_Vector> A;
		/*! @brief covariance for each Gaussian(Gaussian components are correlated)*/
		float ***covariance_cluster;

		
	//Operations:
 	 	/*! @brief constructor initialise the Gaussian mixture
		*
		*After a memory  allocation step, components of the mixture are randomly spread in the feature space according to the mean, and variance of the data. The data points are grouped in clusters according to their Euclidian distance to the Gaussian component means (this is the first step of a K-Means). Then, an adaptation step is performed. This procedure is iterated until there are enough Gaussian components in the mixture.
		@param  data pointer to the multiband image data
		 @param O pointer to the outliers (TO BE REMOVED)
		 @param nb minimum number of requiered Gaussian in the mixture*/
        Inliers();
		Inliers(Image<PIXEL>* data,Outliers * O,int nb);
		/*!@brief EM algorthm for ML estimate of the mixture model 
		*
		* For the assumption of non-interfering Gaussian in the mixture, the EM algorithm is reduced to the iteration  of following to steps until convergence 
		* \li a step for calculating the total assignment of the points to the Gaussians according to a criterion of maximum probability.
		*\li  Reevaluating the Gaussian mean and covariance parameters according to the new points memberships. */
        /*!@brief Initialization of the Gaussian components
         *
         *The components of the mixture are randomly spread in the feature space according to the mean, and variance of the data. A few input paramters are fixed as well. The bounds of the definition domain of the variables are finally calculated.*/ 
        void Initialise(Image<PIXEL> * data_point,Outliers * O,int nb);
		void Adaptation();
		/*!@brief 1 iteration of the maximum decrease optimisation
		*
		*This method calls a private method named "Update_RefVectors()" which removes the Gaussian component inducing the maximum decrease of the 2 part MDL code.*/
		void Selection();
		/*!@brief unuseful method to detect outliers to be removed (not working anymore)*/
		void Outliers_detection();
		/*!@brief Method to add Gaussian components during initialisation  */
		void AddRandomRefVectors();
		/*!@brief Method to print on the standard output, the estimated  Gaussian model parameters  */
		void Print_Mean_Covariance(int k);
		/*!@brief Method to print on the standard output, the estimated  Gaussian model parameters, after the simulated removal of  each Gaussian Component (survey method not used anymore)  */
  		void Print_Mean_Covariance_PostRemoval(int k);
		/*!@brief Method to assign to the point samples the most probable Gaussian Component (Should be private!!) */
		void NearestRefVector_StdDevVar();		

	private:
	
	// Attributes:
	
		/*!@brief pointer to Outliers (Not used anymore!!)*/ 
		Outliers *O_point;
		
		/*!@brief maximum values of the data on each band */
		float *maxi_dim;
		/*!@brief minimum values of the data on each band */
		float *mini_dim;
		
		/*!@brief Inverted covariance matrices of the clusters*/
		float ***Inv_covariance_cluster;
		/*!@brief  Covariance matrices of the clusters after the simulated removal of each Gaussian component*/
		float ****covariance_cluster_PostRemoval;
		/*!@brief  Inverted Covariance matrices of the clusters after the simulated removal of each Gaussian component*/
		float ****Inv_covariance_cluster_PostRemoval;
		/*!@brief determinants of the inverted covariance matrices*/
		float *determinant; 
		/*!@brief determinants of the inverted covariance matrices, after the simulated removal of each Gaussian component*/
		float **determinant_PostRemoval;
		/*!@brief  Singularity flag equal "1  if a band of teh Gaussian is considered singular or "0" if not.*/ 
		int *singularity; 
		/*!@brief After the simulated removal of each Gaussian component, singularity flag equal "1  if a band of teh Gaussian is considered singular or "0" if not*/ 
		int **Singularity_PostRemoval;
		/*!@brief   population of the Gaussian components */
		unsigned long int *effectif_cluster;
		/*!@brief   population of each of the Gaussian components, after the simulated removal of each one of them*/
		unsigned long int **cross_effectif_cluster;
		/*!@brief Criterion evaluating the convergence of the EM algorithm */ 
		float ChangingCriterion;
		/*!@brief Constant */ 
		float M_LOG2;
		/*!@brief Minimum number of Gaussian requiered initially in the mixture */ 
		int nb_init;	
	
	// Operations:
		/*!@brief Initialization of the index of the samples. NOT USEFUL because there are no more outliers*/
		void Initialise_IndexInliers();
		/*!@brief Memory allocation to store the mixture model parameters*/ 
		void Allocate_Variable_StdDev();
		/*!@brief Calculation of the bounds of the definition domain of the variables.*/ 
		void MinMax();
		/*!@brief Grouping of the data samples in clusters according to an Euclidian distance (similar to the first step of a K-Means)*/ 
		void Assigne_data_Euclidian();
		/*!@brief Calculation of the Euclidian distance between the i th data sample and the mean of the  k th component of the mixture*/ 
		float Proba_Assigne_Euclidian(unsigned long int i, int k);
		/*!@brief Removal of the k th component of the mixture.
		@param k index of the component of the mixture to remove
		@param AllCov parameter equal "1" if all the Gaussian attributes (mean vector, covariance and inverted covariance matrices, determinants, singularity flag, Gaussian component assignement index after removal) or equal "0" if only themean should be removed (when, at the beginning, assignements were only performed using an Euclidian distance).*/
		void Remove_RefVector(int k, int AllCov);
		/*!@brief Removal of all the component of the mixture which do not contain any data points
		@param AllCov parameter equal "1" if all the Gaussian attributes (mean vector, covariance and inverted covariance matrices, determinants, singularity flag, Gaussian component assignement index after removal) or equal "0" if only themean should be removed (when, at the beginning, assignements were only performed using an Euclidian distance)*/
		void Remove_Unaffected_RefVector(int AllCov);
		/*!@brief Assignement of the data points to the most probable component*/ 
		void Reassigne_inliers_StdDevVariable();
		/*!@brief Calculation of the probability of the i th data point when belonging to the k th Gaussian component of the mixture.
		*
		*When a band of the multivariate Gaussian is considered singular (singular components possess a Dirac distribution), we assume, for simplification reasons, independance between the bands and decompose the probability into a product.
		*@param i index of the image data sample (pixel index)
		*@param k index of the most likely Gaussian component assignement
		*@return probability of the point according to the most likely Gaussian component */ 
		float Proba_Assigne_Gauss( unsigned long int i, int k);
		/*!@brief Calculation of the probability of the i th data point when belonging to the k th Gaussian component of the mixture.
		*
		*When a band of the multivariate Gaussian is considered singular (singular components possess a Dirac distribution), we assume, for simplification reasons, independance between the bands and decompose the probability into a product.
		*@param i index of the image data sample (pixel index)
		*@param k index of the most likely Gaussian component assignement after the simulated removal of the l th component 
		*@param l index of the removed Gaussian component
		*@return probability of the point according to the most likely Gaussian component after the simulated removal of the l th component */ 
		float Proba_Assigne_Gauss_PostRemoval( unsigned long int i, int k, int l);
		/*!@brief Calculates the weights (population of the class in the special case where Gaussian do not interfer) of the different Gaussian components*/
		void Update_effectif();
		/*!@brief Reevaluate the centers of the Gaussians*/
		void Update_Centers();
		/*!@brief Estimation of the covariance matrices of the Gaussian components together with their inverse matrices  and the inverse matrix determinants. Singularities are also detected and "flagged" by this method.*/
		void Estimation_Stdev_RefVector();
		/*!@brief Estimation of the covariance matrices of the Gaussian components*/
		void Estimation_CovarianceMatrix();
		/*!@brief Detection of the zero values inthe covariance matrices and flag assignements */
		void Singularity_Detection();
		/*!@brief Inversion of the covariance matrices and calculation of the inverted matrice determinants.
		*
		* The inversion is based on a LU decomposition performed using a function of the Numerical recepies free library */
		void Invert_CovarianceMatrix();
		/*!@brief Estimation of the mean vectors, the covariance matrices of the Gaussian components together with their inverse matrices  and the inverse matrix determinants, after the simulated removal of each Gaussian component. Singularities, after the simulated removal of each Gaussian component, are also detected and "flagged" by this method.*/ 
		void Estimation_Stdev_RefVector_PostRemoval();
		/*!@brief Estimation of the mean vectors of the Gaussian components after the simulated removal of each Gaussian component*/
		void Estimation_Center_RefVector_PostRemoval();
		/*!@brief  Calculates the weights (population of the class in the special case where Gaussian do not interfer) of the different Gaussian components after the simulated removal of each Gaussian component*/
		void Estimation_CrossEffectif();
		/*!@brief  Estimation of the covariance matrices of the Gaussian components after the simulated removal of each Gaussian component*/
		void Estimation_CovarianceMatrix_PostRemoval();
		/*!@brief  Detection of the zero values inthe covariance matrices and flag assignements  after the simulated removal of each Gaussian component*/
		void Singularity_Detection_Postremoval();
		/*!@brief Inversion of the covariance matrices and calculation of the inverted matrice determinants after the simulated removal of each Gaussian component.
		*
		* The inversion is based on a LU decomposition performed using a function of the Numerical recepies free library */
		void Invert_PostCovarianceMatrix();
		/*!@brief Method to assign to the point samples the most probable Gaussian Component after the simulated removal of each Gaussian component. */
		void NearestRefVector_StdDevVar_PostRemoval();
		/*!@brief Calculation of the length variation for encoding the data points assignement to the Gaussians.
		*
		* This code length calculus is based on the Shannon-Fano coding source theorem (See theoritical note).
		@param aux array of the Variation length in the assignement encoding,  induced by the simulated removal of each Gaussian component */
		void Inliers_Variation_Index_term_cost2(float* aux);
		/*!@brief Calculation of the length variation for encoding the data points assignement to the Gaussians after the simulated removal of each Gaussian component.
		*
		* This code length calculus is based on the Shannon-Fano coding source theorem (See theoritical note).
		@param SumErrorRatio array of the Variation length of the likelihood, induced by the simulated removal of each Gaussian component*/
		void Inliers_Variation_Error_term_cost4(float * SumErrorRatio);
		/*!@brief Method calculating the likelihood code length variation for the special case where th data points belong to a component with singular bands.
		*
		*The likelihood varaiations for the bands without a singularity (before and after the simulated removal of Gaussians) are calculated separately, using the full covariance matrices. For the bands with a singularity (except case "singularity->singularity" : zero cost exclusion), we calculate separately the likelihood variations. We then assume independance between these bands and the bands without singularities, so that a simple product (sum for the log-likelihood) gives us the complete likelihood varaiation.
		@param i index of the image data sample (pixel index)
		@return likelihood variation for the i th data point*/
		float Error_FreeInliers_Singularity(unsigned long int i);
		/*!@brief Method calculating the likelihood code length variation for the i th inliers which was belonging to the simulated removed Gaussian*/
		float Error_FreeInliers(unsigned long int i);
		/*!@brief Method calculating the likelihood code length variation for the inliers which are not belonging to simulated removed Gaussians
		@param SumErrorRatio array of the Variation length of the likelihood of the points not belonging to the simulated removed Gaussians, induced by the simulated removal of each Gaussian component. */
		void Error_OtherInliers(float * SumErrorRatio);
		/*!@brief removes the Gaussian component inducing the maximum decrease of the 2 part MDL code.
		*
		*The 2 part code length variation is composed of 
		\li the variation of code lenght induced by saving the encoding of a Gaussian component (see theoritical note approximation). However, THIS IS A VERY ROUGH APPROXIMATION AND SHOULD BE REPLACED EASILY BY THE TERM DEFINED IN THE THEORITICAL NOTE.
		\li neglected  variation terms (see theoritical note approximation)
		\li the variation in the length of encoding  of the data points assignement to the Gaussians. This variation is calculated by the following method :" Inliers_Variation_Index_term_cost2(float* aux)".
		\li the variation of the encoding of the data points according to the so defined Gaussian model calculated by the following method "Inliers_Variation_Error_term_cost4(float * SumErrorRatio)". This is the varaiation of the likelihood parameter.*/
		void Update_RefVectors();

		
};
#endif
