/*---------------------------------------------------------------------------*/
/*!
 * @file TimeGaussMIX_MDL.hh
 * @brief   File containing the Kernel Object of the program, supporting the algorithm organisation in 4 steps : Initialization, Adaptation, Selection and Convergence*/
 /*---------------------------------------------------------------------------*/

#include <iostream>
#include <string>  
#include <vector>
#include <fstream>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>

#include "./Inliers.hh"

using namespace std;
/*!@brief Kernel Object of the program, supporting the algorithm organisation in 4 steps : Initialization, Adaptation, Selection and Convergence.*/
class Temporal_MDL_classifier 
{	
	public:
		
		Temporal_MDL_classifier(){};
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief initialization of a first Gaussian mixture*/
		 /*! Initialization :  A initial Gaussian mixture is produced. It is composed of a high number of Gaussian (number provided *by the user). The initialization is done by randomly spreading the clusters according to a Gaussian distribution of mean and variance learned from each data feature component.*/
		void Initialise(char * path_in, char * path_fileout, int dimx_tmp, int dimy_tmp, int nb_bands, int ofx_tmp, int ofy_tmp, int Dy_tmp, int subx_tmp, int suby_tmp,int nb) 
	
		 {
			Acquire_path(path_in,nb_bands,path_fileout);
			Acquire_parameters_allocateImage(dimx_tmp,dimy_tmp,ofx_tmp,ofy_tmp,Dy_tmp,subx_tmp,suby_tmp);
			Curent_images_index=0;
			Acquire_next_image();
			Begin_procedure(nb);
			Allocate_classfile();
			Allocate_OutMeanCov();
			
			
		 }
		/*--------------------------------------------------------------------------------------------*/
		 /*!@brief chained optimisation procedures for consecutive modeling */
		 /*!the optimisation procedure comprises 4 steps :  
		 *\li 1- If the initialisation has not been performed yet (all except first multi-band image to model), the modeling parameters infered at the last modeling itration are used 
		 *\li 2- Adaptation : At a given iteration, we consider K Gaussians in the mixture. An Expectation Maximization (EM) algorithm is used to perform a ML estimation of the Gaussian mixture parameters.
		 *\li 3- Selection : For the same iteration, we remove the Gaussian which induces the biggest decrease in the description length. Then we go back to step 2 if a decrease is observed. If not, then we do not remove any Gaussians and go to the convergence step. 
		 *\li 4- Convergence : To avoid local minima, we iterate step 2 and 3 until the variation in the description lenght does not exceed a given treshold. When stability is reached, we obtain the estimated number of Gaussians with the ML estimates of the parameters of the mixture model */
		 void Cluster_optimiser()
		 {
		 	int nb_outliers_temp,nb_RefVector_temp;
			int compt;
			time_t tdeb, t1,t2,t3;
			while (Curent_images_index<nb_images){
				if (Curent_images_index!=0)
				{
					Acquire_next_image();
					I[0].AddRandomRefVectors();
				}
				
time(&tdeb);
t1=tdeb;
				I[0].Adaptation();
time(&tdeb);
t2=tdeb;
				I[0].Selection();
time(&tdeb);
t3=tdeb;
cout<<"Adaptation time : "<<t2-t1<<" sec / Selection time : "<<t3-t2<<" sec"<<endl;
cout<<"Mean adaptation iteration : "<<(float)I[0].nbMeanIt/(float)I[0].nbMeanAdapt<<endl;

				I[0].Outliers_detection();
				nb_outliers_temp=0;
				nb_RefVector_temp=0;
				compt=0;
				while(nb_outliers_temp!=O->nb_outliers || nb_RefVector_temp!=(&I[0])->nb_RefVector || (&I[0])->Change) 
				{ 
					if (compt > 25 && nb_outliers_temp==O->nb_outliers && nb_RefVector_temp==(&I[0])->nb_RefVector)
					{
						break;
					}
					nb_outliers_temp=O->nb_outliers;
					nb_RefVector_temp=(&I[0])->nb_RefVector;
					compt++;
					//cout<< ""<<endl;
					//cout<< "Selection step number "<<compt<< ":		 number of RefVector = "<<nb_RefVector_temp<<", number of Outliers = "<<nb_outliers_temp<<endl;
					I[0].Adaptation();
					I[0].Selection();	
			
						
				}
					cout<< "Number of Selection steps "<<compt<< ",   number of RefVector = "<<nb_RefVector_temp<<", number of Outliers = "<<nb_outliers_temp<<endl;
time(&tdeb);
t3=tdeb;
cout<<"Optimization time duration : "<<t3-t1<<" sec"<<endl;

			I[0].Print_Mean_Covariance(-1);
			Create_classfile();
			Create_OutMeanCov();
			Curent_images_index++;	

			}
		 }
		 
	private:
		 /*!@brief number of multiband images to classify*/
		int nb_images;
		/*!@brief image attributes*/
		int dimx; int dimy; int ofx; int ofy; int Dy; int subx; int suby;
		/*!@brief array of paths for the output files*/
		char *** PathOutPut;
		/*!@brief array of paths for the input files*/
		char ***PathInPut;
		/*!@brief Index of the Image time sample which is currently processed*/
		int Curent_images_index;
		/*!@brief Paths of the Image which is currently processed*/
		char ***Current_Path;
		/*!@brief Multiband image data*/
		Image<PIXEL> data;
		/*!@brief Classfile*/
		Image<unsigned char> classfile;
		/*!@brief attribute file*/
		Image<float> OutMeanCov;
		/*!@brief Inliers object used for managing the feature points  grouped into clusters  */
		//Inliers *I;
        vector<Inliers> I;
		/*!@brief Outliers object not used anymore (TO BE REMOVED!)  */
		Outliers *O;
		/*--------------------------------------------------------------------------------------------*/
		 /*!@brief  Allocation of memory for storing the data */
		void Acquire_parameters_allocateImage(int dimx_tmp, int dimy_tmp, int ofx_tmp, int ofy_tmp, int Dy_tmp, int subx_tmp, int suby_tmp)
		{
		dimx=dimx_tmp; dimy=dimy_tmp; ofx=ofx_tmp;  ofy=ofy_tmp; Dy=Dy_tmp; subx=subx_tmp; suby=suby_tmp;
		data.Allocate((int)(dimx/subx),(int)(dimy/suby),data.nb_bands,1);
					
		}
		/*--------------------------------------------------------------------------------------------*/
		 /*!@brief Reading the next image */
		void Acquire_next_image()
		{
			int m;
			for (m=0;m<data.nb_bands;m++){Current_Path[0][m]=PathInPut[Curent_images_index][m];}
			data.Read_Cut_SubSample(Current_Path,ofx,ofy,Dy,subx,suby);
			
		}
 		/*--------------------------------------------------------------------------------------------*/
		/*!@brief Reading the text input file names and counting the number of multi-band images to classify */
		void Acquire_path(char * path_in, int nb_bands, char* path_fileout)  
		{
			
			int m,k;
			Image<unsigned char> pippo;
			nb_images=pippo.NumberImages_pathfiles_in(path_in,nb_bands);
			
			PathInPut= new char **[nb_images];
			for(m=0;m<nb_images;++m){PathInPut[m] =new char*[nb_bands];} 
			for(m=0;m<nb_images;++m){for(k=0;k<nb_bands;++k){PathInPut[m][k]=new char[128];}}							
			pippo.Read_pathfiles_in(path_in,PathInPut,nb_bands);

			Current_Path= new char **[1];
			for(m=0;m<1;++m){Current_Path[m] =new char*[nb_bands];} 
			for(m=0;m<1;++m){for(k=0;k<nb_bands;++k){Current_Path[m][k]=new char[128];}}				
			
			data.nb_images=1;
			data.nb_bands=nb_bands;
			//dataSmoothBin.nb_images=1;
			//dataSmoothBin.nb_bands=nb_bands;
			
			Acquire_Pathoutput(path_fileout);
			
		}
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief  Reading the text file names for saving the classifiles*/
		void Acquire_Pathoutput(char *path_fileout)
		{
		
		int m;
			
			PathOutPut= new char **[nb_images];
			for(m=0;m<nb_images;++m){PathOutPut[m] =new char*[1];PathOutPut[m][0]=new char[128];	} 
			
			Image<unsigned char> pippo;
			pippo.nb_bands=1;
			pippo.nb_images=nb_images;
			pippo.Read_pathfiles_out(path_fileout,PathOutPut);
		
		}
		/*--------------------------------------------------------------------------------------------*/
		/*! @brief Allocating memory to store  the Gaussian Mixture parameters  */
		void Allocate_OutMeanCov()
		{
			OutMeanCov.Allocate(255,(data.nb_bands+1)*data.nb_bands,1,1);	
		}
		/*--------------------------------------------------------------------------------------------*/
		/*! @brief Storing the Gaussian Mixture parameters */
		void Make_OutMeanCov()
		{
			int k,j,j1,j2;
			//storing on the 1st line the nb of classes followed by zeros...
			*(OutMeanCov.position)=I[0].A[0].nb_ref;
			//storing on each line :
			for (k=0;k<I[0].A[0].nb_ref;k++)
			{
				for (j=0;j<data.nb_bands;j++)
				{
					//the cluster centroids
					*(OutMeanCov.position+(k+1)*OutMeanCov.ny+j)=I[0].A[0][k][j];//

				}
				for (j1=0;j1<data.nb_bands;j1++)
				{
					for (j2=0;j2<data.nb_bands;j2++)
					{
						//the cluster covariance matrix
							*(OutMeanCov.position+(k+1)*OutMeanCov.ny+(j1+1)*data.nb_bands+j2)=I[0].covariance_cluster[k][j1][j2];
					}
				}
			}
		}	
		/*--------------------------------------------------------------------------------------------*/	
		/*! @brief Writing the  infered Gaussian parameters on the disk*/
		void Write_OutMeanCov()
			{
			char *path=new char[128];
			strcpy(path,"./ClusterAttributes_00");
			path[20]=char((int)(Curent_images_index+1)/10+48);
			path[21]=char((Curent_images_index+1)%10+48);
			OutMeanCov.Write_singleImage(path);
			
			cout<<" Cluster Attributes of image "<< Curent_images_index+1 << " written"<<endl;
			delete [] path;
			}
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief  Storing and writing the Gaussian Mixture parameters */
		 void Create_OutMeanCov()
		 {
			Make_OutMeanCov();
			//Write_OutMeanCov();
		 }
		
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief  Allocate memory  for classfiles*/
		void Allocate_classfile()
		{
			classfile.Allocate(data.nx,data.ny,1,1);	
		}
		/*--------------------------------------------------------------------------------------------*/
		/*! @brief Storing classfiles*/
 		void Make_classfile()
			{
			unsigned long int i;
			I[0].im_point=&data;
			I[0].NearestRefVector_StdDevVar();
			for (i=0;i<(unsigned long int)data.nx*(unsigned long int)data.ny;i++)
				{
				*(classfile.position+i)=255;
				
				}
			for (i=0;i<I[0].nb_inliers;i++)
				{
				*(classfile.position+(I[0].Index_Inliers[i]))=(unsigned char)I[0].firstNearestA[i];
				
				}
			
			
			}	
		/*--------------------------------------------------------------------------------------------*/
		/*! @brief Writing classfiles on the disk*/	
		void Write_classfile()
			{
			char ***path=new char **[1];
			path[0] =new char*[1];
			path[0][0]=new char[128];
			path[0][0]=PathOutPut[Curent_images_index][0];
			classfile.Write(path);
			cout<<" Classfile of image "<< Curent_images_index+1 << " written"<<endl;
			delete [] path[0][0];
			}
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief  Storing and writing classfiles on the disk*/
		 void Create_classfile()
		 {
			Make_classfile();
			Write_classfile();
		 }
		
		/*--------------------------------------------------------------------------------------------*/
		/*! @brief Objects initialisation.
		 *
		 *Initialisation of the Inliers (and Outliers) and initialisation of the pointer to the image data*/
		void Begin_procedure(int nb)
		{
			Image<PIXEL>* im_point=&data;
			O=new Outliers[1]();
			//I=new Inliers[1](im_point,O,nb);            
			//I  = new Inliers[1];
            Inliers tmp(im_point,O,nb);
            I.push_back(tmp);
            

		}
		/*--------------------------------------------------------------------------------------------*/
		




};
/**********************************************************************************************/
