/*!@file DynaClass.hh
*@brief File containing the DynaClass object used for storing the MT cluster trajectories*/
#include <iostream>
#include <fstream>
#include <string>  
#include <vector>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
/*!@brief  object used for storing one MT cluster trajectory. Almost all the attribute are public and can be accessed from  a main procedure. */
class DynaClass 
{
  public:
  	
  	// Attributes
	/*!@brief Collections of TL cluster assigned to the given MT cluster for each time sample\n
	CAREFULL!! The name is confusing, it would be better to name it  TLAssigned2MT*/
	vector<int> *AssignedTimelessClass;
	/*!@brief Vectors of divergence value associated to the assigned TL cluster for each time sample*/
	vector<float>  *DivergenceValue;
	/*!@brief Vectors of mean values associated to the assigned TL cluster for each time sample*/
	vector<float*> *Centroids;
	/*!@brief Matrices of covariance values associated to the assigned TL cluster for each time sample*/
	vector<float**> *Covariances;
	/*!@brief number of sample points belonging to the assigned TL cluster and the MT cluster for each time sample*/
	vector<unsigned long int> *Weight;
	/*!@brief mean vectors of the  MT cluster for each time sample*/
	vector<float*> *MTCentroids;
	/*!@brief Covariance matrices  of the  MT cluster for each time sample*/
	vector<float**> *MTCovariances;
	/*!@brief number of sample points belonging to the MT class\n
	SHOULD NOT BE A VECTOR BUT A SCALAR! */
	vector<unsigned long int> *MTWeight;
	
	//Joint Distribution for 2 consecutive times;
	/*!@brief mean vectors of the joint distribution of a MT clusters, in 2 consecutive time samples.
	*
	*This attribute is used for mutual information calculation*/
	vector<float*> *JointCentroids;
	/*!@brief Covariance matrix of the joint distribution of a MT clusters, in 2 consecutive time samples.
	*
	*This attribute is used for mutual information calculation*/
	vector<float**> *JointCovariances;
	/*!@brief number of sample points belonging to the MT class\n
	IDENTICAL to MTWeight! SHOULD NOT BE AN ARRAY BUT A SCALAR! 
	*
	*This attribute is used for mutual information calculation*/
	vector<unsigned long int> *JointWeight;
	/*!@brief Array of mutual information values calculated  between a  MT cluster in consecutive time samples */
	float *MutualInfo;
	
	//trajectory attributes between 2 consecutive times
	/*!@brief  graph edge attributes between 2 consecutive times : flow of points between consecutive TL clusters*/
	float ***flow;
	/*!@brief  graph edge attributes between 2 consecutive times : divergence between consecutive TL clusters*/
	float ***div;
	
	// Operations
    	DynaClass(){}
	virtual~DynaClass(){}
	/*!@brief Method to allocate memory for storing a MT trajectory.
	*
	* The object attributes are almost all vector objects. Thus the memory allocation must be done progressively. This method allocates memory only for the first TL association. Further allocation procedures are performed later.
	@param size number of time samples
	@param nb_bands dimensionnality of the TL feature spaces*/
	void Allocate(int size, int nb_bands)
	{
	
		int k,j;
		nb_TemporalSample=size;
		nb_dim=nb_bands;
		AssignedTimelessClass=new vector<int>[nb_TemporalSample];
		Weight=new vector<unsigned long int>[nb_TemporalSample];
		DivergenceValue=new vector<float>[nb_TemporalSample];
		Centroids=new vector<float*>[nb_TemporalSample];
		Covariances=new vector<float**>[nb_TemporalSample];
		
		MTWeight=new vector<unsigned long int>[nb_TemporalSample];
		MTCentroids=new vector<float*>[nb_TemporalSample];
		MTCovariances=new vector<float**>[nb_TemporalSample];
		
		JointCentroids=new vector<float*>[nb_TemporalSample-1];
		JointCovariances=new vector<float**>[nb_TemporalSample-1];
		JointWeight=new vector<unsigned long int>[nb_TemporalSample-1];
		
		int *AuxAssignedTimelessClass=new int[nb_TemporalSample];
		unsigned long int *AuxWeight=new unsigned long int[nb_TemporalSample];
		float  *AuxDivergenceValue=new float[nb_TemporalSample];
		float  **AuxCentroids=new float*[nb_TemporalSample];
		float ***AuxCovariances=new float**[nb_TemporalSample];
		unsigned long int *AuxMTWeight=new unsigned long int[nb_TemporalSample];
		float  **AuxMTCentroids=new float*[nb_TemporalSample];
		float ***AuxMTCovariances=new float**[nb_TemporalSample];
		float  **AuxJointCentroids=new float*[nb_TemporalSample-1];
		float ***AuxJointCovariances=new float**[nb_TemporalSample-1];
		unsigned long int  *AuxJointWeight=new unsigned long int[nb_TemporalSample-1];
	
	
		for (k=0;k<nb_TemporalSample;k++)
		{
			AuxCentroids[k]=new float[nb_dim];
			AuxCovariances[k]=new float*[ nb_dim];
			AuxMTCentroids[k]=new float[nb_dim];
			AuxMTCovariances[k]=new float*[ nb_dim];
			for (j=0;j<nb_dim;j++)
			{
				AuxCovariances[k][j]=new float[nb_dim];
				AuxMTCovariances[k][j]=new float[nb_dim];
			}
			if (k<nb_TemporalSample-1){
				AuxJointCentroids[k]=new float[nb_dim*2];
				AuxJointCovariances[k]=new float*[nb_dim*2];
				for (j=0;j<nb_dim*2;j++){
					AuxJointCovariances[k][j]=new float[nb_dim*2];
				}
				JointCentroids[k].push_back(AuxJointCentroids[k]);
				JointCovariances[k].push_back(AuxJointCovariances[k]);
				JointWeight[k].push_back(AuxJointWeight[k]);
			}
			Weight[k].push_back(AuxWeight[k]);
			AssignedTimelessClass[k].push_back(AuxAssignedTimelessClass[k]);
			DivergenceValue[k].push_back(AuxDivergenceValue[k]);
			Centroids[k].push_back(AuxCentroids[k]);
			Covariances[k].push_back(AuxCovariances[k]);
			MTWeight[k].push_back(AuxMTWeight[k]);
			MTCentroids[k].push_back(AuxMTCentroids[k]);
			MTCovariances[k].push_back(AuxMTCovariances[k]);
			
		}
		for (k=0;k<nb_TemporalSample;k++){AssignedTimelessClass[k][0]=-1;}
		MutualInfo=new float[nb_TemporalSample-1];
		for (k=0;k<nb_TemporalSample-1;k++){MutualInfo[k]=-1;}
	}
	/*!@brief Method to fix the objects private variables.
	@param size number of time samples
	@param nb_bands dimensionnality of the TL feature spaces*/
	void InitRead(int size, int nb_bands){
	
		nb_TemporalSample=size;
		nb_dim=nb_bands;
		AssignedTimelessClass=new vector<int>[nb_TemporalSample];
		Weight=new vector<unsigned long int>[nb_TemporalSample];
		DivergenceValue=new vector<float>[nb_TemporalSample];
		Centroids=new vector<float*>[nb_TemporalSample];
		Covariances=new vector<float**>[nb_TemporalSample];
		
		MTWeight=new vector<unsigned long int>[nb_TemporalSample];
		MTCentroids=new vector<float*>[nb_TemporalSample];
		MTCovariances=new vector<float**>[nb_TemporalSample];
						
		MutualInfo=new float[nb_TemporalSample-1];
		flow= new float**[nb_TemporalSample];
		div= new float**[nb_TemporalSample];
	}
	/*!@brief Method to allocate memory for storing a MT cluster at a given time sample.
	  @param t image time samples*/
	void MTAlloc(int t){
			   
		float  *AuxMTCentroids=new float[nb_dim];
		float **AuxMTCovariances=new float*[nb_dim];
		for (int j=0;j<nb_dim;j++){ AuxMTCovariances[j]=new float[nb_dim];}
		unsigned long int  *AuxMTWeight=new unsigned long int[1];
	
		MTCentroids[t].push_back(AuxMTCentroids);
		MTCovariances[t].push_back(AuxMTCovariances);
		MTWeight[t].push_back(AuxMTWeight[0]);

	
	}
	/*!@brief Method to allocate memory for storing a TL cluster attributes and association attributes with a MT cluster, at a given time sample.
	*
	* allocation of a number "lenght" of associations. These memory spaces are pushed back at the end of the vectors.
	@param t image time samples 
	@param length number of TL cluster associations with a MT cluster*/
	void TLAlloc(int t, int length){
	
	    for (int k=0;k<length;k++){
		
		unsigned long int *AuxWeight=new unsigned long int[1];
		int *AuxAssignedTimelessClass=new int[1];
		float *AuxDivergenceValue=new float[1];	   
		float *AuxCentroids=new float[nb_dim];
		float **AuxCovariances=new float*[ nb_dim];
		for (int j=0;j<nb_dim;j++)
		{
			AuxCovariances[j]=new float[nb_dim];
			AuxCovariances[j]=new float[nb_dim];
		}			
		Weight[t].push_back(AuxWeight[0]);
		AssignedTimelessClass[t].push_back(AuxAssignedTimelessClass[0]);
		DivergenceValue[t].push_back(AuxDivergenceValue[0]);
		Centroids[t].push_back(AuxCentroids);
		Covariances[t].push_back(AuxCovariances);
	    }
	}
	/*!@brief Method to allocate memory for storing a TL cluster association with a MT cluster, at a given time sample (comprised in method "TLAlloc(...)").
	*
	* allocation of a number "lenght" of associations. These memory spaces are pushed back at the end of the vectors.
	@param t image time samples 
	@param length number of TL cluster associations with a MT cluster*/
	void TLAlloc2(int t, int length){
	
	    for (int k=0;k<length;k++){
		
		unsigned long int *AuxWeight=new unsigned long int[1];
		int *AuxAssignedTimelessClass=new int[1];
		float *AuxDivergenceValue=new float[1];	   
		Weight[t].push_back(AuxWeight[0]);
		AssignedTimelessClass[t].push_back(AuxAssignedTimelessClass[0]);
		DivergenceValue[t].push_back(AuxDivergenceValue[0]);
	    }
	}
	/*!@brief Method to allocate memory for storing a TL cluster attributes, at a given time sample (comprised in method "TLAlloc(...)").
	*
	* allocation of a number "lenght" of associations. These memory spaces are pushed back at the end of the vectors.
	@param t image time samples 
	@param length number of TL cluster associations with a MT cluster*/
	void TLAllocCenterCov(int t, int length){
	
	    for (int k=0;k<length;k++){
				   
		float *AuxCentroids=new float[nb_dim];
		float **AuxCovariances=new float*[ nb_dim];
		for (int j=0;j<nb_dim;j++)
		{
			AuxCovariances[j]=new float[nb_dim];
			AuxCovariances[j]=new float[nb_dim];
		}			
		Centroids[t].push_back(AuxCentroids);
		Covariances[t].push_back(AuxCovariances);
	    }
	}
	
	/*!@brief Method to allocate memory for storing a the flows of points between 2 consecutive times.
	*
	* allocation of a number "lenght1*lenght2" of flow values in an array.
	@param t image time samples 
	@param length number of TL cluster associations with a MT cluster at time t 
	@param length number of TL cluster associations with a MT cluster at time t+1 */
	void flowDivAlloc(int t, int length1, int length2){
   
	   	flow[t]=new float*[length1];div[t]=new float*[length1];
	   	for (int k=0;k<length1;k++){flow[t][k]=new float[length2];div[t][k]=new float[length2];}
	}
	
	
  private:

  	/*! @brief number of image time samples*/
	int nb_TemporalSample;
	/*! @brief number of dimension of the feature space*/
	int nb_dim;
};
