/*---------------------------------------------------------------------------*/
/*!
 * @file MI.hh
 * @brief  File defining the object MI for calculation of mutual information beween 3 dimensionnal Gaussians in consecutive times.
 */
/*---------------------------------------------------------------------------*/
#ifndef MI_HH
#define MI_HH
#define covMin 0.01;// related to bin of histogrammes
#define Infini 1e20;
/*! @brief  Object performing the analytical calculus of mutual information beween 3 dimensionnal Gaussians in consecutive times.
*
* The procedure is based on the determinants of covaraint matrices corresponding to the 2 Gaussian distributions (T1 and T2) and to the joint distribution T12*/
class MI {
  
  // Attributes
  private:
    /*! @brief mean vector time 1*/
    vector<float*> MeanT1;
    /*! @brief mean vector time 2*/		
    vector<float*> MeanT2;
    /*! @brief mean vector time 1 and 2*/
    vector<float*> MeanT12;
    /*! @brief covariance matrix time 1*/		
    vector<float**> CovT1;
     /*! @brief covariance matrix time 2*/		
    vector<float**> CovT2;
    /*! @brief covariance matrix time 1 and 2*/
    vector<float**> CovT12; 
    /*! @brief Determinant covariance matrix time 1 */
    vector<float> determinantT1;
     /*! @brief Determinant covariance matrix time 2 */
    vector<float> determinantT2; 
    /*! @brief Determinant covariance matrix time 1 and 2 */
    vector<float> determinantT12;
    
    /*! @brief  dimensionnality*/
    int nb_dim;
        
  // Operations
  public:
    MI();
    virtual ~MI();
    /*!@brief Memory allocation 
    *
    *This method should be simpler. Indeed, the vector objects used here could be replace by a simple scalar.  
    @param nb_dimensions dimensionnality of the TL feature spaces*/
    void Allocate(int nb_dimensions); 
    /*! @brief Initialize variable at T1 and calculates the covariance matrix determinant
    @param C1 covariance matrix of the  MT cluster at time 1
    @param M1 mean vector of the MT  cluster at time 1*/
    void Initialise1(vector<float**> *C1,vector<float*> *M1);
    /*! @brief Initialize variable at T2 and calculates the covariance matrix determinant
     @param C2 covariance matrix of the  MT cluster at time 1
    @param M2 mean vector of the MT  cluster at time 1 */
    void Initialise2(vector<float**> *C2,vector<float*> *M2);
    /*! @brief Initialize variable at T12 and calculates the covariance matrix determinant
    @param C12 covariance matrix of the  MT cluster projected at time 1 and 2
    @param M12 mean vector of the MT  cluster projected at time 1 and 2*/
    void Initialise12(vector<float**> *C12,vector<float*> *M12);
    /*! @brief Mutual information calculation 
    *
    *This function computes a simple ration of determinants for the mutual information calculation. Note that if there are singular covariance matrices, the mutual information is still equal to the ratio of determinants, but the singular dimensions must be removed from the marginal and joint covariance matrices.
    @return mutual information value*/
    float MICalculus();
    
  private:
    /*!@brief determinant calculation at T1 
    *
    *method based on a LU decomposition provided by Numerical recepies
    *
    *Note that when the matrix is singular the determinants is calculated with a covariance matrix where the singular dimensions have been removed.*/
    void DetCovMatrixT1();
    /*! @brief determinant calculation at T2 
    *
    **method based on a LU decomposition provided by Numerical recepies
    *
    *Note that when the matrix is singular the determinants is calculated with a covariance matrix where the singular dimensions have been removed.*/
    void DetCovMatrixT2();
    /*! @brief determinant calculation at T12 
    *
    **method based on a LU decomposition provided by Numerical recepies
    *
    *Note that when the matrix is singular the determinants is calculated with a covariance matrix where the singular dimensions have been removed.*/
    void DetCovMatrixT12();    
    
};

#endif
