/*---------------------------------------------------------------------------*/
/*!
 * @file SuperposeClass.hh
 * @brief  File Containing the SuperposeClass object which performs basic operations with MT and TL spatial classes	  */
/*---------------------------------------------------------------------------*/
#include <iostream>
#include <fstream>
#include <string>  
#include <vector>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>

/*******************************************************************************************************/
/*************GENRIC MERGING-SPLITING CLASS*************************************************************/
/*******************************************************************************************************/

/*! @brief object which performs basic operations with MT and TL spatial classes.
 *
 * The spatial constraints introduced in the probabilities of association of MT with TL clusters are calculated with the object methods : The number of pixels of the intersection of a TL and a MT spatial map, the flow of pixels between consecutive TL clusters, or the number of pixels belonging to a MT class.*/
class SuperposeClass
{
    public:
  	
	// Operations
	
	SuperposeClass(){}
    	virtual ~SuperposeClass(){}
	/*******************************************************************************************************/
	/*! @brief Method for initializing the object, particularly by defining the pointer to the image classfiles */
	void AcquirePara(int t,int nb_dimension,int nb_im,int nb_Class, Image<unsigned char> *classfiles, Image<unsigned char> *Timeless_classfile)
	{
		TimeLocation=t;
		nb_dim=nb_dimension;
		nb_images=nb_im;
		nb_TimelessClass=nb_Class;
		TempClassfiles=classfiles;
		TimelessClassfile=Timeless_classfile;
		
	}
/*******************************************************************************************************/
	/*! @brief Method to quantify the number of pixel belonging to intersections of a TL and a  MT spatial class */
	float DivWeight(int classIndex,int Index_TimelessClass)
	{
		float ClassPopulation;
		ClassPopulation=(float)SuperposedPopulation(classIndex,TimeLocation,Index_TimelessClass);	
		return ClassPopulation;
	}
	/*******************************************************************************************************/
	/*! @brief Method to quantify the flow of point between two TL classes consecutivein time */
	float FlowWeight(int classIndex,int preClassIndex,int Index_TimelessClass)
	{
		float ClassPopulation;
		unsigned long int res=0;
		int j1,j2;
		int aux;
		int aux1;
		int aux2;
		unsigned long int nxny=TempClassfiles->nx*TempClassfiles->ny;
		int sideX;
		int timeXNxny=TimeLocation*nxny;
		int preTimeXNxny=(TimeLocation-1)*nxny;
		for (j1=0;j1<TempClassfiles->nx;j1++)
		{
		   sideX=j1*TempClassfiles->ny;
		   for (j2=0;j2<TempClassfiles->ny;j2++)
		   {
		      aux2=*((*TimelessClassfile).position+sideX+j2);
		      if (aux2==(Index_TimelessClass)){
		        aux=*((*TempClassfiles).position+timeXNxny+sideX+j2);
		      	aux1=*((*TempClassfiles).position+preTimeXNxny+sideX+j2);
		      	if ((aux==classIndex)&&(aux1==preClassIndex)){res++;}
		      }
		   }
		}
		ClassPopulation=(float)res;	
		return ClassPopulation;
	}
	
/*******************************************************************************************************/
	/*!@brief  Method to quantify the weight of a given MT class */
	float MTPopulation(int kMT){
		float res=0;
		int aux;
		for (int j1=0;j1<TimelessClassfile->nx;j1++)
		{
		   for (int j2=0;j2<TimelessClassfile->ny;j2++)
		   {
		      aux=*((*TimelessClassfile).position+j1*TimelessClassfile->ny+j2);
		      if (aux==kMT){res++;}
		   }
		}
		return res;
	}	/*******************************************************************************************************/
   private:
   	
	// Attributes
	/*! @brief Current temporal sample*/
   	int TimeLocation;
	/*! @brief Dimensionnality of the feature space (3)*/
	int nb_dim;
	/*! @brief number of image time samples*/
	int nb_images;
	/*! @brief number of MT clusters*/
	int nb_TimelessClass;
	/*! @brief TL classfiles*/
	Image<unsigned char>* TempClassfiles;
	/*! @brief MT classfiles*/
	Image<unsigned char> * TimelessClassfile;
		
	//Operations
	/*******************************************************************************************************/
	/*!@brief Number of samples attached to a given TL class*/
	unsigned long int Population(int class_index, int time)
	{
		unsigned long int res=0;
		int j1,j2;
		int aux;
		unsigned long int nxny=TempClassfiles->nx*TempClassfiles->ny;
		for (j1=0;j1<TempClassfiles->nx;j1++)
		{
		   for (j2=0;j2<TempClassfiles->ny;j2++)
		   {
		      aux=*((*TempClassfiles).position+time*nxny+j1*TempClassfiles->ny+j2);
		      if (aux==class_index){res++;}
		   }
		}
		return res;
	}
	/*******************************************************************************************************/
	/*!@brief Number of samples belonging to the spatial intersection of a MT and an  a given TL class
	@param class_index TL class index
	@param time image time sample
	@Index_TimelessClass MT class index
	@return number of pixels in the intersection of the classes */
	unsigned long int SuperposedPopulation(int class_index,int time,int Index_TimelessClass)
	{
		unsigned long int res=0;
		int j1,j2;
		int aux;
		int aux2;
		unsigned long int nxny=TempClassfiles->nx*TempClassfiles->ny;
		for (j1=0;j1<TempClassfiles->nx;j1++)
		{
		   for (j2=0;j2<TempClassfiles->ny;j2++)
		   {
		      aux=*((*TempClassfiles).position+time*nxny+j1*TempClassfiles->ny+j2);
		      aux2=*((*TimelessClassfile).position+j1*(*TimelessClassfile).ny+j2);
		      if ((aux==class_index)&&(aux2==(Index_TimelessClass))){res++;}
		   }
		}
		return res;
	}
	
	/*******************************************************************************************************/
	/*!@brief Number of samples belonging to the spatial intersection of a MT class and to the spatial intersection of two given TL classes
	@param class_index TL class index at time t
	@param time image time sample t
	@preClassIndex TL class index at time t-1
	@param preClassIndex image time sample t-1
	@Index_TimelessClass MT class index
	@return number of pixels in the intersection of the classes */
	unsigned long int FlowPopulation(int class_index,int time,int preClassIndex,int preTimeLocation,int Index_TimelessClass)
	{
		unsigned long int res=0;
		int j1,j2;
		int aux;
		int aux1;
		int aux2;
		unsigned long int nxny=TempClassfiles->nx*TempClassfiles->ny;
		int sideX;
		int timeXNxny=time*nxny;
		int preTimeXNxny=preTimeLocation*nxny;
		for (j1=0;j1<TempClassfiles->nx;j1++)
		{
		   sideX=j1*TempClassfiles->ny;
		   for (j2=0;j2<TempClassfiles->ny;j2++)
		   {
		      aux=*((*TempClassfiles).position+timeXNxny+sideX+j2);
		      aux1=*((*TempClassfiles).position+preTimeXNxny+sideX+j2);
		      aux2=*((*TimelessClassfile).position+sideX+j2);
		      if ((aux==class_index)&&(aux2==(Index_TimelessClass))&&(aux1==preClassIndex)){res++;}
		   }
		}
		return res;
	}
	/*******************************************************************************************************/
	/*!@brief returns the sum of an array*/
	int SUM(int * aux,int size)
	{
		int res=0;
		for (int j=0;j<size;j++)
		{
			res +=aux[j];
		}
		return res;
	}
	/*******************************************************************************************************/

};
