/*---------------------------------------------------------------------------*/
/*!
 * @file graph.cpp
 * @brief Main routine for the inference of a graph of dynamic cluster trajectories
 
 * This progamm creats first a  GraphDynaCluster object used for infering and  storing the dynamic cluster trajectories.\n
  *  Then a GraphDynaCluster method is called for memory allocation and for reading the input data (classfiles and attribute files)\n
  *  The kernel method of the  GraphDynaCluster object, which infers the MT cluster trajectories using Kullback-Leibler divergence and a spatial constraint, is then called.\n
  * Last, GraphDynaCluster methods are called for characterization of the edges and writing the output file : 1) the characterization of the edges with mutual information 2) the characterization of the other edge attributes. This last characterization is done in the last procedure,  while writing the graphe output. 
 
 
  @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/

#include <iostream>
#include <string>  
#include <vector>
#include <fstream>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <ctime>

#include "graphCharac.hh"

using namespace std;
 

/*! 	@brief  Usage function 	
	@param argv0 String containing the imput output file/paramters*/
void usage(std::string argv0) {
    std::cout <<"\nAuthor : Patrick HEAS	\n";
    std::cout << "\nusage: " << ">"
              << argv0 <<std::endl; 
    std::cout << "TLs_FileName\tMT_FileName\tFSDimension\tlines\trows\tGraphOutPutFormat\tIntegrationResolutionFactor (default:10.)\tIntensiveComputation (1= yes/0= no)\n" << std::endl;     
    std::cout <<"for the GraphOutPutFormat:\n\t1:standard\n\t2:adapted to the actual learning algorithms and visualisation by the GUI\n\t3:adapted to the actual learning algorithms and visualisation without edges by the GUI"<<std::endl; 
    std::cout << "\n*****************************************************************************************************************************************\n";
  }

/*!	@brief Main procedure which takes as input TL and MT classfiles and attributes files and creates a graphe in various format as an output*/
	/*!The graphe created are saved on the working directory. 
	@param argv0[1]  Text file name of input TL classfiles and attribute files : the classfiles and attribute files are listed one after the other; The number of lines of this texte file must be equal to 2 X numberOfClassfiles.   
*    	@param argv0[2]  Text file name of input MT classfile and attribute file : the classfile and attribute file are listed one after the other; The number of lines of this texte file must be equal to 2   
*	@param argv0[3]  Dimensionnality of the TL fetaure space 
*	@param argv0[4]  Number of lines of the images
*	@param argv0[5]  Number of columns of the images 
*	
*	Optional parameters  :
*	@param argv0[6]   Format of the Graph OutPut\n
			 1:standard (for future), 2: adapted to the actual interactive learning algorithm and GUI. 3: adapted to the actual interactive learning algorithm and  adapted to the actual GUI where the edge are not displayed (default is 2).
*	@param argv0[7]  Resolution factor for the divergence integration\n
	 (default equals 10).
*	 @param argv0[8]  (0= mean/1= divergence). If "0" the divergence calculation between consective TL clusters for edge caracterization is replaced by a simaple mean difference (default equals 0) 
	
*	
*/	
int main(int argc, char *argv [])
{   
 
	std::string ifn1;
	std::string ifn2;
	unsigned long   int         nb_bands	   	 = 0;
	unsigned long   int         dimx           	 = 0;
	unsigned long   int         dimy	   	 = 0;
	int 			    format		 = 2;
	float 			    PrecisionFactor      = 10.;
	int 			    intensiveCalculus    =0;
	switch (argc-1) {
		case 8:  intensiveCalculus      = atoi(argv[8]);
		case 7:  PrecisionFactor        = atof(argv[7]);
		case 6:  format		        = atoi(argv[6]);
        	case 5:  dimy       		= atoi(argv[5]); 
	        case 4:  dimx       		= atoi(argv[4]);
		case 3:  nb_bands   		= atoi(argv[3]);
        	case 2:  ifn2	 		= std::string(argv[2]);
	        case 1:  ifn1	 		= std::string(argv[1]); break;
        	case 0:
	        default: usage(argv[0]); return -1;
    	}
    	if ((argc-1) < 5) { usage(argv[0]); return -1;}
	std::cout << "> " << argv[0] 
              << " " << ifn1 
              << " " << ifn2
	      << " " << nb_bands 
              << " " << dimx 
              << " " << dimy
	      << " " << format
	      << " " << PrecisionFactor
	      << " " << intensiveCalculus<< std::endl;
	     
	srand((unsigned)time(NULL));
	time_t tdeb, t1,t2;
	time(&tdeb);
	t1=tdeb;
     	
/*****************Building the Graph*****************************************************************/
	/*! Creating the GraphDynaCluster object used for infering and  storing the graph of dynamic cluster trajectories*/
	GraphDynaCluster graph;
	/*!Initialisation the MT cluster trajectories : memory allocation, reading the input data */ 
	graph.Initialise((char *)ifn1.c_str(),(char *)ifn2.c_str(),dimx,dimy,nb_bands);
	/*!Inference of the MT cluster trajectories*/ 
	graph.Build(PrecisionFactor);
	
/*****************Mutual Information in consecutive time for each MT class **************************/
	std::cout<<"MI"<<std::endl;
	/*!Characterization of the trajectories with mutual information */ 	
	graph.MIEstimation();
	
/*****************OutPutSaving***********************************************************************/		
	std::cout<<"OutPut..."<<std::endl;
	bool aux=false;
	if (intensiveCalculus==1){aux=true;}
	/*! Caracterization of the trajectories and writing the graph of trajectories on the disk */ 	
	graph.Out_Evolutions(format, aux);
/****************************************************************************************************/
	
	time(&t2);
	cout<< "Processing time duration : "<<t2-t1<<" seconds."<<endl;
	return 0;
}
