/*---------------------------------------------------------------------------*/
/*!
 * @file MakeClusterAtt.cpp
 * @brief Program creating an attribute file associated to a classfile*/
 /*!
 *
 * This is the main program for generating an attribute file associated to a classfile. It uses as an input a multidimenionnal image and a clasfile to create an attribute file in a specific format
 *
 *
  *
 *
 * @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/
#include <iostream>
#include <string>  
#include <vector>
#include <fstream>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <ctime>

#include "Image_basics.hh"

#define PIX  short int

using namespace std;

/*!@brief Main procedure for reading the input data, performing basic calculation, and saving a attribute file.
*
* The mean vectors, the covariances matrices and the class weights are calculated according to the input classfile and the multidimensionnal data.
*/
class Mapping
{		
	public:
		/*!@brief Reading input files output file names and allocating memory*/
		void Acquire(char * path_in, char * path_in_class, char * path_out, int dimx_tmp, int dimy_tmp,int ofx_tmp, int ofy_tmp, int Dy_tmp, int subx_tmp, int suby_tmp) 
		{
			Acquire_path(path_in,path_in_class,path_out);
			Acquire_parameters_allocateImage(dimx_tmp,dimy_tmp,ofx_tmp,ofy_tmp,Dy_tmp,subx_tmp,suby_tmp);
			Allocate_OutMeanCov();
			
		}
		
		/*--------------------------------------------------------------------------------------------*/
		 /*!@brief Calculating and writing onthe disk the Gaussian Mixture parameters attached to the classification*/
		 void Create()
		 {
			Calculate();
			Write_OutMeanCov();
		 }
		
		/*--------------------------------------------------------------------------------------------*/
		
	private:

		/*--------------------------------------------------------------------------------------------*/ 
		/*!@brief image data*/
		Image<PIX> data;
		/*!@brief as we are dealing with a single multidimensionnal image, this is the same attribute as the attribute "data" ( thus one of these two has to be suppress)*/
		Image<PIX> CurrentIm;// to be removed because using a single image
		/*!@brief image classfile*/
		Image<unsigned char> classfile;
		/*!@brief  image where the Gaussian mixture attributes are stored */
		Image<float> OutMeanCov;
		/*!@brief output paths*/
		char * PathOutPut;
		/*!@brief input image path (a simple  pointer would be sufficiengt here!!)*/
		char ***PathInPut;
		/*!@brief input classfile path (a simple  pointer would be sufficient here!!)*/
		char ***PathInPutClass;
		/*!@brief array storing the population of each class*/
		unsigned long int * effectifSave;
		/*--------------------------------------------------------------------------------------------*/
		
		
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief acquire the input data paths*/
		void Acquire_path(char * path_in, char *path_in_class,char * path_out)  
		{
			
			int m,k,nb_images=1;
			Image<unsigned char> pippo;
			int nb_bands=pippo.NumberImages_pathfiles_in(path_in,nb_images);

			PathInPut= new char **[nb_images];
			for(m=0;m<nb_images;++m){PathInPut[m] =new char*[nb_bands];} 
			for(m=0;m<nb_images;++m){for(k=0;k<nb_bands;++k){PathInPut[m][k]=new char[128];}}							
			pippo.Read_pathfiles_in(path_in,PathInPut,nb_bands);

			data.nb_images=1;
			data.nb_bands=nb_bands;

			PathInPutClass=new char** [1];PathInPutClass[0]=new char*[1];
			PathInPutClass[0][0]=path_in_class;
			PathOutPut=path_out;
								
			
		}		
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief  Allocation of memory for storing the data and reading theinput data*/
		void Acquire_parameters_allocateImage(int dimx, int dimy, int ofx, int ofy, int Dy, int subx, int suby)
		{
			data.Allocate((int)(dimx/subx),(int)(dimy/suby),data.nb_bands,1);
			CurrentIm.Allocate((int)(dimx/subx),(int)(dimy/suby),data.nb_bands,1);
			classfile.Allocate((int)(dimx),(int)(dimy),1,1);
			data.Read_Cut_SubSample(PathInPut,ofx,ofy,Dy,subx,suby);
			classfile.Read_Cut_SubSample(PathInPutClass,0,0,dimy,1,1);
			
		}
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief  Allocation of memory for the classification attribute variable*/
		void Allocate_OutMeanCov()
		{
			OutMeanCov.Allocate(255,(data.nb_bands+1)*data.nb_bands,1,1);
			for (int j=0;j<(data.nb_bands+1)*data.nb_bands*255;j++)
				*(OutMeanCov.position+j)=0;
		}
		/*--------------------------------------------------------------------------------------------*/
		/*!@brief  Calculation of the classification attribute variable*/
		void Calculate()
		{
			int k,j,j1,j2;
			int nb_class=(int)classfile.MAX(0)+1;
			//cout<<"nb class : "<<nb_class<<endl;
			float *mean=new float[data.nb_bands];
			int auxB;
			float auxCov;
			unsigned long int effectif;
			//for (j=0;j<data.nb_bands;j++){mean[j]=data.Mean(j);cout<<"Mean "<<j<<" : "<<mean[j]<<endl;}
			//storing on the 1st line the nb of classes followed by zeros...
			*(OutMeanCov.position)=nb_class;
			effectifSave=new unsigned long int[nb_class];
			unsigned long int nbSample=data.nx*data.ny;	
		
			for (k=0;k<nb_class;k++)
			{
				effectif=0;
				for (unsigned long int i=0;i<nbSample;i++)
				{
					  auxB=*(classfile.position+i);
					
					  if (auxB==k)
					  {	
					  	effectif++;
					  	for (j=0;j<data.nb_bands;j++){
						   *(CurrentIm.position+(j)*nbSample+i)=*(data.position+(j)*nbSample+i);
						}
					  }
					  else
					  {
					  	for (j=0;j<data.nb_bands;j++){
					  	   *(CurrentIm.position+(j)*nbSample+i)=0;
						}
					  }
					
				}
				effectifSave[k]=effectif;
				for (j=0;j<data.nb_bands;j++){mean[j]=(CurrentIm.Mean(j)*nbSample/(float)effectif);}
				
				//storing on each line :
				for (j=0;j<data.nb_bands;j++)
				{
					//the cluster centroids
					*(OutMeanCov.position+(k+1)*OutMeanCov.ny+j)=mean[j];//
	
				}
				for (j1=0;j1<data.nb_bands;j1++)
				{
					for (j2=j1;j2<data.nb_bands;j2++)
					{
					    auxCov=0;
					    for (unsigned long int i=0;i<nbSample;i++){
					      auxB=*(classfile.position+i);
					      if (auxB==k){
						   auxCov=auxCov+(*(CurrentIm.position+(j1)*nbSample+i)-mean[j1])*(*(CurrentIm.position+(j2)*nbSample+i)-mean[j2]);	
					      }
					    }
					    auxCov=(auxCov/(float)effectif);
					   //the cluster covariance matrix
				 	  *(OutMeanCov.position+(k+1)*OutMeanCov.ny+(j1+1)*data.nb_bands+j2)=auxCov;
					}
				}
				for (j1=0;j1<data.nb_bands;j1++){
					for (j2=j1;j2<data.nb_bands;j2++){
						*(OutMeanCov.position+(k+1)*OutMeanCov.ny+(j2+1)*data.nb_bands+j1)=*(OutMeanCov.position+(k+1)*OutMeanCov.ny+(j1+1)*data.nb_bands+j2);
					}
				}
				

			}
			
			
			
		}
			
		/*! @brief Writing the attribute file on the standard output*/
		void Write_OutMeanCov()
			{
			
			OutMeanCov.Write_singleImage(PathOutPut);
			cout<<data.nb_bands<<" "<<*(OutMeanCov.position)<<endl;
			for (int k=0;k<*(OutMeanCov.position);k++){
				cout<<k<<" "<<(float)effectifSave[k]/(float)(data.nx*data.ny)<<endl;
				for (int j=0;j<data.nb_bands;j++){
					cout<<*(OutMeanCov.position+(k+1)*OutMeanCov.ny+j)<<" ";
				}
				cout<<""<<endl;
				for (int j1=0;j1<data.nb_bands;j1++)
				{
					for (int j2=0;j2<data.nb_bands;j2++)
					{
						cout<<*(OutMeanCov.position+(k+1)*OutMeanCov.ny+(j1+1)*data.nb_bands+j2)<<" ";
					}
				}
				cout<<""<<endl;
			}
			}
		/*--------------------------------------------------------------------------------------------*/	
		
}; 



/*! 	@brief  Usage function 	
	@param argv0 String containing the imput output file/paramters*/
void usage(std::string argv0) {
    std::cout << "\nusage: 				" << ">"
              << argv0 
              << "input_images_file_name input_image_classfile.cl output_classfile.att nb_lignes ncolonnes " << std::endl; 
    std::cout << "option for cutting:		 " 
    	      << "offset_colonnes offset_lignes nb_colonnes_original"<< std::endl;
    std::cout << "option for subsampling :	 " 
    	      << "sub_sample_ligne sub_sample_colonne ([ny]div[ny] must = 0 )"<< std::endl;
  }

/*!	@brief Main procedure.
*
*Generating an attribute file associated to a classfile. The calculation of the Gaussian mixture attributes is done via the class Mapping.
	@param argv0[1]  Text file name of an input mutidimensionnal image : the image bands  are listed one after the other; The number of lines of this file must be equal to the nuber of image bands   
*    	@param argv0[2]  Name of the input image classfile
*	@param argv0[3]  Name of the output attribute file 
*	@param argv0[4]  Number of lines of the images
*	@param argv0[5]  Number of columns of the images 
*	
*	Optional parameters for selecting a spatial window  :
*	@param argv0[6]  Column offset when reading the image file 
*	@param argv0[7]  Line offset when reading the image file
*	@param argv0[8]  Number of columns of the image file 

	Optional parameters for spatial subsampling   :
*	@param argv0[9]   factor to reduce each line (factor 1 = no subsampling)
*	@param argv0[10]  factor to reduce each column (factor 1 = no subsampling)
**/	
int main(int argc, char *argv []){    
std::string ifn1;
std::string ifn2;
std::string ifn3;
unsigned long  int          dimx           = 0;
unsigned long   int         dimy	   = 0;
unsigned long   int         ofx		   = 0;
unsigned long   int         ofy		   = 0;
unsigned long   int         Dy	 	   = 0;
unsigned long   int         subx	   = 1;
unsigned long   int         suby	   = 1;
switch (argc-1) {
    	case 10: suby 			= atoi(argv[10]);
	case 9:  subx  			= atoi(argv[9]);
        case 8:  Dy 			= atoi(argv[8]);
	case 7:  ofy  			= atoi(argv[7]);
        case 6:  ofx   			= atoi(argv[6]);
        case 5:  dimy       		= atoi(argv[5]);
        case 4:  dimx       		= atoi(argv[4]);
	case 3:  ifn3	 		= std::string(argv[3]);
        case 2:  ifn2	 		= std::string(argv[2]);
        case 1:  ifn1	 		= std::string(argv[1]); break;
        case 0:
        default: usage(argv[0]); return -1;
    }
    if ((argc-1) < 5) { usage(argv[0]); return -1;}
    if ((argc-1)==5) {Dy=dimy;} 
    std::cout << "> " << argv[0] 
              << " " << ifn1 
              << " " << ifn2
	      << " " << ifn3
              << " " << dimx 
              << " " << dimy 
              << " " << ofx
	      << " " << ofy  
              << " " << Dy 
	      << " " << subx  
              << " " << suby<< std::endl;
	 
/*******mapping **********************************************************/
 	srand((unsigned)time(NULL));
	time_t tdeb, t1,t2;
	time(&tdeb);
	t1=tdeb;
 	Mapping map;
	map.Acquire((char *)ifn1.c_str(),(char *)ifn2.c_str(),(char *)ifn3.c_str(),dimx,dimy,ofx,ofy,Dy,subx,suby);
	map.Create();
	time(&t2);
	return 0;
}

/*****************************************************************/
