/*---------------------------------------------------------------------------*/
/*!
 * @file MatchCost+.hh
 * @brief  file containing the MatchCost object used by GraphManipulator to perform the matching of sub-graphs according to a parametric similarity model
  @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/
#include <iostream>
#include <fstream>
#include <string>  
#include <vector>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>

#include "./SubGraph.hh"
#include "./SearchTree.hh"
#include "./divergence.h"


#define MIN(x, y) (((x) > (y)) ? (y) : (x))
#define semiInf 1e+2
 /*! @brief  Object used by GraphManipulator to perform sub-graph matchings according to a parametric similarity model*/
class MatchCost 
{


  
   /***************ATTRIBUTES****************************/
  
   public: 	
   //component weights of the inexact matching cost function
     /*! @brief  parameter weighting the population of each nodes*/
     float factor_weightMTinTLDiff;
     /*! @brief similarity model parameter weighting the nodes' distribution*/
     float factor_TL1TL2Div;
     /*! @brief  similarity model parameter weighting the flow of pixels exchanged between nodes	*/
     float factor_flowDiff;
     /*! @brief similarity model parameter weighting the change in time of the nodes' distribution*/
     float factor_divDiff;
     /*! @brief   similarity model parameter weighting the intra-node change in time*/
     float factor_MIDiff;
     /*! @brief similarity model parameter weighting the interaction of the 2 MT classes (only if 2 MT class are consideredper sub-graph)*/
     float factor_divMTaMTbCost;
     /*! @brief similarity model parameter weighting the time lattice on which the sub-graph is defined*/
     float factor_timeDiff;
     /*! @brief  mean calculation '1' or Kullback-Leibler divergence '0' used in the matching nodes and edges comparisons
     *
     *We remark that the Kullback-Leibler divergence is not adapted to the actual graph format 2, in which the divergence calculation is replaced by a simple distribution mean difference */
     float meanCalculation;

   
   private:     
   //subGraphs objects 
     /*! @brief reference SubGraph Object for matching */
     SubGraph *subGraph1;
      /*! @brief candidate SubGraph Object to match with the reference sub-graph */
     SubGraph *subGraph2;
     
     /*! @brief mean calculation 'true' or Kullback-Leibler divergence 'false' used in the matching nodes and edges comparison
     *
     *We remark that the Kullback-Leibler divergence is not adapted to the actual graph format 2, in which the divergence calculation is replaced by a simple distribution mean difference */
     bool meanCalcul;     
     /*! @brief divergence objects*/ 
     //Divergence *divObject;
      vector<Divergence> divObject;
     /*!@brief Current covariance matrice considered for divergence  */
     float **C_A; 
     /*!@brief Current mean vector considered for divergence  */
     float *M_A;
     /*!@brief Current covariance matrice considered for divergence  */
     float **C_B; 
     /*!@brief Current mean vector considered for divergence  */
     float *M_B;

  /************METHODES****************************/
  
  public:
    /*! @brief constructor Allocating memory and creating a Divergence object*/ 
     MatchCost(SubGraph *Graph1,SubGraph *Graph2){
     	     
     	     subGraph1=Graph1;
     	     subGraph2=Graph2;
             Divergence tmp(subGraph1->nbDim); 
     	     //divObject=new Divergence[1](subGraph1->nbDim);
             divObject.push_back(tmp);
            //divObject.erase(divObject.begin());
         //Divergence tmp2(subGraph1->nbDim); 
         // divObject.push_back(tmp2);
         C_A=new float *[subGraph1->nbDim];C_B=new float *[subGraph1->nbDim];
      	     M_A=new float[subGraph1->nbDim];M_B=new float[subGraph1->nbDim];
             for (int i=0;i<subGraph1->nbDim;i++){
       		C_A[i]=new float [subGraph1->nbDim];C_B[i]=new float [subGraph1->nbDim]; 
            }
        }
     /*! @brief destructor desAllocating memory */ 
     virtual~MatchCost(){
     	     
         //cout<<"desctructiopn de MatchCost"<<endl;
         //delete [] divObject;
         //divObject.erase(divObject.begin());
	     //for (int i=0;i<subGraph1->nbDim;i++){delete [] C_A[i];delete [] C_B[i];}
	     //delete [] C_A;delete [] M_A;delete [] C_B;delete [] M_B;

     }
     /*! @brief This method introduce in the similarity model  a given parameter vector. This method fixes also another parameter specific to the matching calculus.
	*
	*@param weightMTinTLDiff  parameter weighting the population of each nodes				
	*@param	TL1TL2Div parameter weighting the nodes' distribution			
	*@param	flowDiff parameter weighting the flow of pixels exchanged between nodes					
	*@param	divDiff parameter weighting the change in time of the nodes' distribution					
	*@param	MIDiff parameter weighting the intra-node change in time					
	*@param	divMTaMTbCost parameter weighting the interaction of the 2 MT classes (only if 2 MT class are considered);		
	*@param	timeDiff parameter weighting the time lattice on which the sub-graph is defined
	*@param meanCalcul mean calculation 'true' or Kullback-Leibler divergence 'flase' used in the matching calculation */
     void FixCostFuncWeigths(float weightMTinTLDiff,float TL1TL2Div,float flowDiff,float divDiff,float MIDiff, float divMTaMTbCost, float timeDiff,bool meanCalcul){
     	     
     	     factor_weightMTinTLDiff=weightMTinTLDiff;
     	     factor_TL1TL2Div=TL1TL2Div;
     	     factor_flowDiff=flowDiff;
     	     factor_divDiff=divDiff;
     	     factor_MIDiff=MIDiff;
	     factor_timeDiff=timeDiff;
	     factor_divMTaMTbCost=divMTaMTbCost;
	     meanCalculation=meanCalcul;
     		     
     }
      /*! @brief Method to update the pointer to a new candidate SubGraph Object*/
     void UpdateGraph2(SubGraph *Graph2){
     
		subGraph2=Graph2;
     }
     /*! @brief Method to match the reference sub-graph with the candidate sub-graph based on the parametric similarity model.
     *
     * The idea of this algorithm is to find a collection of elementary transformation wich maps, with a minimal cost, each node and related edges of the reference graph to an existing node or a virtual node of the sub-graph candidate at the same time sample.\n
     The cost assigned to each elementary matching result from the weighted sum of costs related to each graph attributes (population of the node, flow of points, etc...). The weighting is done using the current estimate of the similarity model parameter vector.\n
      Each element of this sum is thus a comparison, by scalar difference or distribution divergence, of the attributes similarities between the couple of nodes (or edges).\n
       When matching with an existing node, this comparison is straight fowards; but when the matching is performed with a virtual node, the virtual node attribute are simply 0 or attribute related to the MT cluster associated to the node. Inddeed, a node represents a TL cluster which is associated to a MT class. \n
       The problem gets complicate when we consider simultaneously 2 MT classes per graph. In this case, at each time sample, the nodes of the 2 MT classes are considered separately one by one. And, as sub-graph nodes related to a given MT class can not be matched with nodes of the candidate sub-graph  belonging to 2 different MT classes, the consistencyof the matching must be checked inthis case. This checking is done by the methods  'CheckMTCoherentEdge1(..)' and 'CheckMTCoherentEdge2'.   
     To find the best collection of elementary transformation (mapping function), a SearchTree object is used. This object  is a tree constituted  of all possible mapping functions:\n
     \li at the beginning a first node of the reference graph is matched with all posssible existing or virtual nodes of the sub-graph candidate present at the same time sample.  The collection costs assigned to these elementary tranformations is stored in the first level of the tree.\n
     NB: note that the mapping of edges are related to nodes. Thus they are considered simultaneously with the node matchings.  
     \li Then a second node of the reference graph is matched with all posssible existing or virtual nodes of the sub-graph candidate present at the same time sample. This collection of costs assigned to these elementary tranformations is linked to the each  element of the collection of costs of the firts level. This collection of collection constitutes  the second levels of the tree. 
     \li the procedure is reiterated for each node of the reference sub-graph until no nodes are left. \n
     * Once the search tree has been built, the best mapping function is the path in the tree associated to a minimal sum of node matching costs. However, because the number of ramification of such a tree is much too high with the combinatorial explosion of solutions, only 'nbMiniBranches' of possible mapping functions (paths in the tree from root to the curent leaves) are preserved after each elementary matching of a new node.\n
     *\n
     *As we are interested in decomposing each possible mapping function cost in a sum of costs related to a given graph attribute, the cost assigned to the best mapping function is decomposed into several parts related to each attribute contribution. These different cost constribution are stored in the array 'costs'.\n
     NB: note that the search tree created by this procedure will be used later to update the cost of the mapping function with a new similarity model parameter vector estimate. This update will only by the estimation of the best path in the search tree according to the new parameter vector. 
     @param nbMiniBranches number of possible mapping functions (paths in the tree from root to the curent leaves)  preserved after each elementary matching of a new node
     @param searchTree   SearchTree Object resulting form the matching procedure. A number of 'nbMiniBranches' possible mapping functions are stored in it. 
     @param costs  collection of costs related to each graph attribute contribution, resulting of the decomposition of the best mapping function costs.*/
     
    
     float InexactMatchCost(int nbMiniBranches, SearchTree* searchTree, float *costs){
 
       float cost, partialCost;
       float interactTimeCost=-1; 
       float auxCost_weight,tmpCost_weight;
       
       float tmpCost_TL1TL2Div; 
       float *resolution=new float[subGraph1->nbDim];
       float *min_bound=new float[subGraph1->nbDim];float *max_bound=new float[subGraph1->nbDim];
       
       float tmpCost_flow=0;
       float tmpCost_div=0;int calculateDiv;
       float tmpCost_MI=0;
       float tmpCost_time=0;
       int indexaux;
       float auxabs;
       int* nbNodes1=new int[subGraph1->nbTimeSamp];
       for (int time=0;time<subGraph1->nbTimeSamp;time++){
       		nbNodes1[time]=0;
       		for (int nodek1=0;(unsigned int)nodek1<subGraph1->meanTL[time].size();nodek1++){
			nbNodes1[time]+=subGraph1->weightMTinTL[time][nodek1].size();
		}
       }
       
       searchTree[0].Init(subGraph1->nbTimeSamp,subGraph1->nbMTClusters,nbNodes1);
       
       int currentLevel=0;
       bool root;
       int *contMTkforMI=new int[subGraph1->nbMTClusters];
       int nbTrunc;
       for (int time=0;time<subGraph1->nbTimeSamp;time++){
         //cout<<"matching time : "<<time<<endl;
	 interactTimeCost=0;
         if (subGraph1->nbMTClusters==2){
	    	LoadMT1aMT1bMeanCov(time,0,1);
		divObject[0].Initialise (C_A,C_B,M_A,M_B);
		if (meanCalculation){ interactTimeCost+=divObject[0].MeanDifference();}
		else{
			divObject[0].InvertCovMatrix();
			calculateDiv=divObject[0].BoundResolution_Calculation(min_bound,max_bound,resolution,10.); 
			if(calculateDiv<1){interactTimeCost+=MIN(divObject[0].Divergence2GaussModel(min_bound,max_bound,resolution,1),divObject[0].MeanDifference());}
			else{interactTimeCost+=divObject[0].MeanDifference();}
		}
		
		LoadMT2aMT2bMeanCov(time,0,1);
		divObject[0].Initialise (C_A,C_B,M_A,M_B);
		if (meanCalculation){ interactTimeCost-=divObject[0].MeanDifference();}   
		else{
			divObject[0].InvertCovMatrix();
			calculateDiv=divObject[0].BoundResolution_Calculation(min_bound,max_bound,resolution,10.);   
			if(calculateDiv<1){interactTimeCost-=MIN(divObject[0].Divergence2GaussModel(min_bound,max_bound,resolution,1),divObject[0].MeanDifference());}
			else{interactTimeCost-=divObject[0].MeanDifference();}
		} 
		interactTimeCost=factor_divMTaMTbCost*sqrt(interactTimeCost*interactTimeCost);
         }
	 tmpCost_time=0;
	 if(time>0){auxabs=subGraph1->tDelay[time-1]-subGraph2->tDelay[time-1];tmpCost_time=sqrt(auxabs*auxabs)*factor_timeDiff;}
	 tmpCost_MI=0;
	 for (int MTkforMI=0;MTkforMI<subGraph1->nbMTClusters;MTkforMI++){contMTkforMI[MTkforMI]=0;}
         for (int nodek1=0; (unsigned int)nodek1<subGraph1->meanTL[time].size();nodek1++){
	 //cout<<"matching nodek1 : "<<nodek1<<endl;	     
	   for (int MTk1=0;(unsigned int)MTk1<subGraph1->weightMTinTL[time][nodek1].size();MTk1++){
	     //weightMTinTLDiff
	     auxCost_weight =(float)subGraph1->weightMTinTL[time][nodek1][MTk1];
	     auxCost_weight /= (float)subGraph1->weightMT[subGraph1->indexMTinTL[time][nodek1][MTk1]];
	     /**/
	     //existing nodes 
	     root=(currentLevel==0);
	     if (root){nbTrunc=1;}else{nbTrunc=searchTree[0].tree[currentLevel-1].size();}
     	     for (int preIndex=0;preIndex<nbTrunc;preIndex++){
 	       for (int nodek2=0;(unsigned int)nodek2<subGraph2->meanTL[time].size();nodek2++){
     	         for (int MTk2=0;(unsigned int)MTk2<subGraph2->weightMTinTL[time][nodek2].size();MTk2++){	       	 	  		 
		   if (searchTree[0].CheckUnusedNode(time,currentLevel,preIndex,nodek2,subGraph2->indexMTinTL[time][nodek2][MTk2])){		  
		    if (searchTree[0].CheckMTCoherentNode(time,currentLevel,preIndex,subGraph1->indexMTinTL[time][nodek1][MTk1],subGraph2->indexMTinTL[time][nodek2][MTk2])){				
     	              //weightMTinTLDiff
		      tmpCost_weight=auxCost_weight;
		      tmpCost_weight-=(float)subGraph2->weightMTinTL[time][nodek2][MTk2]/(float)subGraph2->weightMT[subGraph2->indexMTinTL[time][nodek2][MTk2]];
		      tmpCost_weight=sqrt(tmpCost_weight*tmpCost_weight);//(absolute value)
     	              tmpCost_weight*=factor_weightMTinTLDiff;
		      partialCost=tmpCost_weight;
		      /**/
		      //TL1TL2Div
		      if (factor_TL1TL2Div!=0){
		      	LoadTL1TL2MeanCov(time,nodek1,nodek2);
		      	divObject[0].Initialise (C_A,C_B,M_A,M_B);
			if (meanCalculation){
				tmpCost_TL1TL2Div=divObject[0].MeanDifference();
				tmpCost_TL1TL2Div=sqrt(tmpCost_TL1TL2Div*tmpCost_TL1TL2Div);
			}
			else{
		      		calculateDiv=divObject[0].BoundResolution_Calculation(min_bound,max_bound,resolution,10.);
				if (calculateDiv<1){
					divObject[0].InvertCovMatrix();
					tmpCost_TL1TL2Div=MIN(divObject[0].Divergence2GaussModel(min_bound,max_bound,resolution,1),divObject[0].MeanDifference());
				}
				else{tmpCost_TL1TL2Div=divObject[0].MeanDifference();}
			}
			
		      }
		      else{tmpCost_TL1TL2Div=0;}
		      tmpCost_TL1TL2Div*=factor_TL1TL2Div;
		      partialCost+=tmpCost_TL1TL2Div;	  		      
		      /**/
		      //flowDiff & divDiff
		      if (time>0){
		        tmpCost_flow=0;
			tmpCost_div=0;		
			for (int k=0;(unsigned int)k<subGraph1->meanTL[time-1].size();k++){
			 if (CheckMTCoherentEdge1(time,MTk1,k,nodek1)){
			  indexaux=searchTree[0].getPreTimeCorresNodeIndex(time,currentLevel,preIndex,k,subGraph1->indexMTinTL[time][nodek1][MTk1]);			  
			   //linked to an existing node
			   if (indexaux != -1){
			     if (CheckMTCoherentEdge2(time,MTk2,indexaux,nodek2)){	auxabs=(float)subGraph1->flowMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1]/(float)subGraph1->weightMT[subGraph1->indexMTinTL[time][nodek1][MTk1]]-(float)subGraph2->flowMatrix[time][subGraph2->indexMTinTL[time][nodek2][MTk2]][indexaux][nodek2]/(float)subGraph2->weightMT[subGraph2->indexMTinTL[time][nodek2][MTk2]];
			   	auxabs=sqrt(auxabs*auxabs);//(absolute value)
		           	tmpCost_flow+=auxabs*factor_flowDiff;
			        
				auxabs=0;
				if (!meanCalculation){
					auxabs=subGraph1->divMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1]-subGraph2->divMatrix[time][subGraph2->indexMTinTL[time][nodek2][MTk2]][indexaux][nodek2];

				}
				if (meanCalculation||sqrt(auxabs*auxabs)>semiInf){	
					LoadTL1preTL1MeanCov(time,nodek1,k);
					divObject[0].Initialise (C_A,C_B,M_A,M_B);
					auxabs=divObject[0].MeanDifference();
					LoadTL2preTL2MeanCov(time,nodek2,indexaux);
					divObject[0].Initialise (C_A,C_B,M_A,M_B);
					auxabs-=divObject[0].MeanDifference();											}
			   	auxabs=sqrt(auxabs*auxabs);//(absolute value)
		           	tmpCost_div+=auxabs*factor_divDiff;
				
				if(contMTkforMI[subGraph1->indexMTinTL[time][nodek1][MTk1]]==0){				
				   auxabs=(float)subGraph1->MIMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1]-(float)subGraph2->MIMatrix[time][subGraph2->indexMTinTL[time][nodek2][MTk2]][indexaux][nodek2];
				   auxabs=sqrt(auxabs*auxabs);
				   tmpCost_MI+=auxabs*factor_MIDiff;
				   contMTkforMI[subGraph1->indexMTinTL[time][nodek1][MTk1]]=1;
				}
			     }
				
			   }
			   //linked to an added extra node
			   else{
			   	
				auxabs=(float)subGraph1->flowMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1]/(float)subGraph1->weightMT[subGraph1->indexMTinTL[time][nodek1][MTk1]];
				tmpCost_flow+=auxabs*factor_flowDiff; 	
				
				indexaux=searchTree[0].getPreTimeCorresMTIndex(time,currentLevel,preIndex,k,subGraph1->indexMTinTL[time][nodek1][MTk1]);
				auxabs=0;
				if (!meanCalculation){
					auxabs+=subGraph1->divMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1];	
				}
				if (meanCalculation||auxabs>semiInf){
					auxabs=0;
					LoadTL1preTL1MeanCov(time,nodek1,k);
					divObject[0].Initialise (C_A,C_B,M_A,M_B);
					auxabs+=divObject[0].MeanDifference();					}
				
				LoadTL2preMT2MeanCov(time,nodek2,indexaux); 
				divObject[0].Initialise (C_A,C_B,M_A,M_B);
				if (meanCalculation){auxabs-=divObject[0].MeanDifference();}
				else{			
		      			calculateDiv=divObject[0].BoundResolution_Calculation(min_bound,max_bound,resolution,10.);
					if (calculateDiv<1){
						divObject[0].InvertCovMatrix();
						auxabs-=MIN(divObject[0].Divergence2GaussModel(min_bound,max_bound,resolution,1),divObject[0].MeanDifference());
					}
					else{auxabs-=divObject[0].MeanDifference();}
				}
				tmpCost_div+=sqrt(auxabs*auxabs)*factor_divDiff;
			   }
			  
			 }
			}
			partialCost+=tmpCost_flow;
			partialCost+=tmpCost_div;
		      }
		      /**/
		      
		      //a completer avec MI!!!
		      partialCost+=tmpCost_MI;
		      partialCost+=tmpCost_time;
 		      partialCost+=interactTimeCost;
		      cost=partialCost;
		     searchTree[0].AddNode(time,currentLevel,preIndex,nodek1,subGraph1->indexMTinTL[time][nodek1][MTk1],nodek2,subGraph2->indexMTinTL[time][nodek2][MTk2],cost,tmpCost_weight,tmpCost_TL1TL2Div,tmpCost_flow,tmpCost_div,tmpCost_MI,interactTimeCost,tmpCost_time);
		      
		    }
		   }  
	         }
     	       }
	       //extra node
	       
	       //TL1TL2Div
	       //int MTindex=-1;
	       for (int MTk2=0;MTk2<subGraph2->nbMTClusters;MTk2++){
	        if (searchTree[0].CheckMTCoherentNode(time,currentLevel,preIndex,subGraph1->indexMTinTL[time][nodek1][MTk1],MTk2)){
	         //weightMTinTLDiff
	         tmpCost_weight*=factor_weightMTinTLDiff;
	         partialCost=tmpCost_weight;
	         /**/
		 if (factor_TL1TL2Div!=0){
	       	 	LoadTL1MT2MeanCov(time,nodek1,MTk2);
		 	divObject[0].Initialise (C_A,C_B,M_A,M_B);
			if (meanCalculation){ 
				tmpCost_TL1TL2Div=divObject[0].MeanDifference();
				tmpCost_TL1TL2Div=sqrt(tmpCost_TL1TL2Div*tmpCost_TL1TL2Div);
			}
			else{
				calculateDiv=divObject[0].BoundResolution_Calculation(min_bound,max_bound,resolution,10.);
				if (calculateDiv<1){	
					divObject[0].InvertCovMatrix();
					tmpCost_TL1TL2Div=MIN(divObject[0].Divergence2GaussModel(min_bound,max_bound,resolution,1),divObject[0].MeanDifference());
				}
				else{tmpCost_TL1TL2Div=divObject[0].MeanDifference();}
			}
			
		 	//MTindex=MTk2;
		 }
		 else{tmpCost_TL1TL2Div=0;}
	         tmpCost_TL1TL2Div*=factor_TL1TL2Div;
	         partialCost+=tmpCost_TL1TL2Div;
	         /**/
	         //flowDiff
	         if (time>0){
		   tmpCost_flow=0;
		   tmpCost_div=0;
		   for (int k=0;(unsigned int)k<subGraph1->meanTL[time-1].size();k++){
		    if (CheckMTCoherentEdge1(time,MTk1,k,nodek1)){
		      tmpCost_flow+=(float)subGraph1->flowMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1]/(float)subGraph1->weightMT[subGraph1->indexMTinTL[time][nodek1][MTk1]]*factor_flowDiff;
		 
		      indexaux=searchTree[0].getPreTimeCorresNodeIndex(time,currentLevel,preIndex,k,subGraph1->indexMTinTL[time][nodek1][MTk1]);
		      if (indexaux != -1){
		      //extra node linked to an existant node
		         auxabs=0;
			 if (!meanCalculation){
				auxabs+=subGraph1->divMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1];
			 }
			 if (meanCalculation||auxabs>semiInf){
				auxabs=0;
				LoadTL1preTL1MeanCov(time,nodek1,k);
				divObject[0].Initialise (C_A,C_B,M_A,M_B);
				auxabs+=divObject[0].MeanDifference();
			}
		     	 LoadpreTL2MT2MeanCov(time,indexaux,MTk2); 
			 divObject[0].Initialise (C_A,C_B,M_A,M_B);
			 if (meanCalculation){auxabs-=divObject[0].MeanDifference();}
			 else{
			 	 divObject[0].InvertCovMatrix();
				 calculateDiv=divObject[0].BoundResolution_Calculation(min_bound,max_bound,resolution,10.);
			 	 if (calculateDiv<1){auxabs-=MIN(divObject[0].Divergence2GaussModel(min_bound,max_bound,resolution,1),divObject[0].MeanDifference());}
				 else{auxabs-=divObject[0].MeanDifference();}
			}
			 auxabs=sqrt(auxabs*auxabs);//(absolute value)
		         tmpCost_div+=auxabs*factor_divDiff;
		      }
		      else{
		      //extra node linked to another extra node
		         auxabs=0; 
			 if (!meanCalculation){
			 	auxabs+=subGraph1->divMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1];
			 }
			 if (meanCalculation||auxabs>semiInf){
				auxabs=0;
				LoadTL1preTL1MeanCov(time,nodek1,k);
				divObject[0].Initialise (C_A,C_B,M_A,M_B);
				auxabs+=divObject[0].MeanDifference();
			 }	 
			 indexaux=searchTree[0].getPreTimeCorresMTIndex(time,currentLevel,preIndex,k,subGraph1->indexMTinTL[time][nodek1][MTk1]);
		    	 LoadpreMT2MT2MeanCov(time,indexaux,MTk2); 
			 divObject[0].Initialise (C_A,C_B,M_A,M_B);
			 if (meanCalculation){auxabs-=divObject[0].MeanDifference();}
			 else{
			 	divObject[0].InvertCovMatrix();
			 	calculateDiv=divObject[0].BoundResolution_Calculation(min_bound,max_bound,resolution,10.);
			 	 if (calculateDiv<1){auxabs-=MIN(divObject[0].Divergence2GaussModel(min_bound,max_bound,resolution,1),divObject[0].MeanDifference());}
				 else{auxabs-=divObject[0].MeanDifference();}
			}
			 auxabs=sqrt(auxabs*auxabs);//(absolute value)
		         tmpCost_div+=auxabs*factor_divDiff;
		      }
		    }
		   }
		   partialCost+=tmpCost_flow;
		   partialCost+=tmpCost_div;
	           }
	           /**/
		   partialCost+=tmpCost_MI;
		   partialCost+=tmpCost_time;
		   partialCost+=interactTimeCost;
	           cost=partialCost;
	           searchTree[0].AddNode(time,currentLevel,preIndex,nodek1,subGraph1->indexMTinTL[time][nodek1][MTk1],-1,MTk2,cost,tmpCost_weight,tmpCost_TL1TL2Div,tmpCost_flow,tmpCost_div,tmpCost_MI,interactTimeCost,tmpCost_time);
		  
		  }
		 }
	       }
	     searchTree[0].PruneTree(nbMiniBranches,currentLevel);
	     currentLevel++;
	     interactTimeCost=0;
	     tmpCost_time=0;
	     tmpCost_MI=0;
     	   }
     	 }  
       }
       
       float miniCostmatch=searchTree[0].LowestCosts(costs);  

      delete [] contMTkforMI;
       return miniCostmatch;
 
     }

 
 
 private:     
 	/*!@brief method to load a node of the reference graph at time sample (t) and a node of the candidate graph at time sample (t)*/
     void LoadTL1TL2MeanCov(int time,int nodek1,int nodek2){
     	
	 for (int i=0;i<subGraph1->nbDim;i++){
	 	M_A[i]=subGraph1->meanTL[time][nodek1][i];M_B[i]=subGraph2->meanTL[time][nodek2][i];
		for (int j=0;j<subGraph1->nbDim;j++){
			C_A[i][j]=subGraph1->covTL[time][nodek1][i][j];C_B[i][j]=subGraph2->covTL[time][nodek2][i][j];
		}
     	 }
     }
     /*!@brief method to load one node of the reference graph at time sample (t) and a  MT class node related to the candidate graph at time sample (t) 
	*/
     void LoadTL1MT2MeanCov(int time,int nodek1, int MTk2){
     	
	 for (int i=0;i<subGraph1->nbDim;i++){
	 	M_A[i]=subGraph1->meanTL[time][nodek1][i];M_B[i]=subGraph2->meanMT[time][MTk2][i];
		for (int j=0;j<subGraph1->nbDim;j++){
			C_A[i][j]=subGraph1->covTL[time][nodek1][i][j];C_B[i][j]=subGraph2->covMT[time][MTk2][i][j];
		}
     	 }
     }
     /*!@brief method to load one node of the candidate graph  at time sample (t) and a  MT class node related to the candidate graph at time sample (t-1).
	 */
     void LoadTL2preMT2MeanCov(int time,int nodek1, int MTk2)
     {
         for (int i=0;i<subGraph2->nbDim;i++){
	 	M_A[i]=subGraph2->meanTL[time][nodek1][i];M_B[i]=subGraph2->meanMT[time-1][MTk2][i];
		for (int j=0;j<subGraph2->nbDim;j++){
			C_A[i][j]=subGraph2->covTL[time][nodek1][i][j];C_B[i][j]=subGraph2->covMT[time-1][MTk2][i][j];
		}
     	 }
     }
     /*!@brief method to load one nodes of the candidate graph at time sample (t-1) and a  MT class node related to the candidate graph at time sample (t).
	*/
     void LoadpreTL2MT2MeanCov(int time,int nodek1, int MTk2)
     {
         for (int i=0;i<subGraph2->nbDim;i++){
	 	M_A[i]=subGraph2->meanTL[time-1][nodek1][i];M_B[i]=subGraph2->meanMT[time][MTk2][i];
		for (int j=0;j<subGraph2->nbDim;j++){
			C_A[i][j]=subGraph2->covTL[time-1][nodek1][i][j];C_B[i][j]=subGraph2->covMT[time][MTk2][i][j];
		}
     	 }
     }
      /*!@brief method to load a  MT class node related to the candidate graph at time sample (t-1) and  a  MT class node related to the candidate graph at time sample (t-1).
	 */
     void LoadpreMT2MT2MeanCov(int time,int preMTk2, int MTk2)
     {
         for (int i=0;i<subGraph2->nbDim;i++){
	 	M_A[i]=subGraph2->meanMT[time-1][preMTk2][i];M_B[i]=subGraph2->meanMT[time][MTk2][i];
		for (int j=0;j<subGraph2->nbDim;j++){
			C_A[i][j]=subGraph2->covMT[time-1][preMTk2][i][j];C_B[i][j]=subGraph2->covMT[time][MTk2][i][j];
		}
     	 }
     } 
     /*!@brief method to load a  MT class node related to the reference graph at time sample (t) and a  MT class node related to the reference graph at time sample (t).
	 */
     void LoadMT1aMT1bMeanCov(int time,int MTk1, int MTk2)
     {
         for (int i=0;i<subGraph1->nbDim;i++){
	 	M_A[i]=subGraph1->meanMT[time][MTk1][i];M_B[i]=subGraph1->meanMT[time][MTk2][i];
		for (int j=0;j<subGraph1->nbDim;j++){
			C_A[i][j]=subGraph1->covMT[time][MTk1][i][j];C_B[i][j]=subGraph1->covMT[time][MTk2][i][j];
		}
     	 }
     }
     /*!@brief method to load a  MT class node related to the candidate graph at time sample (t) and a  MT class node related to the candidate graph at time sample (t).*/
     void LoadMT2aMT2bMeanCov(int time,int MTk1, int MTk2)
     {
         for (int i=0;i<subGraph2->nbDim;i++){
	 	M_A[i]=subGraph2->meanMT[time][MTk1][i];M_B[i]=subGraph2->meanMT[time][MTk2][i];
		for (int j=0;j<subGraph2->nbDim;j++){
			C_A[i][j]=subGraph2->covMT[time][MTk1][i][j];C_B[i][j]=subGraph2->covMT[time][MTk2][i][j];
		}
     	 }
     }
     /*!@brief method to load a  node of the reference graph at time sample (t) and a  node od the the reference graph at time sample (t-1).*/
     void LoadTL1preTL1MeanCov(int time,int nodek1,int preNodek1){
     	
	 for (int i=0;i<subGraph1->nbDim;i++){
	 	M_A[i]=subGraph1->meanTL[time][nodek1][i];
		M_B[i]=subGraph1->meanTL[time-1][preNodek1][i];
		for (int j=0;j<subGraph1->nbDim;j++){
			C_A[i][j]=subGraph1->covTL[time][nodek1][i][j];
			C_B[i][j]=subGraph1->covTL[time-1][preNodek1][i][j];
		}
     	 }
     }
     /*!@brief method to load a  node of the candidate graph at time sample (t-1) and a  node od the the candidate graph at time sample (t-1).*/
     void LoadTL2preTL2MeanCov(int time,int nodek2,int preNodek2){
     	
	 for (int i=0;i<subGraph2->nbDim;i++){
	 	M_A[i]=subGraph2->meanTL[time][nodek2][i];
		M_B[i]=subGraph2->meanTL[time-1][preNodek2][i];
		for (int j=0;j<subGraph2->nbDim;j++){
			C_A[i][j]=subGraph2->covTL[time][nodek2][i][j];
			C_B[i][j]=subGraph2->covTL[time-1][preNodek2][i][j];
		}
     	 }
     } 
     /*!@brief method to check if the edge related to a node to match of the reference graph are associated to the same MT class as the other previously matched nodes .*/
     bool CheckMTCoherentEdge1(int time,int MTk1,int k,int nodek1){
     
     	return (subGraph1->divMatrix[time][subGraph1->indexMTinTL[time][nodek1][MTk1]][k][nodek1]!=-1);
	
     }
     /*!@brief method to check if the edge related to a node to match of the candidate graph are associated to the same MT class as the other previously matched nodes .*/
     bool CheckMTCoherentEdge2(int time,int MTk2,int k,int nodek2){
    
     	return (subGraph2->divMatrix[time][subGraph2->indexMTinTL[time][nodek2][MTk2]][k][nodek2]!=-1);
	
     }
    
};
