/*---------------------------------------------------------------------------*/
/*!
 * @file ParaSearch.hh
 * @brief  file containing the ParaSearch object. It contains a collection of Multinomial distributions updated by dirichlet model hyper parameters used by GraphManipulator to learn the parameter vector of the similarity model. 
 *
  
  @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/#ifndef PARASEARCH_HH
#define PARASEARCH_HH

#include <iostream>
#include <fstream>
#include <string>  
#include <vector>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>

#define InfCost 1e5
#include "SearchTree.hh"


using namespace std;
/*! @brief   object containing a collection of Multinomial distributions updated by dirichlet model hyper parameters used by GraphManipulator to learn the parameter vector of the similarity model.
*
* For this, the Multinomial distributions, which are the parameter distributions learned by the system, are used to evaluate with MMSE estimator the currently learned parameter vector of the similarity model.*/ 
class ParaSearch
{
 public:
 	//general attributes
	/*! @brief number of elements in the parameter vector of the similarity model*/
	int nbPara;
	/*! @brief number of bins used for the interactive estimation of the multinomial distributions*/
	int nbQuantifLevel;
	/*! @brief number of MT classes composing the global graph.*/
	int nbMTClasses;
	/*! @brief number of MT classes considered per sub-graph in the matching procedure.*/
	int nbSubGMTClasses;
	/*! @brief Minimum time sample considered in the global graph  */
	int minSample;
	/*! @brief Maximum time sample considered in the global graph */
	int maxSample;
	/*! @brief  number of samples of the time-window*/
	int nbTimeSamples;
	
	//data
	/*!@brief pointer to collections of SearchTree objects*/ 
	SearchTree ****treeCollec;
	/*! @brief pointer to collections of distorsion  costs array. 
	*
	*The cost assigned during the matchingprocedure to the best mapping function is decomposed into several parts related to each attribute contribution. These different cost constribution are stored in the array 'costs'.*/ 
	float ****costs;
	/*! @brief vector of MT class indexes attached to the elements of the a posteriori collection of sub-graphs*/
	vector<int> kMTaSorted;
	/*! @brief vector of  2nd MT class indexes attached to the elements of the  a posteriori collection of sub-graphs (only when 2 MT classes are considered simultaneously per sub-graph)*/
	vector<int> kMTbSorted;
	/*! @brief array of  first time sample indexes attached to the elements of the a posteriori collection of sub-graphs*/
	vector<int> firstSampleSorted;
	
	//estimated PDF of parameters
	/*! @brief array of multinomial distributions for approximating the paramter vector of the similarity model PDF*/
	float **pdf;
	/*! @brief array of maximum values of the multinomial distributions*/
	float *maxPdf;
	/*! @brief array of mean values of the multinomial distributions*/
	float *meanPdf;
	/*! @brief array of MMSE  parameter estimates of the similarity model*/
	float *estimPara;
	/*! @brief array of MMSE  parameter estimates of the similarity model at the previous learning iteration*/
	float *estimParaPrevious;
	/*! @brief array of maximum cost values */
	float *maxCostValue;
	
 
 public :
	ParaSearch(){}
	virtual~ParaSearch(){}
	/*! @brief memory allocation method */
	void Allocate(int nb_parameters,int nbLevel,int nb_MTClasses, int nb_SubGMTClasses,int min_Sample,int max_Sample,int nb_TimeSamples){
		
		nbPara=nb_parameters;
		nbQuantifLevel=nbLevel;
		pdf=new float*[nbPara];
		maxPdf=new float[nbPara];
		meanPdf=new float[nbPara];
		estimPara=new float[nbPara];for (int i=0;i<nbPara;i++){estimPara[i]=0;}
		estimParaPrevious=new float[nbPara];for (int i=0;i<nbPara;i++){estimParaPrevious[i]=0;}
		for (int i=0;i<nbPara;i++){pdf[i]=new float[nbQuantifLevel];}
		for (int i=0;i<nbPara;i++){for (int j=0;j<nbQuantifLevel;j++){pdf[i][j]=0;}}
		maxCostValue=new float[nbPara];
		nbMTClasses=nb_MTClasses;
		nbSubGMTClasses=nb_SubGMTClasses;
		minSample=min_Sample;
		maxSample=max_Sample;
		nbTimeSamples=nb_TimeSamples;
		
	}
	/*! @brief Initialisation of the different pointers : to the tree structures, the costs collection and the a posteriori collection of sub-graphs*/
	void Init(SearchTree ****searchTreeCollec,float ****costss,vector<int>* kMTa,vector<int> *kMTb, vector<int> *firstSample){
		
		treeCollec=searchTreeCollec;
		costs=costss;
		kMTaSorted=*kMTa;
		kMTbSorted=*kMTb;
		firstSampleSorted=*firstSample;
	
	}
	/*! @brief Initialisation of the pointers  the a posteriori collection of sub-graphs*/
	void Init(vector<int>* kMTa,vector<int> *kMTb, vector<int> *firstSample){
		
		kMTaSorted=*kMTa;
		kMTbSorted=*kMTb;
		firstSampleSorted=*firstSample;
	
	}
	/*! @brief Calculation method of the maximum values of the matching costs */
	void MaxCostalues(){
		for (int kPara=0;kPara<nbPara;kPara++){
		   maxCostValue[kPara]=0;
		   for (int t=minSample;t<maxSample-nbTimeSamples+1;t++){
			for (int kMTa=0;kMTa<nbMTClasses;kMTa++){
				if (nbSubGMTClasses==2){for (int kMTb=0;kMTb<nbMTClasses;kMTb++){ 
				   if (!treeCollec[kMTa][kMTb][t][0].emptyTree){
					if (maxCostValue[kPara]<costs[kMTa][kMTb][t][kPara]&&costs[kMTa][kMTb][t][kPara]<InfCost){maxCostValue[kPara]=costs[kMTa][kMTb][t][kPara];}
				   }
				}}
				else{
				   if (!treeCollec[kMTa][kMTa][t][0].emptyTree){
					if (maxCostValue[kPara]<costs[kMTa][kMTa][t][kPara]&&costs[kMTa][kMTa][t][kPara]<InfCost){maxCostValue[kPara]=costs[kMTa][kMTa][t][kPara];}
				
				   }
				}
		   	}
		   }	
		}
	
	}
 	
	/*! @brief Method to perform a MMSE estimate of the parameter vector of the similarity model */
	void MMSEestim(int interestkMTa,int interestkMTb,int interestfirstSample){
		int aux=0;
		float indexMMSE=0;
		float auxSum=0; float cardinality=0;
		//float *sum=new float[nbPara];
		if (nbSubGMTClasses==1){interestkMTb=interestkMTa;}
		for (int kPara=0;kPara<nbPara;kPara++){
			estimParaPrevious[kPara]=estimPara[kPara];
			//sum[kPara]=0;
			//for (int i=0;i<nbQuantifLevel;i++){sum[kPara]+=pdf[kPara][i];}
			
			if (costs[interestkMTa][interestkMTb][interestfirstSample][kPara]>=maxCostValue[kPara]){
                if (maxCostValue[kPara]!=0){aux=nbQuantifLevel-1;}
                else{aux=0;}
            }
            //cout<<"aux0="<<aux<<"maxCostValue[kPara]="<<maxCostValue[kPara]<<endl;
            

			if (costs[interestkMTa][interestkMTb][interestfirstSample][kPara]<0){aux=0;}
			else{
 				if (costs[interestkMTa][interestkMTb][interestfirstSample][kPara]<maxCostValue[kPara]){
					aux=(int)((float)costs[interestkMTa][interestkMTb][interestfirstSample][kPara]*nbQuantifLevel/(float)maxCostValue[kPara]+0.5);
				}
			}
			//cout<<"aux="<<aux<<endl;
			//Dirichlet hyperparameters update
			if(costs[interestkMTa][interestkMTb][interestfirstSample][kPara]!=0){pdf[kPara][aux]+=1;}
			
			//MMSE estimation
			indexMMSE=-1;auxSum=0;cardinality=0;
			for (int i=0;i<nbQuantifLevel;i++){auxSum+=pdf[kPara][i]*i;cardinality+=pdf[kPara][i];}
			
            if (cardinality!=0){indexMMSE=auxSum/cardinality;}else{indexMMSE=0;}
			estimPara[kPara]=1.-(float)indexMMSE/(float)(nbQuantifLevel-1);
			meanPdf[kPara]=pdf[kPara][(int)(indexMMSE+0.5)];
			//cout<<" cost "<<costs[interestkMTa][interestkMTb][interestfirstSample][kPara]<<endl;
            //cout<<"kPara "<<kPara<<" indexMMSE "<<(int)(indexMMSE+0.5)<<" meanPdf[kPara] "<<meanPdf[kPara]<<" ,estimPara[kPara] "<<estimPara[kPara]<<endl;
		}
		auxSum=0;
		for (int kPara=0;kPara<nbPara;kPara++){
			auxSum+=estimPara[kPara];
		}
		for (int kPara=0;kPara<nbPara;kPara++){estimPara[kPara]/=auxSum;}
        if (nbSubGMTClasses==1){estimPara[5]=0;}
        if (nbTimeSamples==(maxSample-minSample)){estimPara[6]=0;}

        
	}
	/*! @brief Method to update of the similarity with a new parameter vector.
	*
	 *This methods calls the SearchTree object method 'LowestCosts(...)'which  updats the matching costs with a new parameter vector while using the already created tree structure. It returns the updated costs related to the contribution of the different graph's attribute for the updated best mapping function (path of minimal cost)*/	
	float ReevaluateMatch(int i,int j,int k){
		
	    float cost=-1;
	    if (!treeCollec[i][j][k][0].emptyTree){
	    	cost=treeCollec[i][j][k][0].LowestCosts(costs[i][j][k],estimPara);
	
	    }
	    return cost;
			
		
	}
	/*! @brief Method use to transfom into a flat distribution the learned multinomial distributions*/
	void ErasePdf(){
	
		for (int i=0;i<nbPara;i++){for (int j=0;j<nbQuantifLevel;j++){pdf[i][j]=0;}}
		for (int kPara=0;kPara<nbPara;kPara++){estimPara[kPara]=0;}
	}
	

	
};
#endif
