/*---------------------------------------------------------------------------*/
/*!
 * @file SearchTree.hh
 * @brief  file containing the SearchTree object which is  a tree  used for finding the minimal matching cost among all possible mapping functions.
 *
  
  @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/
#ifndef SEARCHTREE_H
#define SEARCHTREE_H

#include <iostream>
#include <fstream>
#include <string>  
#include <vector>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>

#include "TreeNode.hh"

#define Inf 1e20
using namespace std;
/**********************************************************/
/*******************SearchTree class***********************/
/**********************************************************/
 /*! @brief  SearchTree object which is  a tree  used by MatchCost to find the minimal matching cost among all possible mapping functions*/  class SearchTree 
{
 
  
   /***************ATTRIBUTES****************************/
  
  public: 
  	/*!@brief tree structures, with at each level a vector of objects called TreeNode*/	
	vector<TreeNode>* tree;
	/*!@brief number of time samples of the sub-graphs used in the matching function*/ 
	int nbTimeSamp;
	/*!@brief number of MT class considered per sub-graphs*/ 
	int nbMTClusters;
	/*!@brief array of number of nodes to match at the different time samples. In other words, number of level in the tree for each time sample.*/ 
	int *nbTimeLevels;
	/*!@brief Total number of nodes to match. In other words, total number of level in the tree.*/ 
	int nbTotalLevels;
	/*!@brief Index of the path with the minimal cost.*/ 
	int IndexminiCost;
	/*!@brief  minimal cost associated to the best path.*/ 
	float miniCost;
	/*!@brief  flag to precise if there a Tree structure was previously created.*/ 
	bool emptyTree;
	
  /***************METHODES****************************/
   	/*!@brief  constructor setting the attribut 'emptyTree' to true  */
	SearchTree(){
	
		emptyTree=true;
		
	}
	/*!@brief  destructor  clearing the tree structure  */
	virtual~SearchTree(){
		
		/*for (int i=0;i<nbTotalLevels;i++){
			tree[i].clear();
		}*/
		//delete [] tree;
		
	}
	/*!@brief  memory allocation, intialisation of the tree structure (vector of Object TreeNode)  */
	void Init(int nbTimes, int nbMT,int *nbNodes1){
	
		nbTimeSamp=nbTimes;
		nbMTClusters=nbMT;
		nbTimeLevels=new int[nbTimeSamp];
		nbTotalLevels=0;
		for (int time=0;time<nbTimeSamp;time++){
			nbTimeLevels[time]=nbNodes1[time];
			nbTotalLevels+=nbTimeLevels[time];
		}	
		tree=new vector<TreeNode>[nbTotalLevels];
		emptyTree=false;
		
	}
	/*!@brief Method used in the matching procedure to check that the node to match was not previously matched  */
	bool CheckUnusedNode(int time, int currentLevel,int preIndex, int nodek2,int MTk2){

		bool notUsedYet=true;
		int auxlevel=0;
		for (int i=0;i<time;i++){auxlevel+=nbTimeLevels[i];}
		while(currentLevel>auxlevel){
			currentLevel--;
			if (tree[currentLevel][preIndex].g2VertexIndex==nodek2&&tree[currentLevel][preIndex].g2MTIndex==MTk2){
				notUsedYet=false;
			}
			preIndex=tree[currentLevel][preIndex].branch;
		}
		return notUsedYet;

	}
	/*!@brief Method used in the matching procedure to check that the node to match is associated to the same MT class as the other previously matched nodes*/
	bool CheckMTCoherentNode(int time, int currentLevel,int preIndex,int MTk1,int MTk2){

		bool coherent=true;
		int auxlevel=0;
		for (int i=0;i<time-1;i++){auxlevel+=nbTimeLevels[i];}//checking coherence only in the current and past time sample
		while(currentLevel>auxlevel){
			currentLevel--;
			if (tree[currentLevel][preIndex].g1MTIndex==MTk1){
				if (tree[currentLevel][preIndex].g2MTIndex!=MTk2)
					coherent=false;
			}
			if (tree[currentLevel][preIndex].g1MTIndex!=MTk1){
				if (tree[currentLevel][preIndex].g2MTIndex==MTk2)
					coherent=false;
			}
			preIndex=tree[currentLevel][preIndex].branch;
		}
		return coherent;

	}
	/*!@brief Method used in the matching procedure to return the index of the node in the previous level of the tree  linked to the current node to match */
	int getPreTimeCorresNodeIndex(int time, int currentLevel,int preIndex,int nodek1, int MTk1){
	
		int nodek2=-1;
		int timeaux=time;
		while(timeaux>time-1){
			currentLevel--;
			timeaux=tree[currentLevel][preIndex].timeLevel;
			if (timeaux==(time-1)){break;}
			else{preIndex=tree[currentLevel][preIndex].branch;}
		}
		while(timeaux==(time-1)&&currentLevel>=0){
		   if (tree[currentLevel][preIndex].g1VertexIndex==nodek1&&tree[currentLevel][preIndex].g1MTIndex==MTk1){
		   	nodek2=tree[currentLevel][preIndex].g2VertexIndex;
			break;
		   }
		   else{
		   	
		        preIndex=tree[currentLevel][preIndex].branch;
			currentLevel--;
		   	timeaux=tree[currentLevel][preIndex].timeLevel;
		   }
		}
		//if (nodek2==-1){cout<<"linked to extra node"<<endl;}
		return nodek2;
				
	}
	/*!@brief Method used in the matching procedure to return the MT class index related to the node in the previous level of the tree  linked to the current node to match */
	int getPreTimeCorresMTIndex(int time, int currentLevel,int preIndex,int nodek1, int MTk1){
	
		int MTindex=-1;
		int timeaux=time;
		while(timeaux>time-1){
			currentLevel--;
			timeaux=tree[currentLevel][preIndex].timeLevel;
			if (timeaux==(time-1)){break;}
			else{preIndex=tree[currentLevel][preIndex].branch;}
		}
		while(timeaux==(time-1)&&currentLevel>=0){
		   if (tree[currentLevel][preIndex].g1VertexIndex==nodek1&&tree[currentLevel][preIndex].g1MTIndex==MTk1){
		   	 MTindex=tree[currentLevel][preIndex].g2MTIndex;
			 break;
		   }
		   else{
		        preIndex=tree[currentLevel][preIndex].branch;
			currentLevel--;
		   	timeaux=tree[currentLevel][preIndex].timeLevel;
		   }
		}
		
		if (MTindex==-1){cout<<"Error occured in SearchTree.getPreTimeCorresMTIndex(...)"<<endl;}
		return MTindex;
		
	}
	/*!@brief Method used in the matching procedure to add a Node in the tree  */
	void AddNode(int time, int currentLevel, int preIndex,int nodek1,int MTk1,int nodek2,int MTk2,float cost,float weightMTinTLDiffCost,float TL1TL2DivCost,float flowDiffCost, float divDiffCost, float MIDiffCost, float divMTaMTbCost, float timeCost){
		
		TreeNode auxtreeNode;
		auxtreeNode.timeLevel=time;
		auxtreeNode.g2VertexIndex=nodek2;
		auxtreeNode.g2MTIndex=MTk2;
		auxtreeNode.g1VertexIndex=nodek1;
		auxtreeNode.g1MTIndex=MTk1;
		auxtreeNode.branch=preIndex;						
		auxtreeNode.nodeCost=cost;
		auxtreeNode.weightMTinTLDiffCost=weightMTinTLDiffCost;
		auxtreeNode.TL1TL2DivCost=TL1TL2DivCost;
		auxtreeNode.flowDiffCost=flowDiffCost;
		auxtreeNode.divDiffCost=divDiffCost;
		auxtreeNode.MIDiffCost=MIDiffCost;
		auxtreeNode.divMTaMTbCost=divMTaMTbCost;
		auxtreeNode.timeCost=timeCost;
		
		tree[currentLevel].push_back(auxtreeNode);
		
	}
	/*!@brief Method used in the matching procedure to return the costs related to the contribution of the different graph's attribute for the best mapping function (path of minimal cost) */
	float LowestCosts(float * costs){
	
		miniCost=Inf;
		IndexminiCost=-1;
		float currentCost;
		float *currentCosts= new float[7];
		int currentLevel;
		int preIndex;
		if (nbTotalLevels-1<=0){cout<<"empty graph ..."<<endl;}
		//cout<<"number of leaves in the SearchTree : "<<tree[nbTotalLevels-1].size()<<endl;
		for (int k=0;(unsigned int)k<tree[nbTotalLevels-1].size();k++){
			currentCost=0;for(int kCost=0;kCost<7;kCost++){currentCosts[kCost]=0;}
			currentCost+=tree[nbTotalLevels-1][k].nodeCost;
			currentCosts[0]+=tree[nbTotalLevels-1][k].weightMTinTLDiffCost;
			currentCosts[1]+=tree[nbTotalLevels-1][k].TL1TL2DivCost;
			currentCosts[2]+=tree[nbTotalLevels-1][k].flowDiffCost;
			currentCosts[3]+=tree[nbTotalLevels-1][k].divDiffCost;
			currentCosts[4]+=tree[nbTotalLevels-1][k].MIDiffCost;
			currentCosts[5]+=tree[nbTotalLevels-1][k].divMTaMTbCost;
			currentCosts[6]+=tree[nbTotalLevels-1][k].timeCost;
			currentLevel=nbTotalLevels-1;
			preIndex=k;	
			while(currentLevel>0){
				preIndex=tree[currentLevel][preIndex].branch;
				currentLevel--;
				currentCost+=tree[currentLevel][preIndex].nodeCost;
				currentCosts[0]+=tree[currentLevel][preIndex].weightMTinTLDiffCost;
				currentCosts[1]+=tree[currentLevel][preIndex].TL1TL2DivCost;
				currentCosts[2]+=tree[currentLevel][preIndex].flowDiffCost;
				currentCosts[3]+=tree[currentLevel][preIndex].divDiffCost;
				currentCosts[4]+=tree[currentLevel][preIndex].MIDiffCost;
				currentCosts[5]+=tree[currentLevel][preIndex].divMTaMTbCost;
				currentCosts[6]+=tree[currentLevel][preIndex].timeCost;
			}
			if (currentCost<miniCost){
				miniCost=currentCost;
				for(int kCost=0;kCost<7;kCost++){costs[kCost]=currentCosts[kCost];}
				costs[7]=miniCost;
				IndexminiCost=k;
			}
		}
		
		
	
		return miniCost;
	}
	/*!@brief Method used when updating the matching costs with a new parameter vector while using the already created tree structure. It returns the updated costs related to the contribution of the different graph's attribute for the updated best mapping function (path of minimal cost) */
	float LowestCosts(float * costs, float* para){
	
		miniCost=Inf;
		IndexminiCost=-1;
		float currentCost;
		float *currentCosts= new float[7];
		int currentLevel;
		int preIndex;
		if (nbTotalLevels-1<=0){cout<<"empty graph ..."<<endl;}
		for (int k=0;(unsigned int)k<tree[nbTotalLevels-1].size();k++){
			currentCost=0;for(int kCost=0;kCost<7;kCost++){currentCosts[kCost]=0;}
			tree[nbTotalLevels-1][k].nodeCost=tree[nbTotalLevels-1][k].weightMTinTLDiffCost*para[0];
			tree[nbTotalLevels-1][k].nodeCost+=tree[nbTotalLevels-1][k].TL1TL2DivCost*para[1];
			tree[nbTotalLevels-1][k].nodeCost+=tree[nbTotalLevels-1][k].flowDiffCost*para[2];
			tree[nbTotalLevels-1][k].nodeCost+=tree[nbTotalLevels-1][k].divDiffCost*para[3];
			tree[nbTotalLevels-1][k].nodeCost+=tree[nbTotalLevels-1][k].MIDiffCost*para[4];
			tree[nbTotalLevels-1][k].nodeCost+=tree[nbTotalLevels-1][k].divMTaMTbCost*para[5];
			tree[nbTotalLevels-1][k].nodeCost+=tree[nbTotalLevels-1][k].timeCost*para[6];
			currentCost+=tree[nbTotalLevels-1][k].nodeCost;
			currentCosts[0]+=tree[nbTotalLevels-1][k].weightMTinTLDiffCost;
			currentCosts[1]+=tree[nbTotalLevels-1][k].TL1TL2DivCost;
			currentCosts[2]+=tree[nbTotalLevels-1][k].flowDiffCost;
			currentCosts[3]+=tree[nbTotalLevels-1][k].divDiffCost;
			currentCosts[4]+=tree[nbTotalLevels-1][k].MIDiffCost;
			currentCosts[5]+=tree[nbTotalLevels-1][k].divMTaMTbCost;
			currentCosts[6]+=tree[nbTotalLevels-1][k].timeCost;
			currentLevel=nbTotalLevels-1;
			preIndex=k;	
			while(currentLevel>0){
				preIndex=tree[currentLevel][preIndex].branch;
				currentLevel--;
				tree[currentLevel][preIndex].nodeCost=tree[currentLevel][preIndex].weightMTinTLDiffCost*para[0];
				tree[currentLevel][preIndex].nodeCost+=tree[currentLevel][preIndex].TL1TL2DivCost*para[1];
				tree[currentLevel][preIndex].nodeCost+=tree[currentLevel][preIndex].flowDiffCost*para[2];
				tree[currentLevel][preIndex].nodeCost+=tree[currentLevel][preIndex].divDiffCost*para[3];
				tree[currentLevel][preIndex].nodeCost+=tree[currentLevel][preIndex].MIDiffCost*para[4];
				tree[currentLevel][preIndex].nodeCost+=tree[currentLevel][preIndex].divMTaMTbCost*para[5];
				tree[currentLevel][preIndex].nodeCost+=tree[currentLevel][preIndex].timeCost*para[6];				
				currentCost+=tree[currentLevel][preIndex].nodeCost;
				currentCosts[0]+=tree[currentLevel][preIndex].weightMTinTLDiffCost;
				currentCosts[1]+=tree[currentLevel][preIndex].TL1TL2DivCost;
				currentCosts[2]+=tree[currentLevel][preIndex].flowDiffCost;
				currentCosts[3]+=tree[currentLevel][preIndex].divDiffCost;
				currentCosts[4]+=tree[currentLevel][preIndex].MIDiffCost;
				currentCosts[5]+=tree[currentLevel][preIndex].divMTaMTbCost;
				currentCosts[6]+=tree[currentLevel][preIndex].timeCost;
			}
			if (currentCost<miniCost){
				miniCost=currentCost;
				for(int kCost=0;kCost<7;kCost++){costs[kCost]=currentCosts[kCost];}
				costs[7]=miniCost;
				IndexminiCost=k;
			}
		}
		
		
	
		return miniCost;
	}
	/*!@brief Method used during the matching procedure to select only a number of  'nbMiniBranches' best paths in the search tree.  */
	void PruneTree(int nbMiniBranches, int currentLevel){
	
	  	
	   int nb_branches=(int)tree[currentLevel].size();
	   if (nb_branches>nbMiniBranches){ 
	   
		//calculating branches cost
		int pruneLevel=currentLevel;
		int preIndex;
		float *brancheCost=new float[nb_branches];
		float *sortedBrancheCost=new float[nb_branches];
		float *prunedBranches=new float[nb_branches-nbMiniBranches];
		float *sortedPrunedBranches=new float[nb_branches-nbMiniBranches];
		for (int k=0;k<nb_branches;k++){
			brancheCost[k]=0;
			sortedBrancheCost[k]=-1;
			currentLevel=pruneLevel;
			brancheCost[k]+=tree[pruneLevel][k].nodeCost;
			preIndex=k;
			while(currentLevel>0){
				preIndex=tree[currentLevel][preIndex].branch;
				currentLevel--;
				brancheCost[k]+=tree[currentLevel][preIndex].nodeCost;
			}
		}
		//sorting
		float miniCost;
		bool aux=true;
		for (int i=0;i<nb_branches;i++){
		  miniCost=Inf;
		  for (int k=0;k<nb_branches;k++){     
		     if (brancheCost[k]<miniCost){
		     	aux=true;
		     	for (int l=0;l<nb_branches;l++){if (sortedBrancheCost[l]==k){aux=false;}}
			if (aux){
				sortedBrancheCost[i]=k;
		     		miniCost=brancheCost[k];
			}
		     }
		   }
		}
		//sorting the branches to be pruned in index order (for the vector.erase() procedure in the pruning)  
		for (int i=nbMiniBranches;i<nb_branches;i++){
			prunedBranches[i-nbMiniBranches]=sortedBrancheCost[i];
			sortedPrunedBranches[i-nbMiniBranches]=-1;
		}
		float maxiIndex;
		for (int i=0;i<nb_branches-nbMiniBranches;i++){
		    maxiIndex=-Inf;
		    for (int k=0;k<nb_branches-nbMiniBranches;k++){
		      if (prunedBranches[k]>maxiIndex){ 
		    	 aux=true;
		    	 for (int l=0;l<nb_branches-nbMiniBranches;l++){if (sortedPrunedBranches[l]==prunedBranches[k]){aux=false;}}
		    	 if (aux){
		    	    sortedPrunedBranches[i]=prunedBranches[k];
		    	    maxiIndex=prunedBranches[k];
		    	 }
		      }
		    }
		}
		
		//pruning
		vector<TreeNode>::iterator currentIterator;
		for (int i=0;i<nb_branches-nbMiniBranches;i++){
			currentIterator = tree[pruneLevel].begin();
			int k=0;
			while(k<(int)sortedPrunedBranches[i]){currentIterator++;k++;}
			tree[pruneLevel].erase( currentIterator );
		}
				
		delete [] brancheCost;
		delete [] sortedBrancheCost;
		delete [] prunedBranches;
		delete [] sortedPrunedBranches;
		
	  }	
	}
	
	
};

#endif





