/*---------------------------------------------------------------------------*/
/*!
 * @file /SubGraph.hh
 * @brief  File containing the SubGraph object used to load a part of the global graph.
 
  @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/
#include <iostream>
#include <fstream>
#include <string>  
#include <vector>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
 /*!@brief   object used by GraphManipulator to load a part of the global graph.
 *
 *Note that a maximum of 2 MT classes are considered per sub-graph. */
class SubGraph 
{
  public:
  
   /***************ATTRIBUTES****************************/	
   
  // subGraph general attributes:
   	/*!@brief  number of samples of the time-window*/   
	int nbTimeSamp;
	/*!@brief time delay between the different time samples specific to the graph 	*/ 
	float *tDelay;
	/*!@brief dimensionnality of the TL feature space*/
	int nbDim;
	/*!@brief  number of MT classes contained in the the  sub-graph ( Maximum of 2)*/ 
	int nbMTClusters;
	
  //Vertex attributes:
  
	//Projected MT cluster attributes
	/*!@brief Projected MT cluster attributes in successive TL spaces: pointer to a vector of mean array*/  
	vector<float*> *meanMT;
	/*!@brief Projected MT cluster attributes in successive TL spaces: pointer to a vector of covariance matrices*/ 
	vector<float**> *covMT;
	/*!@brief Projected MT cluster attributes in successive TL spaces: pointer to a vector of cluster's weights*/ 
	vector<unsigned long int> weightMT;
	
	//TL associated cluster attributes
	/*!@brief node (TL associated cluster) attributes: index of order of introduction of the nodes */
	vector<int> *orderAddedNode;
	/*!@brief node (TL associated cluster) attributes: pointer to a vector of mean array */
	vector<float*> *meanTL;
	/*!@brief node (TL associated cluster) attributes: pointer to a vector of covariance matrices*/
	vector<float**> *covTL;
	/*!@brief node (TL associated cluster) attributes: pointer to a vector of MT  index associated to the graph nodes*/
	vector<vector<int> > *indexMTinTL;
	/*!@brief node (TL associated cluster) attributes: pointer to a vector of cluster's weights*/
	vector<vector<unsigned long int> > *weightMTinTL;
	/*!@brief node (TL associated cluster) attributes: pointer to a vector of association probabilities (Kullback-Leibler divergence mesurements devided by the number of pixels in the intersection of TL a MT spatial classes)*/
	vector<vector<float> > *divTL_MT;
	
	
  //Edges attributes
  	
	/*!@brief Edges attributes : Flow adjacent matrix*/
	vector<vector<vector<unsigned long int> > > *flowMatrix;
	
	/*!@brief Edges attributes : Flow Divergence adjacent matrix*/
	vector<vector<vector<float> > > *divMatrix;
	
	/*!@brief Edges attributes :  Mutual information adjacent matrix*/
	vector<vector<vector<float> > > *MIMatrix;
	
  /***************METHODES****************************/
  	/*!@brief constructor : allocation of memory*/
    	SubGraph(int nb_dim, int nb_tSamp, float *t_delay){
		
		nbDim=nb_dim;
		nbMTClusters=0;
		nbTimeSamp=nb_tSamp; 
		tDelay=new float[nbTimeSamp-1];
		for (int i=0;i<nbTimeSamp-1;i++){tDelay[i]=t_delay[i];}
		meanMT=new vector<float*>[nbTimeSamp];
		covMT=new vector<float**>[nbTimeSamp];
		//weightMT=new vector<unsigned long int>[nbTimeSamp];
		meanTL=new vector<float*>[nbTimeSamp];
		covTL=new vector<float**>[nbTimeSamp];
		weightMTinTL=new vector<vector<unsigned long int> >[nbTimeSamp];
		indexMTinTL=new vector<vector<int> >[nbTimeSamp];
		divTL_MT=new vector<vector<float> >[nbTimeSamp];
		flowMatrix=new vector<vector<vector<unsigned long int> > >[nbTimeSamp];
		divMatrix=new vector<vector<vector<float> > >[nbTimeSamp];
		MIMatrix=new vector<vector<vector<float> > >[nbTimeSamp];
		orderAddedNode=new vector<int>[nbTimeSamp];
		
	}
	/*!@brief destructor : freeing memory*/
	virtual~SubGraph(){
	/*
		delete [] tDelay;
		for (int t=0;t<nbTimeSamp;t++){
			meanMT[t].clear();
			covMT[t].clear();
			//weightMT[t].clear();
			meanTL[t].clear();
			covTL[t].clear();
			for (int k=0;(unsigned int)k<indexMTinTL[t].size();k++){
				indexMTinTL[t][k].clear();
				weightMTinTL[t][k].clear();
				divTL_MT[t][k].clear();
			}
			indexMTinTL[t].clear();
			weightMTinTL[t].clear();
			divTL_MT[t].clear();
		}
		weightMT.clear();
		delete [] meanTL;
		delete [] covTL;
		delete [] indexMTinTL;
		delete [] weightMTinTL;
		delete [] divTL_MT;
		delete [] meanMT;
		delete [] covMT;
		delete [] orderAddedNode;
		//delete [] weightMT;
	*/	
	}
	/*!@brief Method to load in a subGraph object,  attributes relative to a MT cluster evolution in the defined time-window.
	*
	*This method is the first one to be called when a MT cluster trajectory is added.*/
	void AddAttMT(float **mean, float ***cov,unsigned long int weight){
		
		//adding the attributes of a projected MT Cluster
		for (int t=0;t<nbTimeSamp;t++){
			float *auxmean=new float[nbDim];
			float **auxcov=new float*[nbDim];
			
			for (int j1=0;j1<nbDim;j1++){
				auxmean[j1]=mean[t][j1];
				auxcov[j1]=new float[nbDim];
				for (int j2=0;j2<nbDim;j2++){
					auxcov[j1][j2]=cov[t][j1][j2];
				}
			}
			
			meanMT[t].push_back(auxmean);
			covMT[t].push_back(auxcov);
		}
		
		unsigned long int *auxweight=new unsigned long int[1];
		auxweight[0]=weight;
		weightMT.push_back(auxweight[0]);
		
		nbMTClusters++;
		//deleting MT eigen related  parameters
		for (int i=0;i<nbTimeSamp;i++){orderAddedNode[i].clear();}
		
	}
	/*!@brief Method to load in a subGraph object,  attributes relative to the nodes of the graph in the defined time-window.
	*
	*This method is the second one to be called when a MT cluster trajectory is added. */
	void AddAttTLforMT(int time, int nbClusters, float **mean, float ***cov,unsigned long int *weight, float *div){
		
		int addNewCluster, check,sameClusterIndex=-1;	
		for (int k=0;k<nbClusters;k++){
		   //checking if it is a new vertex
		   addNewCluster=1;
		   check=0;
		   for (int l=0;(unsigned int)l<meanTL[time].size();l++){
		   	check=0;
		   	for (int j=0;j<nbDim;j++){
				if (meanTL[time][l][j]==mean[k][j]&&covTL[time][l][j][j]==cov[k][j][j]){
					check++;
				}
		   	}
			if(check==nbDim){
				addNewCluster=0;sameClusterIndex=l;
			}
		   }	
		   // adding the new vertex
		   if (addNewCluster==1){
			float *auxmean=new float[nbDim];
			float **auxcov=new float*[nbDim];		
			for (int j1=0;j1<nbDim;j1++){
				auxmean[j1]=mean[k][j1];
				auxcov[j1]=new float[nbDim];
				for (int j2=0;j2<nbDim;j2++){
					auxcov[j1][j2]=cov[k][j1][j2];
				}
			}
			meanTL[time].push_back(auxmean);
			covTL[time].push_back(auxcov);
			vector<int> auxindexvector;
			int *auxindex=new int[1];
			auxindex[0]=nbMTClusters-1;
			auxindexvector.push_back(auxindex[0]);
			indexMTinTL[time].push_back(auxindexvector);
			vector<unsigned long int> auxweightvector;
			unsigned long int *auxweight=new unsigned long int[1];
			auxweight[0]=weight[k];
			auxweightvector.push_back(auxweight[0]);
			weightMTinTL[time].push_back(auxweightvector);
			vector<float>  auxdivvector;
		   	float  *auxdiv=new float[1];
			auxdiv[0]=div[k];
		   	auxdivvector.push_back(auxdiv[0]);
			divTL_MT[time].push_back(auxdivvector);
			
		   }
		   else{
		   	//filling the identical vertex with the MT related attributes	
		   	int *auxindex=new int[1];auxindex[0]=nbMTClusters-1;
			indexMTinTL[time][sameClusterIndex].push_back(auxindex[0]);
			unsigned long int *auxweight=new unsigned long int[1];auxweight[0]=weight[k];
			weightMTinTL[time][sameClusterIndex].push_back(auxweight[0]);
			float  *auxdiv=new float[1];auxdiv[0]=div[k];
			divTL_MT[time][sameClusterIndex].push_back(auxdiv[0]);
		   }
		
		}
		
		for (int k=0;k<nbClusters;k++){
		   //checking if it is a new vertex
		   check=0;
		   for (int l=0;(unsigned int)l<meanTL[time].size();l++){
		   	check=0;
		   	for (int j=0;j<nbDim;j++){
				if (meanTL[time][l][j]==mean[k][j]&&covTL[time][l][j][j]==cov[k][j][j]){
					check++;
				}
		   	}
			if(check==nbDim){
				addNewCluster=0;sameClusterIndex=l;
				int *order =new int[1];order[0]=l;
				orderAddedNode[time].push_back(order[0]);
			}
		   }
		 }
		 
	}
	/*!@brief Method to load in a subGraph object,  attributes relative to the edges of the graph in the defined time-window.
	*
	*This method is the last and third one to be called when a MT cluster trajectory is added. */		
	void AddEdgesAttMT(int time,  int nbClusters1,int nbClusters2,unsigned long int **flow, float **div, float **MI ){
	//adding edges (flow, Div & MI) related to the trajectories of a MT between time and time-1
			
		//updating size of adajacent matrices (creating new matrices, copying in it the values of old matrices which are deleted)
		int nbNewNodes1;
		int nbNewNodes2;
		if (flowMatrix[time].size()==0){
		  nbNewNodes1=meanTL[time-1].size();
		  nbNewNodes2=meanTL[time].size();
		}
		else{
		  nbNewNodes1=meanTL[time-1].size()-flowMatrix[time][0].size();
		  nbNewNodes2=meanTL[time].size()-flowMatrix[time][0][0].size();
		} 
		if (nbNewNodes1!=0||nbNewNodes2!=0){
		    vector<vector<unsigned long int> > *auxauxflowvector=new vector<vector<unsigned long int> >[nbMTClusters];
		    vector<vector<float> > *auxauxdivvector=new vector<vector<float> >[nbMTClusters];
		    vector<vector<float> > *auxauxMIvector= new vector<vector<float> >[nbMTClusters];
		   for (int k=0;k<nbMTClusters;k++){	     
		     for (int l1=0;(unsigned int)l1<meanTL[time-1].size();l1++){
			vector<unsigned long int> *auxflowvector=new vector<unsigned long int>[1];
			vector<float> *auxdivvector=new vector<float>[1];
			vector<float> *auxMIvector=new vector<float>[1];
			for (int l2=0;(unsigned int)l2<meanTL[time].size();l2++){
				unsigned long int *auxflow=new unsigned long[1] ;
				float *auxdiv=new float[1];
				float *auxMI=new float[1];
				if (flowMatrix[time].size()!=0){
				  if ((unsigned int)l1<flowMatrix[time][0].size() && (unsigned int)l2<flowMatrix[time][0][0].size() && k!=nbMTClusters-1){
					auxflow[0]=flowMatrix[time][k][l1][l2];
					auxdiv[0]=divMatrix[time][k][l1][l2];
					auxMI[0]=MIMatrix[time][k][l1][l2];	
				  }
				  else{
					auxflow[0]=0;
					auxdiv[0]=-1;
					auxMI[0]=-1;
				  }
				}
				else{
				   auxflow[0]=0;
				   auxdiv[0]=-1;
				   auxMI[0]=-1;
				}
				auxflowvector[0].push_back(auxflow[0]);
				auxdivvector[0].push_back(auxdiv[0]);
				auxMIvector[0].push_back(auxMI[0]);
			}
			auxauxflowvector[k].push_back(auxflowvector[0]);
			auxauxdivvector[k].push_back(auxdivvector[0]);
			auxauxMIvector[k].push_back(auxMIvector[0]);
		     }
		   }
		   flowMatrix[time].clear();
		   divMatrix[time].clear();
		   MIMatrix[time].clear();
		   for (int k=0;k<nbMTClusters;k++){
		     flowMatrix[time].push_back(auxauxflowvector[k]);
		     divMatrix[time].push_back(auxauxdivvector[k]);
		     MIMatrix[time].push_back(auxauxMIvector[k]);
		   }
		}
		else{
		     vector<vector<unsigned long int> > auxauxflowvector;
		     vector<vector<float> > auxauxdivvector;
		     vector<vector<float> > auxauxMIvector;
		     for (int l1=0;(unsigned int)l1<meanTL[time-1].size();l1++){
			 vector<unsigned long int> auxflowvector;
			 vector<float> auxdivvector;
			 vector<float> auxMIvector;
			 for (int l2=0;(unsigned int)l2<meanTL[time].size();l2++){
				unsigned long int *auxflow=new unsigned long int[1];
				auxflow[0]=0;
				auxflowvector.push_back(auxflow[0]);
				float *auxdiv=new float[1];
				auxdiv[0]=-1;
				auxdivvector.push_back(auxdiv[0]);
				float *auxMI=new float[1];
				auxMI[0]=-1;
				auxMIvector.push_back(auxMI[0]);
			 }
			 auxauxflowvector.push_back(auxflowvector);
			 auxauxdivvector.push_back(auxdivvector);
			 auxauxMIvector.push_back(auxMIvector);
		     }
		     flowMatrix[time].push_back(auxauxflowvector);
		     divMatrix[time].push_back(auxauxdivvector);
		     MIMatrix[time].push_back(auxauxMIvector);
		}
		//adding the new values in the adjacent matrices
		for (int v1=0;v1<nbClusters1;v1++){
			for (int v2=0;v2<nbClusters2;v2++){
				flowMatrix[time][nbMTClusters-1][orderAddedNode[time-1][v1]][orderAddedNode[time][v2]]=flow[v1][v2];
				divMatrix[time][nbMTClusters-1][orderAddedNode[time-1][v1]][orderAddedNode[time][v2]]=div[v1][v2];
				MIMatrix[time][nbMTClusters-1][orderAddedNode[time-1][v1]][orderAddedNode[time][v2]]=MI[v1][v2];
			}
		}
		
	
	}
		
};
