/*---------------------------------------------------------------------------*/
/*!
 * @file SupLearnPara.cpp
 * @brief  Program for learning interactively user-specific semantics attached to sub-graphs.
 
  @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/

#include <iostream>
#include <string>  
#include <vector>
#include <fstream>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <ctime>

#include "graphCharac+.hh"


using namespace std;
 

 
/*! 	@brief  Usage function 	*/ 
void usage(std::string argv0) {
    std::cout <<"\nAuthor : Patrick HEAS, Date : 2005\n";
    std::cout << "\nusage: " << ">"
              << argv0 <<std::endl; 
    std::cout << "       Graph_FileName TimeDelays_FileName "<<std::endl; 
    std::cout << "\n*****************************************************************************************************************************************\n";
  }

/*!	@brief Main procedure for interactive learning.
*
*The program follows the following sequential organisation:\n
 * \li 1) This progamm creats first a  GraphManipulator object. A method "ReadGraphVersion3(...)" is then called to read a graph at format 2 (c.f. Documentation on the "Inference of a graph of dynamic clusters trajectories") stored on the disk. Note that this object posses  similarties with the GraphDynaCluster object used previously  for infering the trajectories, with the difference that  the GraphManipulator object does not possess any  methods and attributes related to the  graph inference; It possesses  attributes and methods to read a already created graph, to manipulate sub-graphs, to create sorted collections of sub-graphs with supervised learning methods, etc....\n 
 * \li 2) Then the interactive learning procedure is initializated : the users defines a time window, a part of interest in the graph and the complexity of the graph (via the divergence treshold). Then he defines the spatio-temporal location of a sub-graph of interest. Optionally he fixes the a priori  model similarity parameters and can also set some specific parameters for the search.\n 
* \li 3) Therfore, the matching procedure by the parametric similarity model is performed between all the possible sub-graphs included in the graph part defined by the user (with the input parameters :'minSample' and 'maxSample') and a reference sub-graph defined by the first user example and with a parameter vector filled with ones (or the prior parameters provided by the user are used). A sorted collection of sub-graphs is created. The ordering of the sub-graphs is done according to their likelihood for the user-specific positive semantic. Note that, in the case that  a couple of 2 MT class are considered per sub-graph, each element of the pair is first matched idependently with all the possible sub-graphs. Then unlikely MT class associations  are removed from the collections of sub-graph couple to match.\n
@todo GENERAL REMARK: We remark here that, to reduce the computational complexity and the clarty of  many methods called by this program, it would have been wiser to consider 2 MT class (or more) in a single graph and not, as it is done here, in 2 different graphs which are associated. TO BE MODIFIED IN PRIORITY
*
* 
\n
\n **************************************Beginning of the learning loop**************************************\n
\n
* \li 4) In order to derive a sub-graph collection sorted according to negative semantic likelihood probabilities, a matching procedure is performed with the most unlikely sub-graph of the positive semantic collection (for the first iteration only) or with the last user negative example. The current estimated parameter vector is used (for the first iteration, we use a parameter vector  filled with ones)\n 
*\li 5) Then, positive and negative semantic learning are initialisated and performed with the user last examples. It results in the iference of a collection of sub-graphs sorted according to posterior probabilities. In parallel, the relevance feedback measurement are derived. (until the user has not provided a negative example, the negative learning is on stand by. Thus, in this case, the collection is not sorted according to posterior probabilities but only according to the likelihood probabilities relative to the positive semantic) 
*\li 6) Then the program enters another loop where new positive or negative examples are provided by the user. To update the sub-graph sorted collections (i.e. the sub-graphs likelihood probabilities related to the positive and negative semantic),  the reference graphs (positive and negative) are preserved in this loop, and the matching procedure is performed with the same search trees created previously. Indeed, the matching procedure just finds new solution-paths in the search trees, according to the current estimated similarity model parameters.  The 2 sorted collections of sub-graphs are then used to update the  posterior sub-graph collection. In parallel, the relevance feedback measurement are derived.\n
Note that, for each iteration of this loop, the user can decide to erase the memory of the positive or/and negative semantic learning. He can also decide to update the reference subgraphs and thus to exit the loop.
*\li 7) When the previous loop is exited, the user provides a new positive reference sub-graph. Here the user can decide to concentrate on only a percentage of the collection of sub-graphs.   A matching procedure is then performed with this new subgraph and the current estimated parameters of the similarity model.\n
The program goes then to the beginning of the learning loop (step 4)\n
\n
\n **************************************End of the learning loop**************************************\n
\n
*Note that this program creats text files called "./APostList" containing the relevance feedback measurements and the posterior collection of sub-graphs after each new example provided by the user. This files can then be read by the GUI applet to be visualized.

	@param argv0[1]  Graph file name at format 2  (c.f. Documentation on the "Inference of a graph of dynamic clusters trajectories") 
*    	@param argv0[2]  Text file name containing, on a single line and separated by a white space, the time delay between the different time samples specific to the graph.
@todo Note that the time delay information is not included in the graph at format 2 directly, but is included in the standard format. Thus, this input has to be removed and a  procedure for reading the graph at the standard format has to replace the actual one.  TO BE MODIFIED!!	*/
int main(int argc, char *argv [])
{   
 
	std::string ifn1;
	std::string ifn2;
	switch (argc-1) { 
		case 2:  ifn2	 		= std::string(argv[2]);
	        case 1:  ifn1	 		= std::string(argv[1]); break;
        	case 0:
	        default: usage(argv[0]); return -1;
    	}
    	if ((argc-1) < 2) { usage(argv[0]); return -1;}
	std::cout << "> " << argv[0] << " " << ifn1 << " " << ifn2 << std::endl;
	     
	srand((unsigned)time(NULL));
	time_t tdeb, t1,t2;
	time(&tdeb);
	t1=tdeb;
     	
/*****************Reading the Graph*****************************************************************/

	GraphManipulator Graphtool;
	Graphtool.ReadGraphVersion3((char *)ifn1.c_str(),(char *)ifn2.c_str());
	

/*****************GraphMatching with the Parametric Similarty Model**********************************/	
	std::cout<<"GraphMatching with the Parametric Similarty Model..."<<endl;	
	
	int nb_samples=2;int minSample=0;int maxSample=1;float seuilDiv=5; // part of the graph to search in 
	int firstSample1=0;int kMT1a=0;int kMT1b=-1; // spatio-temporal location
	float weightMTinTLDiff=1;float TL1TL2Div=1;float flowDiff=1;float divDiff=1;float MIDiff=1;float divMTaMTbCost=1;float timeDiff=1;// a priori
	string userAPriori="n";string userParaSearch="n"; //search parameters
	bool meanCalcul=true;int aux=1;
	int nbMiniBranches=30;
	int RefSGTime=0;int RefSGkMT1a=0;int RefSGkMT1b=0;
	

	//User defining a time window,a part of interest in the graph, and the complexity of the graph. 
	cout<<"number of samples of the time-window?"<<endl;
	cin >> nb_samples;	
	cout<<"Minimum sample time of the graph for searching other sub-graphs?"<<endl;cin >> minSample;minSample--;	
	cout<<"Maximum sample time of the graph for searching other sub-graphs?"<<endl;cin >> maxSample;maxSample--;
	cout<<"Divergence treshold?"<<endl;
	cin >> seuilDiv;
	//Spatio-temporal location of sub-graph of interest 			
	cout<<"First sample of the sub-graph of interest?"<<endl;
	cin >> firstSample1; firstSample1--;RefSGTime=	firstSample1;						
	cout<<"Index of the MT Class of interest?"<<endl;
	cin >> kMT1a;kMT1a--;RefSGkMT1a=kMT1a	;		
	cout<<"If you want to consider 2 MT Class per sub-graph, enter the index of the 2nd MT class, otherwise respond -1"<<endl;
	cin >> kMT1b;if (kMT1b!=-1){ kMT1b--;}	RefSGkMT1b=kMT1b;
	//User fixing the a priori parameters of the parametric model similiarity  (OPTIONAL)
	cout<<"Do you want to fix  a priori parameter for the similarity model ('y'/'n')?"<<endl;cin >> userAPriori;
	if (userAPriori=="y"){
		cout<<"a priori parameter (in [0,1]) weighting the population of each node?"<<endl;
		cin >> weightMTinTLDiff;				
		cout<<"a priori parameter (in [0,1]) weighting the nodes' distribution?"<<endl;
		cin >> TL1TL2Div;					
		cout<<"a priori parameter (in [0,1]) weighting the flow of pixels exchanged between nodes?"<<endl;
		cin >> flowDiff;					
		cout<<"a priori parameter (in [0,1]) weighting the change in time of the nodes' distribution?"<<endl;
		cin >> divDiff;						
		cout<<"a priori parameter (in [0,1]) weighting the intra-node change in time?"<<endl;
		cin >> MIDiff; 							
		cout<<"a priori parameter (in [0,1]) weighting the interaction of the 2 MT classes (only if 2 MT class are considered)?"<<endl;
		cin >> divMTaMTbCost;					
		cout<<"a priori parameter (in [0,1]) weighting the time lattice on which the sub-graph is defined?"<<endl;
		cin >> timeDiff;
        if(weightMTinTLDiff==0)weightMTinTLDiff+=1e-5;
        if(TL1TL2Div==0)TL1TL2Div+=1e-5;
        if(flowDiff==0)flowDiff+=1e-5;
        if(divDiff==0)divDiff+=1e-5;
        if(MIDiff==0)MIDiff+=1e-5;
        if(divDiff==0)divDiff+=1e-5;
        if(MIDiff==0)MIDiff+=1e-5;
	}
	//User fixing parameters for the search (OPTIONAL)	
	cout<<"Do you want to fix some specific parameters for the search ('y'/'n')?"<<endl;cin >> userParaSearch;
	if (userParaSearch=="y"){
		cout<<"mean calculation / Kullback-Leibler divergence calculation  (1=mean/0=divergence)?"<<endl;
		cin >> aux;if (aux==0){meanCalcul=false;}		
		cout<<"number of branches considered in the search tree?"<<endl;cin >> nbMiniBranches;
	}	
		
	

	
	
	//Finding the most likely MT class associations for the matching when 2MT classes are considered
	if (kMT1b!=-1){ 
		Graphtool.InitGraphforMatch(nb_samples,firstSample1,seuilDiv,kMT1a,-1);
		Graphtool.CreateMatchCostObject();
		Graphtool.FixWeights(weightMTinTLDiff,TL1TL2Div,flowDiff,divDiff,MIDiff,divMTaMTbCost,timeDiff,meanCalcul);
		Graphtool.Match(1,minSample,maxSample,2);
		Graphtool.SaveSortedList(1,0);
		Graphtool.InitGraphforMatch(nb_samples,firstSample1,seuilDiv,kMT1b,-1);
		Graphtool.CreateMatchCostObject();
		Graphtool.FixWeights(weightMTinTLDiff,TL1TL2Div,flowDiff,divDiff,MIDiff,divMTaMTbCost,timeDiff,meanCalcul);
        Graphtool.Match(1,minSample,maxSample,2);
		Graphtool.SaveSortedList(0,1);
		
	}
	
	//Matching 
	Graphtool.InitGraphforMatch(nb_samples,firstSample1,seuilDiv,kMT1a,kMT1b);
	Graphtool.CreateMatchCostObject();
	Graphtool.FixWeights(weightMTinTLDiff,TL1TL2Div,flowDiff,divDiff,MIDiff,divMTaMTbCost,timeDiff,meanCalcul);
	Graphtool.Match(1,minSample,maxSample,nbMiniBranches);
 	
/*****************learning Parameter Matching***********************************************************************/	
		
	int interestkMT2a;
	int interestkMT2b; 
	int interestfirstSample2;
	int negkMT2a=-1;
	int negkMT2b=-1;
	int negfirstSample2=-1;
	int again=1;
	int posOrNeg=1;
	int newSub=1;
	int erase=0;
	int iteration=0;
    int iterationNeg=0;
	bool scale=true;
	float factorReducList;
	Graphtool.InitParaLearn(1,0);
    if (userAPriori=="y"){Graphtool.PriorMemLearn(weightMTinTLDiff,TL1TL2Div,flowDiff,divDiff,MIDiff,divMTaMTbCost,timeDiff);}
    
	while (newSub==1){	

        /* Matching */
        
 		//Preparing negative semantic matching 
        //     (with negative semantic ref graph as the most unlikely graph of the positive semantic matching)	
        //     (or with user-provided negative sub graph example)
        Graphtool.SaveSortedList(1,1);
		if (negkMT2a==-1){Graphtool.InitGraphforMatchNeg(nb_samples,seuilDiv,kMT1b);}
		else{Graphtool.InitGraphforMatch(nb_samples,negfirstSample2,seuilDiv,negkMT2a,negkMT2b);}
		Graphtool.CreateMatchCostObject();
 		//Fixing negative semantic parameter MMSE estimate
 		//      (initialize with positive semantic parameter MMSE)	
        //      (or remain with learned negative semantic parameter MMSE)
        if (negkMT2a==-1){ 
            if (iteration==0){ Graphtool.FixWeights(weightMTinTLDiff,TL1TL2Div,flowDiff,divDiff,MIDiff,divMTaMTbCost,timeDiff,meanCalcul);}
                else{Graphtool.FixLearnedWeights(1,meanCalcul,1);}  
        }
        else{
            if (iterationNeg<=1){ Graphtool.FixLearnedWeights(1,meanCalcul,1);	}
            else{ Graphtool.FixLearnedWeights(-1,meanCalcul,1); 	}
        }
 		//Negative semantic matching  
		Graphtool.Match(-1,minSample,maxSample,nbMiniBranches);
  
        /* Learning */
        
        //Initializing negative semantic parameter distribution for learning
		//if (iterationNeg<=1){if (negkMT2a!=-1){ Graphtool.CopyPos2NegMemLearn();}}
        if (iteration==0){
            Graphtool.InitParaLearn(0,1);
            Graphtool.CopyPos2NegMemLearn();
        }
		else{
            Graphtool.ScaleParaLearnForSG(1,0,scale);
            Graphtool.ScaleParaLearnForSG(0,1,scale);
        }
		//learning with positive semantic ref graph (only first iteration)
		if (iteration==0){Graphtool.LearnRefGraph(1);}
		else{
            //learning with positive semantic ref graph
            if(negkMT2a==-1){Graphtool.Learn(1,interestkMT2a,interestkMT2b,interestfirstSample2);}
            //learning with negative semantic ref graph
            else{Graphtool.Learn(-1,negkMT2a,negkMT2b,negfirstSample2);}
        }
				
        /* Creating stochastique links & writing output */
        
		Graphtool.PrintProbaLink(); 
		Graphtool.WriteAPostList();
		
		while (again==1){
		
			
			cout<<"LOCATION OF SUB-GRAPH OF INTEREST (1) OR OF NON-INTEREST (-1)?"<<endl;cin >> posOrNeg;
			cout<<"First sample of the sub-graph of interest or non-interest?"<<endl;cin >> interestfirstSample2;interestfirstSample2--;
			cout<<"Index of the MT Class of interest or non-interest?"<<endl;cin >> interestkMT2a;interestkMT2a--;
            if (posOrNeg==-1){negkMT2a=interestkMT2a;negfirstSample2=interestfirstSample2;}
			if (kMT1b!=-1){cout<<"Index of the other MT Class of interest or non-interest?"<<endl;cin >> interestkMT2b;interestkMT2b--;if (posOrNeg==-1){negkMT2b=interestkMT2b;}}
            if (posOrNeg==-1){if (iterationNeg==0){ Graphtool.CopyPos2NegMemLearn();again=0;}iterationNeg++;}
			Graphtool.Learn(posOrNeg,interestkMT2a,interestkMT2b,interestfirstSample2);
			Graphtool.PrintProbaLink();
			Graphtool.WriteAPostList();
           
			if (posOrNeg==-1){
                //if(iterationNeg==0){again=0;} 
                //iterationNeg++;
            }	
            else{
                if (kMT1b==-1){
                    cout<<"Actual maximum a posteriori subgraph: (time sample="<<Graphtool.firstSample2Sorted[0]+1<<", Index MT class="<<Graphtool.kMT2aSorted[0]+1<<"), ";
                    cout<<"do you want to keep the actual reference subgraph: (time sample="<<RefSGTime+1<<", Index MT class="<<RefSGkMT1a+1<<")? (yes='1'/no='0')"<<endl;
                }
                else{
                    cout<<"Actual maximum a posteriori subgraph: (time sample="<<Graphtool.firstSample2Sorted[0]+1<<", Index MT class1="<<Graphtool.kMT2aSorted[0]+1<<", Index MT class2="<<Graphtool.kMT2bSorted[0]+1<<") ";
                    cout<<"do you want to keep the actual reference subgraph: (time sample="<<RefSGTime+1<<", Index MT class1="<<RefSGkMT1a+1<<", Index MT class2="<<RefSGkMT1b+1<<")? (yes='1'/no='0')"<<endl;
                }
                cin >> again;
                if (again==1){cout<<"Erase learning memory (pos='1',neg='-1',both='2'/no='0')?"<<endl;cin >> erase;}
                else{cout<<"Erase negative learning memory (yes='1'/no='0')?"<<endl;cin >> erase;}
                if (erase!=0){Graphtool.CopyPos2NegMemLearn();}
            }
			
		} 

		again=1;
        Graphtool.SaveSortedList(1,1);
		if(iterationNeg!=1||posOrNeg!=-1){
            cout<<"CONTINUE LEARNING WITH NEW REFERENCE SUB-GRAPH ...."<<endl;
            cout<<"First sample of the sub-graph of interest?"<<endl;cin >> interestfirstSample2;
            interestfirstSample2--;RefSGTime=interestfirstSample2;
            cout<<"Index of the MT Class of interest?"<<endl;cin >> interestkMT2a;interestkMT2a--;RefSGkMT1a=interestkMT2a;
            if (kMT1b!=-1){cout<<"Index of the other MT Class of interest?"<<endl;cin >> interestkMT2b;interestkMT2b--;RefSGkMT1b=interestkMT2b;}
            else{interestkMT2b=-1;}
            cout<<"Factor to reduce Sub-Graph list ?"<<endl;
            cin >> factorReducList;scale=(factorReducList==1);
            cout<<"Do you want to fix some specific parameters for the search ('y'/'n')?"<<endl;cin >> userParaSearch;
            if (userParaSearch=="y"){
                cout<<"mean calculation / Kullback-Leibler divergence calculation  (1=mean/0=divergence)?"<<endl;
                cout<<"nunmber of branches considered in the search tree?"<<endl;cin >> nbMiniBranches;
                cin >> aux;if (aux==0){meanCalcul=false;}else{meanCalcul=true;}	
            }	
		
            //Positive semantic matching
            Graphtool.InitGraphforMatch(nb_samples,interestfirstSample2,seuilDiv,interestkMT2a,interestkMT2b);
            Graphtool.CreateMatchCostObject();
            Graphtool.FixLearnedWeights(1,meanCalcul,factorReducList);
            Graphtool.Match(1,minSample,maxSample,nbMiniBranches);
        }
		
		
		iteration++;
	}
	
		
	
/*****************OutPutSaving***********************************************************************/		
/****************************************************************************************************/
	
	time(&t2);
	cout<< "Learning time duration : "<<t2-t1<<" seconds."<<endl;
	return 0;
}
