/*---------------------------------------------------------------------------*/
/*!
 * @file divergence.h
 * @brief  File defining the Divergence object for the  Kullback-Leibler divergence beween  3 dimensionnal Gaussians*/
/*---------------------------------------------------------------------------*/
#ifndef DIVERGENCE_H
#define DIVERGENCE_H

#define INF 1e20;
#define IntLimit 4;

/*! @brief  Object for the  Kullback-Leibler divergence beween  3 dimensionnal Gaussians*/
class Divergence {
  // Associations
  // Attributes
  private:
    /*!@brief  mean vector A*/
    float* MeanA;
    /*!@brief  mean vector B*/		
    float* MeanB;
     /*!@brief  covariance matrix A*/
    float** CovA;	
    /*!@brief  covariance matrix B*/	
    float** CovB; 
    /*!@brief  Inverse covariance matrix A*/
    float** InvCovA;
    /*!@brief  Inverse covariance matrix B*/		
    float** InvCovB;
    /*!@brief  Max bound for integration*/
    float *a;	
    /*!@brief  Min bound for integration*/	
    float *b;
    /*!@brief  auxiliary variable for integration calculus */ 
    float xsav,ysav;  
    /*!@brief  Inverse covariance matrix determinant */ 
    float determinantA, determinantB; 
    /*!@brief  Dimensionnality of the distribution (implemented for 3 dim only) */ 
    int nb_dim;
    /*!@brief  number of dimension with a distribution different from a Dirac*/
    int nb_Gdim;
 
     
  //Operations
  public:
   /*!@brief Constructor : Memory allocation*/
    Divergence(int nb_dimensions);
    /*!@brief Destructor : Memory desallocation*/
    virtual ~Divergence();
    /*!  @brief Initialisation with 2 Gaussian Distributions.
    *
    *    The  components of the multivariate Gaussian distribution which are too narrow are moved to the last dimensions of the covariance matrix. A variable saves the number of dimension presenting a too narrow marginal Gaussian distribution.In this case, the divergence will be equal to a simple multiplication. */
    void Initialise ( float** C_A, float** C_B, float *M_A, float* M_B);
    /*!@brief  Calculation of the integration resolution and the integration bounds.
    
    * For each band, the integration resolution is proportional to the smallest value in the diagonal of the 2 different covariance matrices\n
    The bound of integration are adapted, for each band, to the width of the considered Gaussians. The Dirac special case is also considered, to avoid a future integration  in the next steps and compute instead a multiplication.\n
    If the Gaussians are evaluated too far from each other, a variable greater than 1 is returned by the function to avois a future useless integration.
    @param min_bound output array of min  integration bounds
    @param max_bound output array of max  integration bounds
    @param resolution output array of   integration resolutions
    @param PrecisionFactor input scalar factor for the integration resolutions
    @return number of bands whith too distant Gaussians */
    int BoundResolution_Calculation(float *min_bound,float *max_bound,float *resolution,float PrecisionFactor);
    /*!  @brief Inversion of the covariance matrices using a LU decomposition*/
    void InvertCovMatrix();
    /*!  @brief Rough approximation of Kullback-Leibler distance by Mean vector difference*/
    float MeanDifference();
    /*! @brief  Global method for calculating the ratio of a divergence integration (between a MT and a TL cluster) and  the quantity of pixel belonging to the intersection of the MT and TL spatial classes.
    *
   * The global integration in the 3D space is decomposed. We partition the 3D integration domain into small "cubes", sized according to the previously calculated resolution. All the cube contributions are added to form the global divergence value.\n
   The global integration is performed only in the space composed by the non-Dirac distribution. For the dimensions approximated with a Dirac distribution, because an approximation of band independance, the divergence integration is simplified to a simple multiplication. This product is then added to the global integration value.\n
   The divergence is finally divided by the quantity of pixel belonging to the intersection of the MT and TL spatial classes.
   @param min_bound output array of min  integration bounds
    @param max_bound output array of max  integration bounds
    @param resolution output array of   integration resolutions
    @param weight input quantity of pixel belonging to the intersection of the MT and TL spatial classes
    @return ratio of the Divergence integration and the quantity of pixel belonging to the intersection of the MT and TL spatial classes */
    float Divergence2GaussModel(float *min_bound,float *max_bound,float *resolution, float weight);
  private:
    /*! @brief analytical function to integrate.
    *
    * The function is the kullback-Leibler divergence with 2 Gaussians.*/
    float Div_MultiDimFunc(float x,float y, float z);
    /*! @brief  Method to fix the portion of the integration domain to integrate */
    void  BoundCalculus(float *pas, float *resolution);
    /*!@brief  Integration of a part of the integrable domain.
  
     *The divergence is performed by an analytical integration using a Gaussian approximation. This recursive integration method is inspired from the numerical recepies book. 
     @return integration value*/
    float Integration(); 
    /*!@brief recursive funtions. The 3D function is defined with recursive functions.
    *
    *This recursive integration method is inspired from the numerical recepies book. */
    float f1(float x);
    /*!@brief recursive funtions. The 3D function is defined with recursive functions 
    *
    *This recursive integration method is inspired from the numerical recepies book. */
    float f2(float y);
    /*!@brief recursive funtions. The 3D function is defined with recursive functions
    *
    *This recursive integration method is inspired from the numerical recepies book.  */
    float f3(float z);
    /*!@brief  Function approximating the integral with Gaussian smoothing
    *
    *This integration method is inspired from the numerical recepies book.*/ 
    float Qgaus(float (Divergence::*func)(float),float a, float b);
};

#endif
