/*---------------------------------------------------------------------------*/
/*!
 * @file graphCharac+.hh
 * @brief  file containing the GraphManipulator object able to manipulate collection of sub-graphs included in a graph and drive the supervised learning procedure
 *
 * This file introduce the object methods which are explicitly defined in "readholegraph.cpp" and in "fillsubgraph+.cpp".
 
  @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/
#ifndef TREE_BUIDLER_H
#define TREE_BUIDLER_H
#include "./Image_basics.hh"
#include "./DynaClass.hh"
#include "./divergence.h"
#include "MatchCost+.hh"
#include "ParaSearch.hh"
 /*! @brief  Object designed to manipulate collection of sub-graphs included in a graph and drive the supervised learning procedure.    
 *
 *This object definition present methods defined in "readholegraph.cpp" and in "fillsubgraph+.cpp".\n
  
 *First note that this object posses similarties with the GraphDynaCluster object used previously  for infering the trajectories, with the difference that  the GraphManipulator object does not possess any  methods and attributes related to the  graph inference; It possesses  attributes and methods to read a already created graph, to manipulate sub-graphs, to create sorted collections of sub-graphs with supervised learning methods, etc.... \n
 Second, note this object comprises a large diversity of attributes and methods. This is probably an error : this object should be split in various objects.
 
 @todo redesign this object and the GraphDynaCluster object used for the graph inference so that they share attributes and methods from a commun herited parent object */
class GraphManipulator
{
  
  public:

  // Operations
  
    	GraphManipulator();
	virtual ~GraphManipulator();
	
        /**********************graph reading public methods********************************************************************/
    
       /*! @brief method for reading a graph in format 2 and for reading a  file containing the time delay between the different time samples specific to the graph 
       *@param path Graph file name at format 2  (c.f. Documentation on the "Inference of a graph of dynamic clusters trajectories") 
       @param pathTime Text file name containing, on one line and separated by a white space, the time delay between the different time samples specific to the graph*/  
	void ReadGraphVersion3(char * path,char *pathTime);
     
     
     /**********************graph matching with a parametric similarity model public methods***********************************/
        /*! @brief This method creates a MatchCost Object used for matching*/
        void CreateMatchCostObject();
	/*! @brief This method initializes the matching procedure  relative to the positive semantic definition, with a given reference graph*/
	void InitGraphforMatch(int nb_sample,int firstSample1,float seuilDiv,int kMT1a,int kMT1b);
	/*! @brief This method initializes the matching procedure  relative to the negative semantic definition, with a given reference graph*/
	void InitGraphforMatchNeg(int nb_samples,float seuilDiv, int b);
	/*! @brief This method introduce in the similarity model  a given parameter vector. This method fixes also another parameter specific to the matching calculus.
	*
	*@param weightMTinTLDiff  parameter weighting the population of each nodes				
	*@param	TL1TL2Div parameter weighting the nodes' distribution			
	*@param	flowDiff parameter weighting the flow of pixels exchanged between nodes					
	*@param	divDiff parameter weighting the change in time of the nodes' distribution					
	*@param	MIDiff parameter weighting the intra-node change in time					
	*@param	divMTaMTbCost parameter weighting the interaction of the 2 MT classes (only if 2 MT class are considered);		
	*@param	timeDiff parameter weighting the time lattice on which the sub-graph is defined
	*@param meanCalcul mean calculation '1' or Kullback-Leibler divergence '0' used in the matching calculation */
	void FixWeights(float weightMTinTLDiff,float TL1TL2Div,float flowDiff,float divDiff,float MIDiff,float divMTaMTbCost, float timeDiff,bool meanCalcul);
	/*! @brief This method updates the parameter vector used in the parametric similarity model, with the estimated parameters  which were learned interactively.
	*
	* This method updates  the parameter relative either to the positive  or to the negative semantic. By this method, a couple of paramters are also updated
	@param posOrNeg  parameter vector relative either to the positive '1' or to the negative '-1' semantic
	@param meanCalcul mean calculation '1' or Kullback-Leibler divergence '0' used in the matching calculation  
	@param factorReducList This factor is used to reduce the collection of sub-graphs by factorReducList times 
	 */
	void FixLearnedWeights(int posOrNeg, bool meanCalcul, float factorReducList);
	/*! @brief This method performs the matching of a collection of sub-graphes with a reference sub-graph according to the currently defined similarity model parameter vector.
	*
	* The algorithm runs sequentially the following steps :
	 *\li 1) First a new collection of SearchTree object is created.\n
	 *NB : We focuse either on the positive or on the negative semantic (with the parameter 'posOrNeg'). \n
	 *\n Then, the matching procedure is performed for each sub-graph candidate, one after the other within a loop.  All the possible sub-graphs included in the graph part defined by the user (with the input parameters :'minSample' and 'maxSample') are candidates to be matched with the reference sub-graph. The matching loop is controlled with a while (all the candidates have not been tested) conditions.\n
	 *\n
	 \li 2)  The matching object MatchCost object is then initialisated with a sub-graph candidate. If an a posteriori sub-graph collection was previously inferred, only candidates belonging to  this list are matched with the reference graph.\n
	 *NB : Note that if 2 MT graphs are considered per sub-graph, then for the first matching procedure, only the most likely MT graph associations are preselected according to their rank in the collection. Moreover, 2 identical MT class can not be associated.\n
	 *\li 3) The matching is performed with this candidate using a method of the MatchCost object called 'InexactMatchCost(...)'.\n
	 *NB : Note that a single collection of subgraph is used explictly. The latter is the a posteriori sub-graph collection. For the 2 collections related to the positive and negative semantics,  2 collections of search-tree are saved instead.\n  
	 *\li 4) The index are incremented to focuse on a next sub-graph candidate and the algorithm goes back to step 2). If all the candidates have been matched, the loop is broke and the program exits the method.\n
	 NB: The result produced by this method is only a collection of search-trees
	 *@param posOrNeg matching done for either a positive '1' or negative '-1' example
	 *@param  minSample Minimum sample time considered in the graph
	 *@param  maxSample Maximum sample time considered in the graph
	 *@param  nbMiniBranches number of branches considered in the search tree
	 *@return minimum match-cost among all the sub-graphs candidates*/
	float Match(int posOrNeg, int minSample, int maxSample,int nbMiniBranches);
	/*! @brief Method used to save the a posteriori collection of sub-graphs after a learning step and before beginning a matching procedure which selects candidates of this list. This method is also used to save a sub-graph collection after a matching  procedure with a positive example.
	*
	*The case of saving a sub-graph collection after a matching  procedure with a positive example, is only usefull for the preselection of likely MT assocaitions when considering simultaneously 2 MT class per sub-graph.\n
	*\li if input parameters are (a=1,b=1), the a posteriori collection is saved.\n
	*\li if input parameters are (a=1,b=0), the  collection, resulting from the matching of all possible sub-graphs with the first class of a positive graph example comprising 2 MT classes, is saved.\n
	*\li if input parameters are (a=0,b=1), the  collection, resulting from the matching of all possible sub-graphs with the second class of a positive graph example comprising 2 MT classes, is saved\n*/ 
	void SaveSortedList(int a,int b);
		
     /**********************************Supervised parameter learning public methods*******************************************/
    
    	 /*! @brief Method to 1) calculate, after that the user has fed the system with a positive or negative example, the current estimate of the similarity model parameter vector by calling the ParaSearch method 'MMSEestim(...)'; and 2) calculate the posterior probabilities assigned to each sub-graph candidate and to update the posterior sorted collection of sub-graphs.
	 *
	 *The method runs sequentially the following steps :\n
	 *\li 1) The new estimate of the  the similarity model parameter vector is performed with the new user's example\n
	 *\li 2) The similarity between sub-graphs are updated with the new parameter vector and the maximum of similiraty function is calculated (calculating the range on which the similarity function takes its values)\n
	 *\li 3)After a normalisation step the costs assigned to the sub-graphs matching are converted to probabilities\n
	 *\li 4)Then, the likelihood probabilities relative to the positive and negative semantic are updates and the posterior probabilities are infered by Bayes rule.\n
	 *\li 5)The posterior sub-graph collection is resorted accordingto the new probabilities\n
	 *\li 6) The relevance mesures are updated\n
	 
	 @param posOrNeg learning done for either a positive '1' or negative '-1' example
	 @param interestkMTa index of the MT class provided as an example
	 @param interestkMTb index of a second MT class provided as an example (only in the case when 2 MT class are considered per sub-graph)
	 @param interestfirstSample first time sample index of the graph example   */
	float Learn(int posOrNeg,int interestkMTa,int interestkMTb,int interestfirstSample);	
	 /*! @brief Method calling the 'Learn(..)' method to learn with the as an example the first (positive example) or the last (negative example) element of the current sorted collection of sub-graphs.
	 *
	 *This method is only used once to learn with the first positive example provided by the user.
	 @param learning done for either a positive '1' or negative '-1' example*/
	void LearnRefGraph(int posOrNeg);
	/*! @brief method to print on the standard output the sorted a posteriori collection of sub-graphs.
	*
	*used only duringthe  developpement of the code !!! not usefull anymore*/
	void PrintAPostList();
	/*! @brief Method to write a text file destinated to the GUI, which contains the the relevance feedback measurements and the sorted a posteriori collection of sub-graphs.*/
	void WriteAPostList();	
	/*! @brief Memory allocation and initialization of ParaSearch objects related to the positive or/and negative semantic.
	*
	*The ParaSearch object is used for the estimation of the parameter vector of the similarity model. This method is called once at the first iteration 
	@param a if a='1' the memory allocation and initialization is done for a positive semantic
	@param b if b='1' the memory allocation and initialization is done for a negative semantic  */
	void InitParaLearn(int a,int b);
	/*! @brief Initialization of ParaSearch objects related to the positive or/and negative semantic.
	*
	*The ParaSearch object  is used for the estimation of the parameter vector of the similarity model.
	@param a if a='1' the memory allocation and initialization is done for a positive semantic
	@param b if b='1' the memory allocation and initialization is done for a negative semantic  
	@param scale true if the range on which the matching costs take their values has to be update */ 
	void ScaleParaLearnForSG(int a, int b,bool scale);
	/*! @brief Method to replace by a flat distribution the current multinomial estimated distribution of the similairty model parameters. As the distributions are stored in the ParaSearch object, its method 'ErasePdf(..)' is called.
	*
	*In other words, this methods sets to 1 all the hyperparameters of the dirichlet model related to either the positive or negative semantic learning.
	@param posNeg erase multinomial distribution related to the positive '1' or negative '-1' semantic */
	void EraseMemLearn(int posNeg);
	void PriorMemLearn(float weightMTinTLDiff,float TL1TL2Div,float flowDiff,float divDiff,float MIDiff,float divMTaMTbCost,float timeDiff);
    void CopyPos2NegMemLearn();
    /*! @brief Method to evaluate the divergences measurements between the multinomial distribution related to the positive and negativ semantic. A smoothing of the multinomial distributions by a Gaussian kernel is applied before the symetric Kyullbach-Kleiber divergence calculation.*/ 
	void StrengthProbLink();
	/*! @brief Method to calculate the move of the norm of the parameter vector estimates during the last learning step.*/ 
	void ConvergenceQuality();
	/*! @brief Method to calculate the mean and the variance of the  posterior probabilities attached to the elements of the sub-graph a posteriori collection.*/ 
	void VariancePosterior();
	/*! @brief Method to print on the standard output the relevance feedback measurements*/
	void PrintProbaLink();
	
  // Attributes
  	 
	 /*! @brief number of MT classes */ 
  	 int nb_TimelessClasses;
	 /*! @brief  Minimum/Maximum time sample considered in the graph */ 
	 int minSample,maxSample;
	 /*! @brief Array of time delays between the different time samples specific to the graph */ 
	 float * TimeDelay;
	 /*! @brief number of samples of the time-window*/  
	 int nbSamples;
	 
	 /*! @brief Object used for the estimation of the parameter vector of the similarity model related to a positive semantic*/ 
	 ParaSearch paraSearchObject; 
	 /*! @brief Array of SearchTree objects related to a positive semantic, which is used to save the search tree during a macthing procedure.*/ 
	 SearchTree ****searchTreeCollec;
	 /*! @brief array of distorsion costs related to a positive semantic, assigned to the search tree nodes during the matching procedure*/ 
	 float ****costs;
	 /*! @brief vector of sorted costs or posterior probabilities related to the elements of the collection of sub-graphs*/ 
	 vector<float> costSorted;
	 /*! @brief vector of index of MT class attached to the elements of the collection of sub-graphs*/
	 vector<int> kMT2aSorted;
	 /*! @brief array of the  MT class  indexes attached to the collection of sub-graphs*/
	 int *kMT2aSortedAux;
	 /*! @brief array of  first time sample indexes attached to the collection of sub-graphs*/
	 int *firstSample2aSortedAux;
	  /*! @brief array of  costs attached to the collection of sub-graphs*/
	 float *kMT2aCostAux; 
	 /*! @brief vector of indexes of  2nd MT class  attached to the elements of the collection of sub-graphs. (only when 2 MT classes are considered simultaneously per sub-graph)*/
	 vector<int> kMT2bSorted;
	 /*! @brief array of 2nd MT class indexes attached to the  elements of the collection of sub-graphs (only when 2 MT classes are considered simultaneously per sub-graph)*/
	 int *kMT2bSortedAux;
	 /*! @brief array of  first time sample indexes related to the 2nd MT classes and attached to the elements of the collection of sub-graphs (only when 2 MT classes are considered simultaneously per sub-graph)*/
	 int *firstSample2bSortedAux;
	 /*! @brief array of  costs related to the 2nd MT classes and attached to the elements of the collection of sub-graphs (only when 2 MT classes are considered simultaneously per sub-graph)*/
	 float *kMT2bCostAux;
	 /*! @brief vector of first time sample indexes related to the elements of the collection of sub-graphs*/
	 vector<int> firstSample2Sorted;	 
	 /*! @brief Object used for the estimation of the parameter vector of the similarity model related to a negative semantic*/ 
	 ParaSearch paraSearchObjectNeg;
	 /*! @brief Array of SearchTree objects related to a negative semantic, which is used to save the search tree during a macthing procedure.*/ 
	 SearchTree ****searchTreeCollecNeg;
	 /*! @brief array of distorsion costs related to a negative semantic, assigned to the search tree nodes during the matching procedure*/ 
	 float ****costsNeg;
	 /*! @brief Relevance feedback mesures. Divergence between the positive and negative related parameter distributions*/
	 float * divergences;
	 
  private:

  // Attributes
  	/*! @brief dimensionnality of the  feature space*/
	int nb_dim;
	/*! @brief number of image time samples*/
	int nb_images;
	/*! @brief number of pixels in each image*/
	unsigned long int dim;
	/*!@brief spatial dimension of the images*/
	int nx,ny;
	/*!@brief index of the curent image time sample*/
	int Index_CurrentImage;
		
	/*!@brief DynamicCluster Object for storing dynamic clusters trajectories*/
	DynaClass *TmpClass;
	
	/*!@brief SubGraph Object used to select a 1 st sub-graph for inexact matching*/ 
	//SubGraph *graph1;
    vector<SubGraph> graph1;
     
	/*!@brief SubGraph Object used to select a 2 nd sub-graph for inexact matching*/
	//SubGraph *graph2;
    vector<SubGraph> graph2;
	
	/*!@brief MatchCost Object used for performing the sub-graph matching with the paramteric similarity model */
	vector<MatchCost> matchCostObject;
    //MatchCost *matchCostObject;
	/*!@brief Divergence treshold to fix the graph complexity*/ 
	float seuilDivergence;
	/*!@brief Current number of elements in the posterior sub-graph collection*/
	int sizeList;
	/*!@brief Boolean flag to precise if  a sub-graph collection was previously created or not*/
	bool predefinedList;
	/*!@brief Current factor used to reduce the size of the posterior sub-graph collection */ 
	float factorReducList;
	
	/*!@brief Boolean flag to precise if  a search-tree was previously created by a positive semantic matching */ 
	bool noTree;
	/*!@brief Boolean flag to precise if  a search-tree was previously created by a negative semantic matching */
	bool noTreeNeg;
	/*!@brief Current statistical mean of the  probability attached to the elements of the posterior sub-graph colection */
	float meanPosterior;
	/*!@brief Current statistical variance of the  probability attached to the elements of the posterior sub-graph colection */
	float varPosterior;
	/*!@brief Move of the norm of the parameter vector estimates during the last learning step, for the positive parameter vector estimation */
	float convQualityPos;
	/*!@brief Move of the norm of the parameter vector estimates during the last learning step, for the negative parameter vector estimation */
	float convQualityNeg;
  // Operations
    
      //Graph matching private methods
        /*!@brief Method to load a sub-graph in order to perform after a matching procedure.
	*
	*@param subGraph index of the sub-graph to load (reference graph 1, candidate graph 2)
	@param kMT MT class index 
	@param seuilDiv divergence treshold to consider
	@param firstSample first time sample of the graph example
	@param nb_samples number of samples of the time-window*/ 
	void AddMTtoSubGraph(int subGraph,int kMT,float seuilDiv,int firstSample,int nb_samples);
	
	
	
};
#endif
