/*---------------------------------------------------------------------------*/
/*!
 * @file GaussLUT.cpp
 * @brief Program for the histogram dynamic adaptation of image time-series
 *
 * This procedure transforms, one after the other, the histograms of the different images (or a patial window in the images) composing the time-series into a Gaussian distribution.  For a given spectral band, each image is processed independently from the others.\n
  The algorithm iterates for each image band , the followinng  procdures :
 * \li first the histogram of reflectances related to the current image is equalized using the inverse repartition function of this image; by this procdure the image is requantized from 10 bits (unsigned short) to 16 bits 
 * \li then  the equalized histogram is transformed into a Normal distribution centered and reduced;
 * \li finally, the formed Gaussian histogram defined on 16 bits is rescaled on 8 bits (unsigned char).
 *Finally, it saves multispectral (3 bands) raw images 
 * @author P.Has (IRIT / ENSAE / DLR / CNES)
 * @since January 2005
 *
 */
/*---------------------------------------------------------------------------*/

 

#include <iostream>
#include <string>  
#include <vector>
#include <fstream>

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <ctime>

using namespace std;
#define inf 1e7
#define MIN(x, y) (((x) < (y)) ? (x) : (y))
#define MAX(x, y) (((x) > (y)) ? (x) : (y))
//#define coller(debut,fin) debut ## fin
/*!@brief format of the input images*/ 
#define PIX unsigned short 

/*! 	@brief  Usage function 	
	@param argv0 String containing the imput output file/paramters*/
void usage(std::string argv0) {
    std::cout << "usage: " << ">"
              << argv0 
              << " input_image_file_name output_image_file_name nb_bandsTimesNb_images dimx dimy " << std::endl; 
    std::cout << "usage for cutting: " 
    	      << "column_offset Line_offset DIMY_original"<< std::endl;

}

/*! @brief Gaussian distribution function
*
*Normal distribution function of mean equal to 3 (instead of a zero mean in order to avoid considering negative values) and of standard deviation equal to 1.*/
float gaussian_distribution (float x);
/*!@brief Numerical recepies imported function*/
float qromb(float (*func)(float), float a, float b);

/**************************************************************************/
/*!@brief Main procedure of the program.
*
* The algorithms reads first the input parameters and the input and output files. Then the imput image time-series is read.  The histogram transformations is then done simultaneously for all the images bands by the following processing steps:
\li the repartition function of the image is calculated in order to equalized the image
\li the range of the reflectance values is calculated ( Maximum )
\li the inverse Gaussian repartition function is the calculated using the Numerical recepies function 'qromb(...)'
\li the Gaussian adaptation is then performed
\li the results (3 bands images) are saved on the disk

	@param argv0[1]  input text file name containing on each line the paths of the monospectral image bands to convert
*    	@param argv0[2]  output text file name containing on each line the paths of the 3 bands multispectral images to save on the disk
*	@param argv0[3]  number of spectral bands x number of  images
*	@param argv0[4]  Number of lines of the images
*	@param argv0[5]  Number of columns of the images 
*	
*	Optional parameters for selecting a spatial window  :
*	@param argv0[6]  Column offset when reading the image file 
*	@param argv0[7]  Line offset when reading the image file
*	@param argv0[8]  Number of columns of the image file 
*/
int main(int argc, char *argv []){   

 std::string ifn1;
 std::string ifn2;
 unsigned long   int         nb_bands   = 0;
 unsigned long  int         dimx   = 0;
 unsigned long   int         dimy   = 0;
 unsigned long   int         ofx   = 0;
 unsigned long   int         ofy   = 0;
 unsigned long   int         Dy  = 0;
  
   switch (argc-1) {
   
        case 8:  Dy 			= atoi(argv[8]);
	case 7:  ofy  			= atoi(argv[7]);
        case 6:  ofx   			= atoi(argv[6]);
        case 5:  dimy       		= atoi(argv[5]);
        case 4:  dimx       		= atoi(argv[4]);
	case 3:  nb_bands   		= atoi(argv[3]);
        case 2:  ifn2	 		= std::string(argv[2]);
        case 1:  ifn1	 		= std::string(argv[1]); break;
        case 0:
        default: usage(argv[0]); return -1;
    }
    if ((argc-1) < 5) { usage(argv[0]); return -1;}
    if ((argc-1) == 5) {Dy=dimy;} 
    std::cout << "> " << argv[0] 
              << " " << ifn1 
              << " " << ifn2 
	      << " " << nb_bands 
              << " " << dimx 
              << " " << dimy 
              << " " << ofx
	      << " " << ofy  
              << " " << Dy << std::endl;


unsigned long  int i=0,j=0,k=0,l=0,m=0, n=0;
unsigned long  int dim=dimx*dimy;



/******************************Reading the input_path.txt and output_path.txt files************/
printf("Reading path files\n");

  
string line1;
ifstream inaux;
inaux.open(ifn1.c_str(),ifstream::in);
ifstream in;
in.open(ifn1.c_str(),ifstream::in);
	
n=0;
while(getline(inaux,line1)) {n++;} 	
unsigned long  int nb_images=n/nb_bands;  
unsigned long  int nb_dim=nb_images*nb_bands;

cout << "number of images : " << nb_images <<" "<<nb_bands<<std::endl;
char ***path= new char **[nb_images];
for(m=0;m<nb_images;++m){path[m] =new char*[nb_bands];} 
for(m=0;m<nb_images;++m){for(k=0;k<nb_bands;++k){path[m][k]=new char[100];}}
char ***save_path= new char **[nb_bands/3];
for(m=0;m<nb_bands/3;++m){save_path[m] =new char*[3];} 
for(m=0;m<nb_bands/3;++m){for(k=0;k<3;++k){save_path[m][k]=new char[100];}}

for (i=0;i<(unsigned long int)nb_images;i++){
	for (j=0;j<(unsigned long int)nb_bands;j++){
 		   	getline(in,line1);
		  	strcpy(path[i][j],(char *)line1.c_str());
			std::cout<<path[i][j]<<std::endl;
			
 	}	
}

string line2;
ifstream inaux2;
inaux2.open(ifn2.c_str(),ifstream::in);
ifstream in2;
in2.open(ifn2.c_str(),ifstream::in);

cout<<" GLUT Save Files :"<<endl;				
for (i=0;i<(unsigned long int)nb_bands/3;i++){
	for (j=0;j<(unsigned long int)1;j++){
 		   	getline(in2,line2);
		  	strcpy(save_path[i][j],(char *)line2.c_str());
		  	std::cout<<save_path[i][j]<<std::endl;
 	}	
}



/*******************************Reading the images*********************************************/
printf("Reading images\n");
 
                 
        PIX **a= new PIX *[dimx]; //vector clust contains the data
        for(m=0;m<dimx;++m)
             a[m] =new PIX[dimy];	 
	float **Y= new float *[nb_dim]; //    float data 	
	for(m=0;m<nb_dim;++m)
	{ Y[m]=new float[dim];}
	
       	 char * ifn;
         FILE * ifp;
	 m=0;n=0;
	 for (l = 0 ;l <nb_dim; l++){
	   ifn=path[m][n];printf("%s\n",ifn); // Reading  dimensions R, G and B
	   ifp=NULL;
	   assert (ifp=fopen(ifn,"r"));
	   assert (0==fseek(ifp,(sizeof(PIX)*Dy*ofy+sizeof(PIX)*ofx), SEEK_SET));
	   assert (dimy==fread(&a[0][0],sizeof (PIX),dimy,ifp));    
	   for (i=1;i<dimx;i++){
	     assert (0==fseek(ifp,(sizeof(PIX)*(Dy-dimy)), SEEK_CUR));
	     assert (dimy==fread(&a[i][0],sizeof(PIX),dimy,ifp));
	   }
	   fclose(ifp); 
	   if (n==nb_bands-1){n=0;m++;}
	   else{n++;}
	   for (i = 0 ;i < (dimx); i++){ k=(dimy)*i;          
	     for (j = 0; j <(dimy); j++) { 
	       Y[l][k]=a[i][j];
	       k=k+1;
	     }
	   }
	 }   
	 for(m=0;m<dimx;++m) {delete [] a[m];}
	 delete [] a;            
 /**********************************************************************************************************/ 

unsigned char *sauv= new unsigned char [dim*3];

float *maximum= new float [nb_bands];


/************************************Equalisation for each images by bands*********************************************/ 
  printf("Equalisation\n"); 
  float **Repartition= new float *[nb_bands];
  for(m=0;m<nb_bands;++m)
      { Repartition[m]=new float[1000];}
  for (j=0;j<nb_bands;j++){
    maximum[j]=-inf;    
    for (i=0;i<dim;i++){ 
      for (l=0;l<(nb_dim/nb_bands);l++){

	if  (Y[l*nb_bands+j][i]>maximum[j]) {maximum[j]=Y[l*nb_bands+j][i];}
      }
    }
  }
  for (j=0;j<nb_bands;j++){  printf("%s %d %s %f\n","maximum[",(int)j,"]:",maximum[j]);}
  for (l=0;l<(nb_dim/nb_bands);l++){
    for (i=0;i<nb_bands;i++){for (j=0;j<1000;j++){Repartition[i][j]=0;}}
    for (j=0;j<nb_bands;j++){
      for (i=0;i<dim;i++){
	Repartition[j][(int)(Y[l*nb_bands+j][i])]= Repartition[j][(int)(Y[l*nb_bands+j][i])]+(1.0);
      }
    }
    for (i=0;i<nb_bands;i++){
      Repartition[i][0]=Repartition[i][0]; 
      for (j=1;j<1000;j++){
	Repartition[i][j]=Repartition[i][j]+Repartition[i][j-1];  
      }
    }
    for (j=0;j<nb_bands;j++){
      for (i=0;i<dim;i++){
	Y[l*nb_bands+j][i]= (float)((double)(Repartition[j][(int)Y[l*nb_bands+j][i]])/(double)dim);
      }
    }
  }


/************************************ GAUSS LUT*********************************************/ 
 //adequate for 16 bits with range [0,1]
  printf("%s\n","Gaussian Look up Table");
  float *GRepartition= new float [65536];//16 bits
  float *Inv_repartition= new float [65536];//16 bits
  for (j=0;j<65536;j++){Inv_repartition[j]=0;}
  GRepartition[0]=qromb(&gaussian_distribution,-5,0);
  for (j=1;j<65536;j++){
    GRepartition[j]=GRepartition[j-1]+qromb(&gaussian_distribution,(float)((double)(j-1)*(double)((double)6/(double)65536)),(float)((double)j*(double)((double)6/(double)65536)));
    Inv_repartition[(unsigned long int)((double)GRepartition[j]*(double)65535+(double)0.5)]=j;
  }
  for (j=1;j<65536;j++){  //hole combling
    if (Inv_repartition[j]==0){
      Inv_repartition[j]=Inv_repartition[j-1];   
    }
  }
  for (l=0;l<(nb_dim/nb_bands);l++){
    for (j=0;j<nb_bands;j++){
      for (i=0;i<dim;i++){
	Y[l*nb_bands+j][i]= (unsigned long int)(Inv_repartition[(unsigned long int)((double)Y[l*nb_bands+j][i]*(double)65535+(double)0.5)]/(double)65535*(double)255);// output in 8 bits
     }
    }
  }
  
  printf("Saving results\n");
  for(m=0;m<nb_images;m++){
    for(l=0;l<nb_bands/3;l++){ 
      for (i=0;i<dim;i++){for(k=0;k<3;k++){sauv[i*3+k]=(unsigned char)Y[nb_bands*m+l*3+k][i]; }}
      ifp=NULL; 
      ifn=save_path[l][0];
      assert (ifp=fopen(ifn,"wb"));
      assert ( dim*3==fwrite(&sauv[0], sizeof(unsigned char),dim*3,ifp));
      assert (EOF!=fclose(ifp)); 
    }
  }


/*********************************************************************************************************/


}

/*********************************************************************************************************/
float gaussian_distribution (float x) {  
float s=1;
float m=3;
 float res=1/(sqrt(s*2*M_PI))*exp(-(x-m)*(x-m)/(2*s*s));
return res;
}
/*********************************************************************************************************/
